#module containing object for Iterative Soft-Thresholding with Cooling approach
import python_solver_outcore as solv
import sep_python as sep # import sep functions
import operator_obj as op
import stepper_sample as simp_stpr
import stopper_basic as simp_stop
import math 
import numpy as np
import sys


#Defining the linear problem 1/2*| y - Am |_2 + lambda*| m |_1
class lin_reg_l1_prob(solv.problem):
	"""General linear problem object"""
	def set_prob(self,fwd_cmd_file,adj_cmd_file,op_norm):
		#In this problem we assume that the template series of commands takes one input file (input.H) and spits out one output file (output.H)
		self.op_fwd=op.Operator("operator forward",fwd_cmd_file,"input.H","output.H")
		self.op_adj=op.Operator("operator adjoint",adj_cmd_file,"input.H","output.H")
		self.op_norm=op_norm #Operator Norm necessary for solver
		self.scale_precond = 0.99 * math.sqrt(2) / math.sqrt(self.op_norm); #scaling factor applied to operator A for preconditioning 
		self.lambda_value=None
		return

	def set_lambda(self,lambda_in):
		#In this problem we assume that the template series of commands takes one input file (input.H) and spits out one output file (output.H)
		#Set lambda
		self.lambda_value=lambda_in
		return

	def __del__(self):
		"""Overriding default destructor"""
		solv.problem.__del__(self)
		return


	# define function that computes objective function value
	def objf(self,res):
		"""0.5 norm squared of residual"""
		#Computing separately objective functions
		#data residuals 1/2*| y - Am |_2
		sep.Dot(res[0],res[0],self.obj_list[0])
		cmd="Solver_ops file1=%s scale1_r=0.5 op=scale"%(self.obj_list[0])
		sep.RunShellCmd(cmd)
		#model residuals L1 norm | m |_1
		cmd="Math file1=%s exp='@ABS(file1)' | Stack3d | Stack3d | Stack3d | Stack3d | Stack3d | Stack3d | Stack3d | Transp plane=12 | Stack3d > %s"%(res[1],self.obj_list[1])
		sep.RunShellCmd(cmd)
		#Multiply it by lambda*| m |_1
		cmd="Solver_ops file1=%s scale1_r=%s op=scale"%(self.obj_list[1],self.lambda_value)
		sep.RunShellCmd(cmd)
		#Computing total objective function value
		sep.Add(self.obj_list[0],self.obj_list[1],self.obj)
		obj=self.obj
		return obj

	# define function that computes residuals
	def resf(self,model):
		""" y - alpha * A m = rd (self.res[0]) and m = rm (self.res[1]); alpha = preconditioning factor based on operator norm"""
		self.op_fwd.set_input_output(model,self.res[0])
		#Don't run the forward operator if model is zero vector
		if(sep.Norm_incore(model) != 0.0):
			stat=self.op_fwd.run() #Apply fwd modeling operator
			if(stat!=0): assert False, "problem running forward operator to compute the residual"
			#Applying preconditioning scaling factor
			sep.Scale(self.res[0],self.scale_precond)
		else:
			sep.Cp(self.data,self.res[0])
			sep.Zero(self.res[0]) #Zeroing out a copy of the data
		sep.Sum(self.res[0],self.data,-1.0,1.0)
		#Run regularization part
		sep.Cp(self.model,self.res[1])
		#Return
		res=self.res
		return res

	# define function that projects search direction into data space (Not necessary for ISTC)
	def dresf(self,model,dmodel):
		"""A dm = dr_data (self.dres[0]) and epsilon*A dm = dr_model (self.dres[1])"""
		raise NotImplementedError("dresf is not necessary for ISTC; DO NOT CALL THIS METHOD");
		return 

	# define function to compute gradient (Soft thresholding applied outside in the solver)
	def gradf(self,model,residual):
		"""- A'r_data (residual[0]) = g"""
		#Apply an adjoint modeling
		self.op_adj.set_input_output(residual[0],self.grad)
		stat=self.op_adj.run() #Apply adj modeling operator
		if(stat!=0): assert False, "problem running adjoint operator"
		#Applying preconditioning scaling factor
		sep.Scale(self.grad,-self.scale_precond)
		grad=self.grad
		return grad


#Class for ISTC solver to solve: linear problem 1/2*| y - Am |_2 + lambda*| m |_1
class istc_solver(solv.solver):

	def __init__(self,inner_it,cooling_start,cooling_end,stoppr=simp_stop.stopper_basic(),stpr=solv.stepper()):
		"""Constructor for Iterative Soft Thresholding with Cooling"""
		#Defining stepper and stopper objects as well as ISTC parameters
		self.stoppr=stoppr
		self.stpr=stpr
		if(stoppr.niter<=0): raise ValueError("ERROR! niter for stopper object must be positive and greater than 0!")
		self.outer_it=stoppr.niter #Number of outer iterations taken from the stopper object
		self.inner_it=inner_it #number of inner iterations, the outer iterations are taken care by the stopper
		# cooling_start and cooling_end are numbers between 0 and 1 such that cooling_start <= cooling_end
		if(not (0<=cooling_start<=1) or not (0<=cooling_end<=1) or (cooling_end<cooling_start)): raise ValueError("ERROR! cooling_start and cooling_end must be within [0,1] interval and cooling_start <= cooling_end")
		self.cooling_start=cooling_start # start of cooling continuation as fraction of size of sorted array |A'y| 
		self.cooling_end=cooling_end # end of cooling continuation as fraction of size of sorted array |A'y| 
		#Whether to run steepest descent or not
		self.files_to_clean=[]
		self.removefiles=True
		return

	def __del__(self):
		"""Overriding default destructor"""
		import sep_python as sep
		if hasattr(self, 'removefiles'):
			if (self.removefiles):
				sep.Rm(self.files_to_clean)
			else:
				for ifile in self.files_to_clean:
					print "	Temporary file not removed for debugging in ITSC: %s"%(ifile)
		return


	def run(self,prblm,log_file=None,verbose=True):
		"""Running the solver"""
		#Writing first line in log file if present
		if(not prblm.restart.restarting):
			solv.write_log_file(log_file,info="ITERATIVE SOFT THRESHOLDING with COOLING (ISTC) SOLVER log file")
			solv.write_log_file(log_file,info="Preconditioning scaling factor: %s"%(prblm.scale_precond))
			solv.write_log_file(log_file,info="Restarting folder: %s\n"%(prblm.restart.restart_folder))
			if(not isinstance(prblm,lin_reg_l1_prob)): raise TypeError("Input problem must be a L1-norm regularized problem")
			if(hasattr(self.stpr, 'maxval') and hasattr(self.stpr, 'minval')):
				info = ""
				if(self.stpr.maxval != None): 
					info += "Maximum allowed value for inverted model: %s\n"%(self.stpr.maxval)
				if(self.stpr.minval != None): 
					info += "Minimum allowed value for inverted model: %s\n"%(self.stpr.minval)
				if(info != ""):
					print info
					solv.write_log_file(log_file,info)
		
		#Internal copy of the model
		modl=sep.tmp_file("model_ISTC.H");	self.files_to_clean.append(modl)
		modl0=sep.tmp_file("model_ISTC0.H");	self.files_to_clean.append(modl0)
		#Set internal model file
		modl_prblm=prblm.get_model()
		sep.Cp(modl_prblm,modl)
		sep.Cp(modl_prblm,modl0)
		#Inversion always starts from m = 0 (I need to understand if it is possible to start from m different than 0)
		sep.Zero(modl) 											# modl = 0
		
		if(not prblm.restart.restarting):
			#Creating restart folder
			prblm.restart.create_restart_folder()
			solv.write_log_file(prblm.restart.log_file,info="ITERATIVE SOFT THRESHOLDING with COOLING restart log file\n")	
			#Computing cooling parameters
			grad=prblm.get_grad(modl) 								# Compute the gradient (i.e., grad = - A'y)
			#Computing Cooling parameters by reading values in-core
			grad_arr,_=sep.sep_read_file(grad)
			grad_arr = np.abs(grad_arr.flatten())					# |A'y| and removing zero elements
			grad_arr = grad_arr[np.nonzero(grad_arr)]
			if(grad_arr.size == 0): raise ValueError("ERROR! -- A'y is returning a null vector (i.e., y in the Null space of A')")
			#Sorting the elements in descending order
			grad_arr.sort()
			grad_arr = np.flip(grad_arr,0)
			#Setting fraction of points sampled by the outer loop
			samples = np.round(np.linspace(self.cooling_start,self.cooling_end,self.outer_it)*grad_arr.size)
			samples = samples.astype(np.int)
			#Lambda values to be used during inversion for each outer loop iteration
			lambda_values=grad_arr[samples]
			#Saving lambda values in case of restart restart
			tmp_file_lambda = sep.tmp_file("tmp_lambda.H")
			if(sep.sep_write_file(tmp_file_lambda,[[lambda_values.size,0.,1.,"Outer iterations"]],lambda_values)): raise IOError("ERROR! Problem writing lambda coefficients to file") #Writing a temporary file and then copying it to restart folder
			prblm.restart.write_file(tmp_file_lambda,"lambda_coefficients.H")	
			sep.rm_sep_file(tmp_file_lambda) #removing temporary file for coefficients 
		else:
			#Reading lambda coefficients from restart file
			tmp_file_lambda = sep.tmp_file("tmp_lambda.H")
			prblm.restart.copy_file_from_restart("lambda_coefficients.H",tmp_file_lambda)
			lambda_values,_=sep.sep_read_file(tmp_file_lambda)
			sep.rm_sep_file(tmp_file_lambda) #removing temporary file for coefficients 


		#other temp variables
		iter = 1
		success = True

		#Iteration loop
		while True:
			#Restart inversion from previous run
# 			if(prblm.restart.restarting):
# 				#Restarting the problem and getting iteration number
# 				iter=prblm.restart.set_restart(prblm,log_file)+1
# 				info="Restarting inversion from previous run from: %s"%(prblm.restart.restart_folder)
# 				solv.write_log_file(log_file,info)
# 				if(verbose): print info
# 				prblm.restart.copy_file_from_restart("model_restart_LCG.H",modl)
# 				prblm.restart.copy_file_from_restart("dmodel_restart_LCG.H",dmodl)
# 				#Recomputing delta residuals
# 				dres_prblm=prblm.get_dres(modl,dmodl)
# 				sep.Cp(dres_prblm,dres)
# 				prblm.initial_obj_value = float(prblm.restart.get_info_log("obj",iter_num=0))

			#Setting lambda value for a given outer loop iteration
			prblm.set_lambda(lambda_values[iter-1])
			info = "Outer_iter = %s lambda_value = %s"%(iter-1,lambda_values[iter-1])
			if(verbose): print info
			#Writing on log file
			solv.write_log_file(log_file,info=info)
			solv.write_log_file(prblm.restart.log_file,info="\n"+info)
			inner_iter = 1
			while(inner_iter <= self.inner_it):
				obj=prblm.get_obj(modl) 		#Compute objective function value and put it in a file
				obj0=sep.Get_value(obj) 		#Read the objective function value in the program
				res=prblm.get_res(modl) 		#Compute residuals res = y - Ax
				grad=prblm.get_grad(modl) 		#Compute the gradient g = - A' [y - Ax]
				if(inner_iter == 1):
					info = "	Inner_iter = %s obj = %s residual norm = %s gradient norm= %s feval = %s"%(inner_iter-1,obj0,prblm.get_rnorm(),prblm.get_gnorm(),prblm.get_fevals())
					if(verbose): print info
					#Writing on log file
					solv.write_log_file(log_file,info=info)
					solv.write_log_file(prblm.restart.log_file,info="\n"+info)
					#Check if either objective function value or gradient norm is NaN
					if(math.isnan(obj0) or math.isnan(prblm.get_gnorm())): raise ValueError("Error! Either gradient norm or objective function value NaN!")
				if(prblm.get_gnorm() == 0.):
					print "Gradient vanishes identically"
					break
				#Outputing files if problem was not restarted
				if(not prblm.restart.restarting):
					prblm.output(modl)
					#Removing preconditioning scaling factor from saved inverted model
					sep.Scale(prblm.inverted_model,prblm.scale_precond)
				prblm.restart.restarting=False
				
				#Stepping for internal iteration model update
				sep.Cp(modl,modl0) #Saving model before updating it
				sep.Sum(modl,grad,1.0,-1.0) #Update model x = x + A' [y - Ax]
				#########################################
				#SOFT-THRESHOLDING STEP
				modl_arr,modl_axes = sep.sep_read_file(modl)
				modl_arr = np.sign(modl_arr)*np.clip(np.abs(modl_arr)-prblm.lambda_value,0.,None)
				if(sep.sep_write_file(modl,modl_axes,modl_arr)): raise IOError("ERROR! Problem writing soft-thresholded model vector to file")
				#########################################
				#Apply hard bounds if defined
				clipped=self.stpr.clipping(modl,log_file)

				obj=prblm.get_obj(modl)
				obj1=sep.Get_value(obj)
				if(obj1 >= obj0):
					info = "	Objective function didn't reduce, will terminate solver: obj_new=%s obj_current=%s"%(obj1,obj0)
					if(verbose): print info
					#Writing on log file
					solv.write_log_file(log_file,info=info)
					#Stepping back to the previous solution
					sep.Cp(modl0,modl)
					break

				#Saving current model in case of restart
				prblm.restart.write_file(modl,"model_restart_ISTC.H")
				inner_iter += 1
				#iteration info
				info = "	Inner_iter = %s obj = %s residual norm = %s gradient norm= %s feval = %s"%(inner_iter-1,obj1,prblm.get_rnorm(),prblm.get_gnorm(),prblm.get_fevals())
				if(verbose): print info
				#Writing on log file
				solv.write_log_file(log_file,info="\n"+info)
				solv.write_log_file(prblm.restart.log_file,info="\n"+info)
				#Check if either objective function value or gradient norm is NaN
				if(math.isnan(obj1) or math.isnan(prblm.get_gnorm())): raise ValueError("Error! Either gradient norm or objective function value NaN!")

			iter = iter + 1
			if (self.stoppr.run(iter,prblm,log_file,verbose)): break


		#Writing last model
		prblm.output(modl)
		#Removing preconditioning scaling factor if inverted model movie was requested
		if(prblm.model_movie!=""):
			sep.Scale(prblm.model_movie,prblm.scale_precond)
		#Solver has finished, no need for restart folder
		prblm.restart.clean_problem_restart()
		#Writing on log file
		solv.write_log_file(log_file,info="Solver will terminate\n")
		solv.write_log_file(log_file,info="ITERATIVE SOFT THRESHOLDING with COOLING (ISTC) SOLVER file end")

		return









