!!$=head1 NAME
!!$
!!$Born modeling Adjoint - Full elastic RTM operator.
!!$
!!$=head1 SYNOPSIS
!!$
!!$Bornfast_elastic_adj.x <model.H wavelet=wavelet.H sources=input_sources.p pertdata=delta_data.H recs=input_recs.p wantwavefield=0 readwavefield=0 wavefield=wavefield.H wavefieldr=wavefieldr.H >perturbations.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Elastic RTM operator for velocity-stress formulation. Based on both Vireux's (1986) and Alves' (2015) works.
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history
!!$
!!$B<filename> I<model.H> - model parameters in SEP3D format (axis order: z,x,y,parameter).
!!$
!!$=item from par
!!$
!!$B<integer> I<mod_par> - [0] parameterization: 0 -> lame[GPa], mu[GPa], rho[Kg/m3]; 1 -> Vp[m/s], Vs[m/s], rho[Kg/m3]; 2 -> Ip[N s/m3], Is[N s/m3],rho[Kg/m3] 
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated for each component (Sigmaxx,sigmazz,sigmaxz,vx,vz).
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<filename> I<pertdata> - [delta_data.H] File containing data perturbations (time,Sigmaxx,sigmazz,sigmaxz,vx,vz,receivers,sources)
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the source and back-propagated wavefields.
!!$
!!$B<integer> I<readwavefield> - [0] If 1 the program reads the source wavefield from file.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save (wantwavefield=1) or read (readwavefield=1) only velocity and their time derivatives (x,z,dxdt,dzdt)
!!$
!!$B<filename> I<wavefieldr> - [wavefieldr.H] File where to save back-propagated data wavefield if wantwavefield=1 (x,z,dxdt,dzdt,sigmaxx,sigmazz,sigmaxz)
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<pressure> - [0] If 1 it will save the pressure as well in the written wavefields
!!$
!!$B<integer> I<verbose> - [0] Program will write propagation information
!!$
!!$B<integer> I<free_surf> - [0] Applying free-surface boundary conditions at the surface
!!$
!!$B<integer> I<mod_padded> - [0] Model parameters were padded (120 elements), useful for FWI. If free_surf is requested pad model by 5 samples at the top.
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Bornfast_elastic_adj


  !Used modules
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use resample
  use S_mod
  use Prop_elastic_step
  use Sou_scale_acoustic
  use Add_source2D
  use absorbing
  use full_elastic_imaging_mod

  implicit none
  character*256											               	   :: sources, recs !Files containing source and receiver positions
  integer												                   :: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													                   :: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling
  integer												                   :: nw !Time sample of the source
  real													                   :: ow, dw !Origin and sampling of the source
  real, dimension(:), allocatable						       			   :: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					     			   :: lame, mu, rho  !lame, shear modulus and density models
  real, dimension(:), allocatable						       			   :: wavelet_xx, wavelet_zz, wavelet_xz, wavelet_vx, wavelet_vz !Source wavelets for the different (sigmas,velocities) components
  real, dimension(:,:), allocatable						     			   :: recfield_xx_fine, recfield_zz_fine, recfield_xz_fine, recfield_vx_fine, recfield_vz_fine !Receiver field at fine time scale
  real, dimension(:), allocatable						       			   :: recfield !Receiver field to write
  integer												                   :: wc, srite, rc, sreed
  integer												                   :: wantwavefield, num_threads
  logical												                   :: twoD, inj_xx=.false., inj_zz=.false., inj_xz=.false., inj_vx=.false., inj_vz=.false.
  integer												                   :: is, ig !Cycles variables
  integer												                   :: it2, itw, it
  !Arrays needed for propagation
  real, dimension(:,:,:), allocatable, target			 				   :: cur_xx, old_xx, cur_zz, old_zz, cur_xz, old_xz, cur_vx, old_vx, cur_vz, old_vz
  real, dimension(:,:,:), pointer						       			   :: pnew_xx, pcur_xx, pold_xx, pnew_zz, pcur_zz, pold_zz, pnew_xz, pcur_xz, pold_xz, pnew_vx, pcur_vx, pold_vx, pnew_vz, pcur_vz, pold_vz, ptemp
  real, dimension(:,:,:), pointer						       			   :: lame_pad, mu_pad, rho_pad !Padded models
  real													                   :: dt, dt_inv !Temporal sampling of the propagator
  real													                   :: maxvel, minvel !max and min p-velocity
  real													                   :: minvels, minimum_vel !min s-velocity
  real, dimension(:), allocatable					         			   :: wavelet_xx_fine, wavelet_zz_fine, wavelet_xz_fine, wavelet_vx_fine, wavelet_vz_fine !Source wavelet fine grid
  real, dimension(:), allocatable						       			   :: scaled_vx_wavelet, scaled_vz_wavelet
  integer												                   :: nt !Number of time samples for the propagation
  integer												                   :: nv_pad, pad_offset !Padding for the models and offset from absorbing boundary
  integer												                   :: desample
  integer												                   :: verbose, pressure, free_in
  logical                                          						   :: free=.false.
  real, dimension(:,:), allocatable						     			   :: interp_filter
  real, dimension(:,:,:), allocatable					     			   :: rhoz, rhox, rhoy !Staggered density
  real													                   :: cr
  integer												                   :: count1, count2,ierr, icount, irw
  real, dimension(:,:,:), pointer						       			   :: pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt
  integer												                   :: sk, sseek_block
  !Variables for Born modeling
  integer												                   :: readwavefield, mod_padded, mod_par
  real, dimension(:,:,:), pointer						       			   :: pertlame_pad, pertmu_pad, pertrho_pad   !lame, mu, and rho perturbations padded
  real, dimension(:,:,:), pointer						       			   :: pertrhoz, pertrhox, pertrhoy, pertmuxz !Staggered grid perturbations
  real, dimension(:,:,:,:), pointer						     			   :: pvx, pvz, pvxdt, pvzdt
  real, dimension(:,:,:), pointer						       			   :: dvx_r, dvx_l, dvz_r, dvz_l, dxx_r, dxx_l, dzz_r, dzz_l, dxz_r, dxz_l
  real, dimension(:,:,:), pointer						       			   :: pscat_vxdt_r, pscat_vxdt_l, pscat_vzdt_r, pscat_vzdt_l !Pointer for scattered wavefield interpolation
  real													                   :: scale_scat





  parameter( nv_pad=120, pad_offset=20) !Padding parameters for the absording boundary
  call sep_init (SOURCE)


  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("readwavefield",readwavefield,0)
  call from_par ("mod_padded",mod_padded,0)
  call from_par ("verbose",verbose,0)
  call from_par ("dt",dt,0.)
  call from_par ("pressure",pressure,0)
  call from_par ("mod_par",mod_par,0)
  !Free surface control
  call from_par ("free_surf",free_in,0)
  if (free_in .ne. 0) free = .true.

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)"####################      Elastic 2D wave propagator     #####################"
  if(verbose)write(0,*)"####################      Adjoint Born modeling (RTM)    #####################"
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)



  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading the source wavelets for each component (sigmaxx, sigmazz, sigmaxz, vx, vz)
  allocate(wavelet_xx(nw),wavelet_zz(nw),wavelet_xz(nw),wavelet_vx(nw),wavelet_vz(nw))
  wavelet_xx = 0.; wavelet_zz = 0.; wavelet_xz = 0.; wavelet_vx = 0.; wavelet_vz = 0.;
  rc=sreed("wavelet" , wavelet_xx , nw*4)
  rc=sreed("wavelet" , wavelet_zz , nw*4)
  rc=sreed("wavelet" , wavelet_xz , nw*4)
  rc=sreed("wavelet" , wavelet_vx , nw*4)
  rc=sreed("wavelet" , wavelet_vz , nw*4)

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the lame, shear modulus and rho models and padding if necessary
  if(mod_padded .eq. 0) then
  	!Reading the lame modulus
  	allocate(lame(nz,nx,ny))
  	lame(:,:,:) = 0.
  	rc=sreed("in" , lame , nz*nx*ny*4)
  	!Reading the shear modulus
  	allocate(mu(nz,nx,ny))
  	mu(:,:,:) = 0.
   	rc=sreed("in" , mu , nz*nx*ny*4)
  	!Reading the density model
  	allocate(rho(nz,nx,ny))
  	rho(:,:,:) = 0.
  	rc=sreed("in" , rho , nz*nx*ny*4)
  	!Adding absorbing boundaries to the lame model
  	call Pad_velocity(lame,lame_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD,free)
  	!Adding absorbing boundaries to the mu model
  	call Pad_perturbation(mu,mu_pad,nv_pad,free)
  	!Adding absorbing boundaries to the density model
  	call Pad_perturbation(rho,rho_pad,nv_pad,free)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	!Reading the lame modulus
  	allocate(lame_pad(nz,nx,ny))
  	lame_pad(:,:,:) = 0.
  	rc=sreed("in" , lame_pad , nz*nx*ny*4)
  	!Reading the shear modulus
  	allocate(mu_pad(nz,nx,ny))
  	mu_pad(:,:,:) = 0.
  	rc=sreed("in" , mu_pad , nz*nx*ny*4)
  	!Reading the density model
  	allocate(rho_pad(nz,nx,ny))
  	rho_pad(:,:,:) = 0.
  	rc=sreed("in" , rho_pad , nz*nx*ny*4)
  endif

  !Converting model parameters to lame/mu/rho if they were provided in a different space
  if(mod_par .eq. 0) then
  	lame_pad(:,:,:) = lame_pad(:,:,:)*1e9 ! Conversion from GPa to Pa
  	mu_pad(:,:,:) = mu_pad(:,:,:)*1e9     ! Conversion from GPa to Pa
  elseif(mod_par .eq. 1) then
  	!Converting Vp/Vs/rho to lame/mu/rho
  	mu_pad(:,:,:) = mu_pad(:,:,:)*mu_pad(:,:,:)*rho_pad(:,:,:)
  	lame_pad(:,:,:) = lame_pad(:,:,:)*lame_pad(:,:,:)*rho_pad(:,:,:) - 2.0*mu_pad(:,:,:)
  elseif(mod_par .eq. 2) then
  	!Converting Ip/Is/rho to lame/mu/rho
  	mu_pad(:,:,:) = mu_pad(:,:,:)*mu_pad(:,:,:)/rho_pad(:,:,:)
  	lame_pad(:,:,:) = lame_pad(:,:,:)*lame_pad(:,:,:)/rho_pad(:,:,:) - 2.0*mu_pad(:,:,:)
  endif
  
  !Computing max and min velocities
  minvel = minval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  maxvel = maxval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  if (any(mu_pad .eq. 0.0)) then
  	minvels = 0.0
  	minimum_vel = minvel
  else
  	!minvels = minval(sqrt(mu_pad/rho_pad), mask=(sqrt(mu_pad/rho_pad).gt.0.))
  	minvels = minval(sqrt(mu_pad/rho_pad))
  	minimum_vel = minvels
  endif

  !Allocating model perturbations
  allocate(pertlame_pad(nz,nx,ny));pertlame_pad(:,:,:) = 0.
  allocate(pertmu_pad(nz,nx,ny));pertmu_pad(:,:,:) = 0.
  allocate(pertrho_pad(nz,nx,ny));pertrho_pad(:,:,:) = 0.




  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		desample = desample*2
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.25*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		desample = desample*2
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
  	dt = dw/desample
  	desample = desample*2
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minimum_vel/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  if(verbose)write(0,*)"	Min S vel [m/s]:",minvels
  if(verbose)write(0,*)"	Min P vel [m/s]:",minvel
  if(verbose)write(0,*)"	Max P vel [m/s]:",maxvel
  if(verbose)write(0,*) ""


  !Necessary for time derivative
  dt_inv =1.0/dt
  !Scaling source wavelet by 1/(dz*dx) for 2D (discretization of delta function!)
  wavelet_xx=wavelet_xx/(dz*dx)
  wavelet_zz=wavelet_zz/(dz*dx)
  wavelet_xz=wavelet_xz/(dz*dx)
  wavelet_vx=wavelet_vx/(dz*dx)
  wavelet_vz=wavelet_vz/(dz*dx)


  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw*2 !Because we advance only dt/2 in this implementation
  	 desample = 1
  	 allocate(wavelet_xx_fine(nt),wavelet_zz_fine(nt),wavelet_xz_fine(nt),wavelet_vx_fine(nt),wavelet_vz_fine(nt))
  	 allocate(scaled_vx_wavelet(nt),scaled_vz_wavelet(nt))
  	 !scaling source wavelets by the dt used for propagation
  	 wavelet_xx_fine = dt*wavelet_xx;wavelet_zz_fine = dt*wavelet_zz;wavelet_xz_fine = dt*wavelet_xz;wavelet_vx_fine = dt*wavelet_vx;wavelet_vz_fine = dt*wavelet_vz;
  	 if (any(wavelet_xx_fine .ne. 0.0)) inj_xx=.true.
  	 if (any(wavelet_zz_fine .ne. 0.0)) inj_zz=.true.
  	 if (any(wavelet_xz_fine .ne. 0.0)) inj_xz=.true.
  	 if (any(wavelet_vx_fine .ne. 0.0)) inj_vx=.true.
  	 if (any(wavelet_vz_fine .ne. 0.0)) inj_vz=.true.
  else
   	nt = (nw-1)*desample+1
    	!Source wavelets interpolation to fine time grid
  	allocate(wavelet_xx_fine(nt),wavelet_zz_fine(nt),wavelet_xz_fine(nt),wavelet_vx_fine(nt),wavelet_vz_fine(nt))
  	allocate(scaled_vx_wavelet(nt),scaled_vz_wavelet(nt))
  	call resample_init(nt, nw, desample)
  	!scaling source wavelets by the dt used for propagation
  	if (any(wavelet_xx .ne. 0.0)) then
  		call resample_op(.false.,.false.,dt*wavelet_xx,wavelet_xx_fine)
  		inj_xx=.true.
  	endif
  	if (any(wavelet_zz .ne. 0.0)) then
  		call resample_op(.false.,.false.,dt*wavelet_zz,wavelet_zz_fine)
  		inj_zz=.true.
  	endif
  	if (any(wavelet_xz .ne. 0.0)) then
  		call resample_op(.false.,.false.,dt*wavelet_xz,wavelet_xz_fine)
  		inj_xz=.true.
  	endif
  	if (any(wavelet_vx .ne. 0.0)) then
  		call resample_op(.false.,.false.,dt*wavelet_vx,wavelet_vx_fine)
  		inj_vx=.true.
  	endif
  	if (any(wavelet_vz .ne. 0.0)) then
  		call resample_op(.false.,.false.,dt*wavelet_vz,wavelet_vz_fine)
  		inj_vz=.true.
  	endif
  endif

  !Initialzing the fixed-variable routines
  call S_init(nz,nx,ny,twoD)
  call absorbing_init(nz, nx, ny, nv_pad, pad_offset,twoD,free)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_xx_fine(nt,ng),recfield_zz_fine(nt,ng),recfield_xz_fine(nt,ng),recfield_vx_fine(nt,ng),recfield_vz_fine(nt,ng))
  recfield = 0.
  recfield_xx_fine = 0.; recfield_zz_fine = 0.; recfield_xz_fine = 0.; recfield_vx_fine = 0.; recfield_vz_fine = 0.

  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))

  !Scale for scattering
  scale_scat=1.0*dt

  if (verbose)then
    if(free)write(0,*)"Requested free surface boundary conditions"
  endif

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old_xx(nz,nx,ny),cur_xx(nz,nx,ny),old_zz(nz,nx,ny),cur_zz(nz,nx,ny),old_xz(nz,nx,ny),cur_xz(nz,nx,ny),old_vx(nz,nx,ny),cur_vx(nz,nx,ny),old_vz(nz,nx,ny),cur_vz(nz,nx,ny))
  allocate(pxx_dt(nz,nx,ny), pzz_dt(nz,nx,ny), pxz_dt(nz,nx,ny), pvx_dt(nz,nx,ny), pvz_dt(nz,nx,ny))
  pold_xx => old_xx; pold_zz => old_zz; pold_xz => old_xz; pold_vx => old_vx; pold_vz => old_vz
  pcur_xx => cur_xx; pcur_zz => cur_zz; pcur_xz => cur_xz; pcur_vx => cur_vx; pcur_vz => cur_vz
  pnew_xx => old_xx; pnew_zz => old_zz; pnew_xz => old_xz; pnew_vx => old_vx; pnew_vz => old_vz
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""

  !Staggering density
  allocate(rhoz(nz,nx,ny), rhox(nz,nx,ny))
  call Sz_op(.true., .false., rhoz, rho_pad)
  call Sx_op(.true., .false., rhox, rho_pad)
  !Staggering on y if not twoD problem
  if(.not. twoD) then
  	allocate(rhoy(nz,nx,ny))
  	call Sy_op(.true., .false., rhoy, rho_pad)
  endif

  !allocating memory for linear interpolation of velocity wavefields
  allocate(pvx(nz,nx,ny,nw), pvz(nz,nx,ny,nw), pvxdt(nz,nx,ny,nw), pvzdt(nz,nx,ny,nw))
  !allocating memory for linear interpolation of velocity and stress scattered wavefields
  allocate(dvx_r(nz,nx,ny), dvx_l(nz,nx,ny), dvz_r(nz,nx,ny), dvz_l(nz,nx,ny), dxx_r(nz,nx,ny), dxx_l(nz,nx,ny), dzz_r(nz,nx,ny), dzz_l(nz,nx,ny), dxz_r(nz,nx,ny), dxz_l(nz,nx,ny))

  !Initializing elastic step and imaging modules
  call Prop_elastic_step_init(nz, nx, ny, nt, nv_pad, dt, dz, dx, dy, twoD, rhox, rhoz, lame_pad, mu_pad, free)
  call full_elastic_imaging_init(nz, nx, ny, dz, dx, dy, twoD)

  !Inverting staggered density
  rhox = 1.0/rhox
  rhoz = 1.0/rhoz
  if(.not. twoD) rhoy = 1.0/rhoy
  !Staggering density and mu perturbations
  allocate(pertrhoz(nz,nx,ny), pertrhox(nz,nx,ny), pertmuxz(nz,nx,ny))

  !Staggering on y if not twoD problem
  if(.not. twoD) then
  	allocate(pertrhoy(nz,nx,ny))
  endif
  !Array to interpolating scattered wavefields
  if (wantwavefield) then
  	pscat_vxdt_r => lame_pad; pscat_vxdt_l => mu_pad
  	allocate(pscat_vzdt_r(nz,nx,ny), pscat_vzdt_l(nz,nx,ny))
  endif


  if (wantwavefield) then
  	if(verbose)write(0,*) "#Program will be writing wavefield as requested#"
  		do is = 1,ns
  			do it = 1, nw
  				!Particle Velocities
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				wc=srite("wavefieldr",pcur_vz,nz*nx*ny*4)
  				!Particle Velocities time derivative
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				!Stress components
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  				if(pressure)wc=srite("wavefieldr",pcur_vx,nz*nx*ny*4)
  			enddo
  		enddo
  endif

  do is = 1,ns
  	call system_clock(count_rate=cr)
  	call system_clock(count1)
  	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
  	if(.not. readwavefield) then
  		old_xx = 0.; cur_xx = 0.
  		old_zz = 0.; cur_zz = 0.
  		old_xz = 0.; cur_xz = 0.
  		old_vx = 0.; cur_vx = 0.
  		old_vz = 0.; cur_vz = 0.
  		!Zeroing the source wavefields
  		pvx = 0.
  		pvz = 0.
  		pvxdt = 0.
  		pvzdt = 0.
		!Scaling the source for velocity
		if (inj_vx .or. inj_vz) then
			scaled_vx_wavelet = wavelet_vx_fine
			scaled_vz_wavelet = wavelet_vz_fine
			call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, 1.0, twoD)
			call Sou_scale_acoustic_op(scaled_vx_wavelet,1.0/rho_pad) !Adjoint = Forward (input-output same variable)
			call Sou_scale_acoustic_op(scaled_vz_wavelet,1.0/rho_pad)
		endif

		if(verbose)write(0,*)"Propagating source wavefield..."
    !First current pressure wavefield
    if (wantwavefield) then
      if(pressure) wc=srite("wavefield",0.5*(pcur_xx+pcur_zz),nz*nx*ny*4)
    endif


		do itw = 1, nw-1 !Forward
			it = (itw-1)*desample
			!Propagate source field
			do it2 = 1, desample
				!computing time derivative for each component
				call Prop_elastic_step_op(.false., pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz, pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt)
				!Adding source to the time derivative of each component
				!Adding sigmaxx source
				call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
				if (inj_xx) then
					call Add_source2D_op(.true.,.true.,pxx_dt,wavelet_xx_fine(it+it2))
				endif
				!Adding sigmazz source
				if (inj_zz) then
					call Add_source2D_op(.true.,.true.,pzz_dt,wavelet_zz_fine(it+it2))
				endif
				!Adding sigmaxz source
				if (inj_xz) then
					call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz-(dz/2.0), ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pxz_dt,wavelet_xz_fine(it+it2))
				endif
				!Adding vx source
				if (inj_vx) then
					call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pvx_dt,scaled_vx_wavelet(it+it2))
				endif
				!Adding vz source
				if (inj_vz) then
					call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz-(dz/2.0), ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pvz_dt,scaled_vz_wavelet(it+it2))
				endif


				!Apply adjoint(fine to coarse) of linear interpolation
				!On Velocities
				pvx(:,:,:,itw) = pvx(:,:,:,itw) + pcur_vx*interp_filter(it2,1)
				pvx(:,:,:,itw+1) = pvx(:,:,:,itw+1) + pcur_vx*interp_filter(it2,2)
				pvz(:,:,:,itw) = pvz(:,:,:,itw) + pcur_vz*interp_filter(it2,1)
				pvz(:,:,:,itw+1) = pvz(:,:,:,itw+1) + pcur_vz*interp_filter(it2,2)
				!On derivatives
				pvxdt(:,:,:,itw) = pvxdt(:,:,:,itw) + pvx_dt*interp_filter(it2,1)*dt_inv
				pvxdt(:,:,:,itw+1) = pvxdt(:,:,:,itw+1) + pvx_dt*interp_filter(it2,2)*dt_inv
				pvzdt(:,:,:,itw) = pvzdt(:,:,:,itw) + pvz_dt*interp_filter(it2,1)*dt_inv
				pvzdt(:,:,:,itw+1) = pvzdt(:,:,:,itw+1) + pvz_dt*interp_filter(it2,2)*dt_inv


				!Computing the new time step
				pnew_xx = pold_xx + pxx_dt
				pnew_zz = pold_zz + pzz_dt
				pnew_xz = pold_xz + pxz_dt
				pnew_vz = pold_vz + pvz_dt
				pnew_vx = pold_vx + pvx_dt


				!Applying absorbing boundary condition
				call absorbing_op(pnew_xx)
				call absorbing_op(pnew_zz)
				call absorbing_op(pnew_xz)
				call absorbing_op(pnew_vx)
				call absorbing_op(pnew_vz)


				!Swapping pointers
				pold_xx => pcur_xx
				pcur_xx => pnew_xx
				pnew_xx => pold_xx
				pold_zz => pcur_zz
				pcur_zz => pnew_zz
				pnew_zz => pold_zz
				pold_xz => pcur_xz
				pcur_xz => pnew_xz
				pnew_xz => pold_xz
				pold_vx => pcur_vx
				pcur_vx => pnew_vx
				pnew_vx => pold_vx
				pold_vz => pcur_vz
				pcur_vz => pnew_vz
				pnew_vz => pold_vz

			enddo

			!saving source wavefield
			if (wantwavefield) then
				wc=srite("wavefield",pvx(:,:,:,itw),nz*nx*ny*4)
				wc=srite("wavefield",pvz(:,:,:,itw),nz*nx*ny*4)
				wc=srite("wavefield",pvxdt(:,:,:,itw),nz*nx*ny*4)
				wc=srite("wavefield",pvzdt(:,:,:,itw),nz*nx*ny*4)
				if(pressure) wc=srite("wavefield",0.5*(pcur_xx+pcur_zz),nz*nx*ny*4)
			endif

			if (mod(itw,(nw/10))==0) then
    		if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
    		close(0)
  		endif


		enddo

			!Last step of the source wavefield just saving source wavefield
			pvx(:,:,:,nw) = pvx(:,:,:,nw) + pcur_vx*interp_filter(1,1)
			pvz(:,:,:,nw) = pvz(:,:,:,nw) + pcur_vz*interp_filter(1,1)
			pvzdt(:,:,:,nw) = pvzdt(:,:,:,nw) + pvz_dt*interp_filter(1,1)*dt_inv
			pvxdt(:,:,:,nw) = pvxdt(:,:,:,nw) + pvx_dt*interp_filter(1,1)*dt_inv
			if (wantwavefield) then
				wc=srite("wavefield",pvx(:,:,:,nw),nz*nx*ny*4)
				wc=srite("wavefield",pvz(:,:,:,nw),nz*nx*ny*4)
				wc=srite("wavefield",pvxdt(:,:,:,nw),nz*nx*ny*4)
				wc=srite("wavefield",pvzdt(:,:,:,nw),nz*nx*ny*4)
			endif

			if(verbose)write(0,'(A4)',advance="no")"DONE"
       	 close(0)
		if(verbose)write(0,'(a)',advance="yes")""
		if(verbose)write(0,*)"End Propagation source wavefield"
	else
		if(verbose)write(0,*)"Reading source wavefield"
		do itw = 1, nw
			rc=sreed("wavefield",pvx(:,:,:,itw),nz*nx*ny*4)
			rc=sreed("wavefield",pvz(:,:,:,itw),nz*nx*ny*4)
			rc=sreed("wavefield",pvxdt(:,:,:,itw),nz*nx*ny*4)
			rc=sreed("wavefield",pvzdt(:,:,:,itw),nz*nx*ny*4)
		enddo

	endif

	if(verbose)write(0,*)"Propagating adjoint wavefield..."

	!Scattered wavefield section
	icount = 0
	if (wantwavefield) then
		if (pressure) then
			sk = sseek_block("wavefieldr",8*nw*(is),4*nz*nx*ny,0)
		else
			sk = sseek_block("wavefieldr",7*nw*(is),4*nz*nx*ny,0)
		endif
	endif
	pertrhox=0.;pertrhoz=0.;pertlame_pad=0.;pertmu_pad=0.;pertmuxz=0.
	old_xx = 0.; cur_xx = 0.
	old_zz = 0.; cur_zz = 0.
	old_xz = 0.; cur_xz = 0.
	old_vx = 0.; cur_vx = 0.
	old_vz = 0.; cur_vz = 0.
	dvx_r=0.; dvx_l=0.;
	dvz_r=0.; dvz_l=0.;
	dxx_r=0.; dxx_l=0.;
	dzz_r=0.; dzz_l=0.;
	dxz_r=0.; dxz_l=0.
	if (wantwavefield) then
		pscat_vxdt_r = 0.; pscat_vxdt_l = 0.
		pscat_vzdt_r = 0.; pscat_vzdt_l = 0.
	endif

	!Creating adjoint source at the receivers
	if (dw .le. dt) then
		do ig = 1,ng
			rc=sreed("pertdata",recfield,nw*4)
			recfield_xx_fine(:,ig) = recfield
			rc=sreed("pertdata",recfield,nw*4)
			recfield_zz_fine(:,ig) = recfield
			rc=sreed("pertdata",recfield,nw*4)
			recfield_xz_fine(:,ig) = recfield
			rc=sreed("pertdata",recfield,nw*4)
			recfield_vx_fine(:,ig) = recfield
			rc=sreed("pertdata",recfield,nw*4)
			recfield_vz_fine(:,ig) = recfield
		enddo
	else
		do ig = 1,ng
			rc=sreed("pertdata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_xx_fine(:,ig))
			rc=sreed("pertdata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_zz_fine(:,ig))
			rc=sreed("pertdata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_xz_fine(:,ig))
			rc=sreed("pertdata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_vx_fine(:,ig))
			rc=sreed("pertdata",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_vz_fine(:,ig))
		enddo
	endif
	!Checking with component is non-zero
	if (any(recfield_xx_fine .ne. 0.0)) inj_xx=.true.
	if (any(recfield_zz_fine .ne. 0.0)) inj_zz=.true.
	if (any(recfield_xz_fine .ne. 0.0)) inj_xz=.true.
	if (any(recfield_vx_fine .ne. 0.0)) inj_vx=.true.
	if (any(recfield_vz_fine .ne. 0.0)) inj_vz=.true.

  !First current pressure wavefield
  if (wantwavefield) then
    if(pressure) then
      sk = sseek_block("wavefieldr",-8,4*nz*nx*ny,1)
      wc = srite("wavefieldr",0.5*(pcur_xx+pcur_zz),nz*nx*ny*4)
    endif
  endif

	do itw = nw, 2, -1 !Adjoint
		it = (itw-2)*desample
		!Propagate receiver field
		do it2 = desample+1, 2, -1


			!Applying absorbing boundary condition
			call absorbing_op(pnew_xx)
			call absorbing_op(pnew_zz)
			call absorbing_op(pnew_xz)
			call absorbing_op(pnew_vx)
			call absorbing_op(pnew_vz)


			!computing time derivative for each component
			call Prop_elastic_step_op(.true., pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz, pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt)

			!Adding receiver data to the time derivative of each component
			if (inj_xx .or. inj_zz) then
				do ig = 1, ng
					!Adding sigmaxx and sigmazz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pxx_dt,recfield_xx_fine(it+it2,ig))
					call Add_source2D_op(.true.,.true.,pzz_dt,recfield_zz_fine(it+it2,ig))
				enddo
			endif

			if (inj_xz) then
				do ig = 1, ng
					!Adding sigmaxz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz-(dz/2.0), ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pxz_dt,recfield_xz_fine(it+it2,ig))
				enddo
			endif

			if (inj_vx) then
				do ig = 1, ng
					!Adding sigmavx receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pvx_dt,recfield_vx_fine(it+it2,ig))
				enddo
			endif

			if (inj_vz) then
				do ig = 1, ng
					!Adding sigmavz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz-(dz/2.0), ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pvz_dt,recfield_vz_fine(it+it2,ig))
				enddo
			endif



			!Apply adjoint(fine to coarse) of linear interpolation
			!On Velocities and stress
			dvx_l = dvx_l + pcur_vx*interp_filter(it2,1)
			dvx_r = dvx_r + pcur_vx*interp_filter(it2,2)
			dvz_l = dvz_l + pcur_vz*interp_filter(it2,1)
			dvz_r = dvz_r + pcur_vz*interp_filter(it2,2)
			dxx_l = dxx_l + pcur_xx*interp_filter(it2,1)
			dxx_r = dxx_r + pcur_xx*interp_filter(it2,2)
			dzz_l = dzz_l + pcur_zz*interp_filter(it2,1)
			dzz_r = dzz_r + pcur_zz*interp_filter(it2,2)
			dxz_l = dxz_l + pcur_xz*interp_filter(it2,1)
			dxz_r = dxz_r + pcur_xz*interp_filter(it2,2)
			!On derivatives
			if (wantwavefield) then
				pscat_vxdt_l = pscat_vxdt_l + pvx_dt*interp_filter(it2,1)*dt_inv
				pscat_vxdt_r = pscat_vxdt_r + pvx_dt*interp_filter(it2,2)*dt_inv
				pscat_vzdt_l = pscat_vzdt_l + pvz_dt*interp_filter(it2,1)*dt_inv
				pscat_vzdt_r = pscat_vzdt_r + pvz_dt*interp_filter(it2,2)*dt_inv
			endif


			!Computing the new time step
			pnew_xx = pold_xx + pxx_dt
			pnew_zz = pold_zz + pzz_dt
			pnew_xz = pold_xz + pxz_dt
			pnew_vz = pold_vz + pvz_dt
			pnew_vx = pold_vx + pvx_dt



			!Swapping pointers
			pold_xx => pcur_xx
			pcur_xx => pnew_xx
			pnew_xx => pold_xx
			pold_zz => pcur_zz
			pcur_zz => pnew_zz
			pnew_zz => pold_zz
			pold_xz => pcur_xz
			pcur_xz => pnew_xz
			pnew_xz => pold_xz
			pold_vx => pcur_vx
			pcur_vx => pnew_vx
			pnew_vx => pold_vx
			pold_vz => pcur_vz
			pcur_vz => pnew_vz
			pnew_vz => pold_vz

		enddo


		!Scaling secondary source or back-propagated residual
		dvx_r = dvx_r*rhox*scale_scat; dvz_r = dvz_r*rhoz*scale_scat
		dxx_r = dxx_r*scale_scat; dzz_r = dzz_r*scale_scat; dxz_r = dxz_r*scale_scat
		!Imaging condition
		call full_elastic_imaging_op(.true.,.true.,pvxdt(:,:,:,itw),pvzdt(:,:,:,itw),pvx(:,:,:,itw),pvz(:,:,:,itw),pertrhox,pertrhoz,pertlame_pad,pertmu_pad,pertmuxz,dvx_r,dvz_r,dxx_r,dzz_r,dxz_r)


		!saving receiver wavefield
		if (wantwavefield) then
			pscat_vxdt_r = pscat_vxdt_r*rhox*scale_scat; pscat_vzdt_r = pscat_vzdt_r*rhoz*scale_scat
			if(.not. pressure)then
				sk = sseek_block("wavefieldr",-7,4*nz*nx*ny,1)
			endif
			wc=srite("wavefieldr",dvx_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dvz_r,nz*nx*ny*4)
			wc=srite("wavefieldr",pscat_vxdt_r,nz*nx*ny*4)
			wc=srite("wavefieldr",pscat_vzdt_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dxx_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dzz_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dxz_r,nz*nx*ny*4)
			if(pressure)then
				sk = sseek_block("wavefieldr",-16,4*nz*nx*ny,1)
        wc = srite("wavefieldr",0.5*(pcur_xx+pcur_zz),nz*nx*ny*4)
			else
				sk = sseek_block("wavefieldr",-7,4*nz*nx*ny,1)
			endif
		endif


		!Swap source wavefield left and right pointers
		ptemp => dvx_l
		dvx_l => dvx_r
		dvx_r => ptemp
		dvx_l = 0.
		ptemp => dvz_l
		dvz_l => dvz_r
		dvz_r => ptemp
		dvz_l = 0.
		ptemp => dxx_l
		dxx_l => dxx_r
		dxx_r => ptemp
		dxx_l = 0.
		ptemp => dzz_l
		dzz_l => dzz_r
		dzz_r => ptemp
		dzz_l = 0.
		ptemp => dxz_l
		dxz_l => dxz_r
		dxz_r => ptemp
		dxz_l = 0.
		if (wantwavefield) then
			ptemp => pscat_vxdt_l
			pscat_vxdt_l => pscat_vxdt_r
			pscat_vxdt_r => ptemp
			pscat_vxdt_l = 0.

			ptemp => pscat_vzdt_l
			pscat_vzdt_l => pscat_vzdt_r
			pscat_vzdt_r => ptemp
			pscat_vzdt_l = 0.
		endif


		icount = icount + 1
		if (mod(icount,(nw/10))==0) then
    	if(verbose)write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
    	close(0)
		endif


	enddo
		itw=1
		!Last step of the receiver wavefield
		dvx_r = dvx_r + pcur_vx*interp_filter(it2,2)
		dvz_r = dvz_r + pcur_vz*interp_filter(it2,2)
		dxx_r = dxx_r + pcur_xx*interp_filter(it2,2)
		dzz_r = dzz_r + pcur_zz*interp_filter(it2,2)
		dxz_r = dxz_r + pcur_xz*interp_filter(it2,2)


		!Scaling secondary source or back-propagated residual
		dvx_r = dvx_r*rhox*scale_scat; dvz_r = dvz_r*rhoz*scale_scat
		dxx_r = dxx_r*scale_scat; dzz_r = dzz_r*scale_scat; dxz_r = dxz_r*scale_scat
		!Imaging condition
		call full_elastic_imaging_op(.true.,.true.,pvxdt(:,:,:,itw),pvzdt(:,:,:,itw),pvx(:,:,:,itw),pvz(:,:,:,itw),pertrhox,pertrhoz,pertlame_pad,pertmu_pad,pertmuxz,dvx_r,dvz_r,dxx_r,dzz_r,dxz_r)


		if (wantwavefield) then
			pscat_vzdt_r(:,:,:) = pscat_vzdt_r(:,:,:) + pvz_dt*interp_filter(1,2)*dt_inv
			pscat_vxdt_r(:,:,:) = pscat_vxdt_r(:,:,:) + pvx_dt*interp_filter(1,2)*dt_inv
			pscat_vxdt_r = pscat_vxdt_r*rhox*scale_scat; pscat_vzdt_r = pscat_vzdt_r*rhoz*scale_scat

			if(.not. pressure)then
				sk = sseek_block("wavefieldr",-7,4*nz*nx*ny,1)
			endif
			wc=srite("wavefieldr",dvx_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dvz_r,nz*nx*ny*4)
			wc=srite("wavefieldr",pscat_vxdt_r,nz*nx*ny*4)
			wc=srite("wavefieldr",pscat_vzdt_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dxx_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dzz_r,nz*nx*ny*4)
			wc=srite("wavefieldr",dxz_r,nz*nx*ny*4)
			if(pressure)then
        sk = sseek_block("wavefieldr",-8,4*nz*nx*ny,1)
      else
				sk = sseek_block("wavefieldr",-7,4*nz*nx*ny,1)
			endif
		endif

		if(verbose)write(0,'(A4)',advance="no")"DONE"
        close(0)
	if(verbose)write(0,'(a)',advance="yes")""


	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing model for source n°: ", is
	!Performing last conversions and staggered grid movements
	call Sx_op(.false., .false., pertmuxz, ptemp)
	call Sz_op(.false., .true., ptemp, pertmu_pad)
	call Sz_op(.false., .false., pertrhoz, pertrho_pad)
	call Sx_op(.false., .true., pertrhox, pertrho_pad)
	
	!Converting model parameter perturbations from lame/mu/rho if they were requested in a different space
	if(mod_par .eq. 0) then
  		pertlame_pad = pertlame_pad*1e9 !Conversion from GPa to Pa
  		pertmu_pad = pertmu_pad*1e9 !Conversion from GPa to Pa
  	elseif(mod_par .eq. 1) then 
  		!Converting lame/mu/rho to Vp/Vs/rho perturbations
  		pertrho_pad = pertrho_pad + (mu_pad*pertmu_pad + lame_pad*pertlame_pad)/rho_pad
  		pertmu_pad =  sqrt(mu_pad*rho_pad)*(2.0*pertmu_pad - 4.0*pertlame_pad)
  		pertlame_pad = 2.0*sqrt((lame_pad+2.0*mu_pad)*rho_pad)*pertlame_pad
  	elseif(mod_par .eq. 2) then
  		!Converting lame/mu/rho to Ip/Is/rho perturbations
  		pertrho_pad = pertrho_pad - (lame_pad*pertlame_pad+mu_pad*pertmu_pad)/rho_pad
  		pertmu_pad = (2.0*pertmu_pad-4.0*pertlame_pad)*sqrt(mu_pad/rho_pad)
  		pertlame_pad = 2.0*sqrt((lame_pad+2.0*mu_pad)/rho_pad)*pertlame_pad
  	endif
  	
	!Write model for is source
	wc=srite("out",pertlame_pad(:,:,:),nz*nx*ny*4)
	wc=srite("out",pertmu_pad(:,:,:),nz*nx*ny*4)
	wc=srite("out",pertrho_pad(:,:,:),nz*nx*ny*4)

  enddo

  call Prop_elasti_step_close()
  call full_elastic_imaging_close()
  call Sou_scale_acoustic_close()
  call absorbing_close()

  !Writing information to headers
  call to_history("n1",nz)
  call to_history("o1",oz)
  call to_history("d1",dz)
  call to_history("label1","z [m]")
  call to_history("n2",nx)
  call to_history("o2",ox)
  call to_history("d2",dx)
  call to_history("label2","x [m]")
  call to_history("n3",3)
  call to_history("o3",1.)
  call to_history("d3",1.)
  if(mod_par .eq. 0) then
  	call to_history("label3","Parameter (Lame,shear modulus,density)")
  elseif(mod_par .eq. 1) then 
  	call to_history("label3","Parameter (Vp,Vs,density)")
  elseif(mod_par .eq. 2) then
  	call to_history("label3","Parameter (Ip,Is,density)")
  endif
  call to_history("n4",ns)
  call to_history("o4",1.)
  call to_history("d4",1.)
  call to_history("label4","Source number")
  if (wantwavefield .and. .not. readwavefield) then
  	if(twoD) then
  		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
  		call to_history("n2",nx,"wavefield")
    	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
  		call to_history("n3",4,"wavefield")
  		if(pressure) call to_history("n3",5,"wavefield")
  		call to_history("o3",1.,"wavefield")
  		call to_history("d3",1.,"wavefield")
  		call to_history("label3","Component [Vx,Vz,dVxdt,dVzdt]","wavefield")
  		if(pressure)call to_history("label3","Component [Pressure,Vx,Vz,dVxdt,dVzdt]","wavefield")
  		if (desample .eq. 1) then
  		call to_history("n4",nt,"wavefield")
  		else
  			call to_history("n4",nt/desample+1,"wavefield")
  		endif
  		call to_history("o4",ow,"wavefield")
  	 	call to_history("d4",dw,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")
  	else
  		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
  		call to_history("n2",nx,"wavefield")
    	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
  		call to_history("n3",ny,"wavefield")
    	call to_history("o3",oy,"wavefield")
    	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
  		call to_history("n4",6,"wavefield")
  		if(pressure) call to_history("n3",7,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Component [Vx,Vy,Vz,dVxdt,dVydt,dVzdt]","wavefield")
  		if(pressure)call to_history("label4","Component [Pressure,Vx,Vy,Vz,dVxdt,dVydt,dVzdt]","wavefield")
  		if (desample .eq. 1) then
    		call to_history("n5",nt,"wavefield")
  		else
  			call to_history("n5",nt/desample+1,"wavefield")
  		endif
  		call to_history("o5",ow,"wavefield")
   	 	call to_history("d5",dw,"wavefield")
  		call to_history("label5","Time [s]","wavefield")
  		call to_history("n6",ns,"wavefield")
  		call to_history("o6",1.,"wavefield")
  		call to_history("d6",1.,"wavefield")
  		call to_history("label6","Source number","wavefield")
  	endif
  endif
  if (wantwavefield) then
  	if(twoD) then
  		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
  		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
  		call to_history("n3",7,"wavefieldr")
  		if(pressure) call to_history("n3",8,"wavefieldr")
  		call to_history("o3",1.,"wavefieldr")
  		call to_history("d3",1.,"wavefieldr")
  		call to_history("label3","Component [Vx,Vz,dVxdt,dVzdt,Sigmaxx,Sigmazz,Sigmaxz]","wavefieldr")
  		if(pressure)call to_history("label3","Component [Pressure,Vx,Vz,dVxdt,dVzdt,Sigmaxx,Sigmazz,Sigmaxz]","wavefieldr")
  		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefieldr")
		else
			call to_history("n4",nt/desample+1,"wavefieldr")
		endif
  		call to_history("o4",ow,"wavefieldr")
 	 	call to_history("d4",dw,"wavefieldr")
  		call to_history("label4","Time [s]","wavefieldr")
  		call to_history("n5",ns,"wavefieldr")
  		call to_history("o5",1.,"wavefieldr")
  		call to_history("d5",1.,"wavefieldr")
  		call to_history("label5","Source number","wavefieldr")
  	else
		call to_history("n1",nz,"wavefieldr")
  		call to_history("o1",oz,"wavefieldr")
  		call to_history("d1",dz,"wavefieldr")
  		call to_history("label1","z [m]","wavefieldr")
		call to_history("n2",nx,"wavefieldr")
	  	call to_history("o2",ox,"wavefieldr")
  		call to_history("d2",dx,"wavefieldr")
  		call to_history("label2","x [m]","wavefieldr")
		call to_history("n3",ny,"wavefieldr")
	  	call to_history("o3",oy,"wavefieldr")
	  	call to_history("d3",dy,"wavefieldr")
  		call to_history("label3","y [m]","wavefieldr")
  		call to_history("n4",12,"wavefieldr")
  		if(pressure) call to_history("n3",13,"wavefieldr")
  		call to_history("o4",1.,"wavefieldr")
  		call to_history("d4",1.,"wavefieldr")
  		call to_history("label4","Component [Vx,Vy,Vz,dVxdt.dVydt,dVzdt,Sigmaxx,Sigmayy,Sigmazz,Sigmaxz,Sigmayz,Sigmaxy]","wavefieldr")
  		if(pressure)call to_history("label4","Component [Pressure,Vx,Vy,Vz,dVxdt.dVydt,dVzdt,Sigmaxx,Sigmayy,Sigmazz,Sigmaxz,Sigmayz,Sigmaxy]","wavefieldr")
  		if (desample .eq. 1) then
			call to_history("n5",nt,"wavefieldr")
		else
			call to_history("n5",nt/desample+1,"wavefieldr")
		endif
  		call to_history("o5",ow,"wavefieldr")
 	 	call to_history("d5",dw,"wavefieldr")
  		call to_history("label5","Time [s]","wavefieldr")
  		call to_history("n6",ns,"wavefieldr")
  		call to_history("o6",1.,"wavefieldr")
  		call to_history("d6",1.,"wavefieldr")
  		call to_history("label6","Source number","wavefieldr")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"



  !Deallocating memory
  if (wantwavefield) deallocate(pscat_vzdt_l,pscat_vzdt_r)
  if(associated(pertrhoy)) deallocate(pertrhoy)
  deallocate(pertmuxz,pertrhox,pertrhoz)
  deallocate(pvzdt,pvxdt,pvz,pvx)
  if(allocated(rhoy)) deallocate(rhoy)
  deallocate(rhox,rhoz)
  deallocate(dxz_l,dxz_r,dzz_l,dzz_r,dxx_l,dxx_r,dvz_l,dvz_r,dvx_l,dvx_r)
  deallocate(pvz_dt, pvx_dt, pxz_dt, pxx_dt, pzz_dt)
  deallocate(cur_vz,old_vz,cur_vx,old_vx,cur_xz,old_xz,cur_zz,old_zz,cur_xx,old_xx)
  deallocate(interp_filter)
  deallocate(recfield_vz_fine,recfield_vx_fine,recfield_xz_fine,recfield_zz_fine,recfield_xx_fine,recfield)
  deallocate(wavelet_vz_fine,wavelet_vx_fine,wavelet_xz_fine,wavelet_zz_fine,wavelet_xx_fine)
  deallocate(rho_pad,mu_pad,lame_pad)
  deallocate(pertrho_pad,pertmu_pad,pertlame_pad)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(scaled_vz_wavelet,scaled_vx_wavelet)
  deallocate(wavelet_vz,wavelet_vx,wavelet_xz,wavelet_zz,wavelet_xx)
  call sep_close()
  call exit (0)


end program Bornfast_elastic_adj
