!!$=head1 NAME
!!$
!!$Born modeling Forward - Acoustic Born modeling with constant density. (Without adjoint interpolation filter for source wavefield)
!!$
!!$=head1 SYNOPSIS
!!$
!!$Bornfast_fw.x <velocity.H deltavel=deltavel.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation Born modeling for acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the velocity model, the perturbation in the model and return the perturbed data in data.H
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<velocity.H> - Velocity model in SEP3D format (theoretically [m/s]). (Constructed using Velpar and/or Vel)
!!$
!!$=item from par
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<deltavel> - Velocity perturbation in SEP3D format.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save perturbed wavefield if wantwavefield=1
!!$
!!$B<filename> I<wavefields> - [wavefields.H] File where to save source wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional)
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Born_fw
  
  !Used modules 
  use sep
  use read_pos
  use sec_deriv
  use Pad_vel
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_step
  use weight_sec
  use Pad_pert


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: vel   !velocity model 
  real, dimension(:,:,:), allocatable					:: pertvel   !velocity perturbation 
  real, dimension(:), allocatable						:: wavelet, wavelet_buff,scaled_wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), allocatable, target			:: sec_source_left, sec_source_right
  real, dimension(:,:,:), pointer						:: psec_source_left, psec_source_right
  real, dimension(:,:,:), allocatable					:: sec_scale
  real, dimension(:,:,:), pointer						::  ptemp
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), allocatable, target			:: curpert, oldpert
  real, dimension(:,:,:), pointer						:: pnewpert, pcurpert, poldpert
  real, dimension(:,:,:), pointer						:: vel_pad, pertvel_pad !Padded velocity model
  real, dimension(:,:,:), pointer						:: scat_sou_right !for imaging condition
  real													:: dt !Temporal sampling of the propagator
  real, dimension(:), allocatable						:: wavelet_fine !Source wavelet fine grid 
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the velocity 
  integer												:: is, ig, it, it2, itw !Cycles variables
  integer												:: wc, srite, rc, sreed
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr
  real, dimension(:,:), allocatable						:: interp_filter
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
		  

  write(0,*)""
  write(0,*)""
  write(0,*)"#######################################################################"
  write(0,*)"####################Acoustic 2D/3D wave propagator#####################"
  write(0,*)"####################         Born modeling        #####################"  
  write(0,*)"#######################################################################"
  write(0,*)""
  write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("desample",desample,1)
  call from_par ("dt",dt,0.)
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the velocity and the perturbation
  allocate(vel(nz,nx,ny),pertvel(nz,nx,ny))
  vel(:,:,:) = 0.;pertvel(:,:,:) = 0.
  rc=sreed("in" , vel , nz*nx*ny*4)
  rc=sreed("deltavel" , pertvel , nz*nx*ny*4)
  !compute model perturbation from velocity perturbation
  pertvel(:,:,:) = -2.0*pertvel(:,:,:)/(vel(:,:,:)*vel(:,:,:)*vel(:,:,:))

  !Adding absorbing boundaries to the velocity model
  call Pad_velocity(vel,vel_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  call Pad_perturbation(pertvel,pertvel_pad,nv_pad)
  
  write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxval(vel_pad)*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  write(0,*) "	Number of samples for 50 Hz wave = ", (minval(vel_pad)/(50.*max(dx,dy,dz)))
  write(0,*) "	CFL number: ",dt*maxval(vel_pad)/min(dx,dy,dz)
  write(0,*)"	Min vel [m/s]:",minval(vel_pad)
  write(0,*)"	Max vel [m/s]:",maxval(vel_pad)
  write(0,*) ""
  
  !Scaling source wavelet by 1/(dz*dx) for 2D or 1/(dz*dx*dy) (discretization of delta function!)
  if(twoD)then
  	wavelet=wavelet/(dz*dx)
  else
  	wavelet=wavelet/(dz*dx*dy)
  endif

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	 wavelet_fine = wavelet
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt),scaled_wavelet(nt))
  	wavelet_fine = 0.
	!call resample_init(nw, nt)
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine)
  endif
  !Computing the second order derivative on the source wavelet
  allocate(wavelet_buff(nt))
  call sec_deriv_init(nt, dt)
  call sec_deriv_op(.false., .false., wavelet_fine, wavelet_buff)
  wavelet_fine = wavelet_buff
  
  
  !Initialzing the fixed-variable routines
  call Prop_step_init(nz, nx, ny, nt, nv_pad, vel_pad, dt, dz, dx, dy,twoD)
  call weight_sec_init(nz,nx,ny,dt,dz,dx,dy,nv_pad,vel_pad,twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  pold => old
  pcur => cur
  pnew => old
  allocate(oldpert(nz,nx,ny),curpert(nz,nx,ny))
  poldpert => oldpert
  pcurpert => curpert
  pnewpert => oldpert
  allocate(scat_sou_right(nz,nx,ny),sec_scale(nz,nx,ny))
  sec_scale(:,:,:) = -dt*dt*vel_pad*vel_pad
  allocate(sec_source_left(nz,nx,ny),sec_source_right(nz,nx,ny))
  psec_source_left=>sec_source_left
  psec_source_right=>sec_source_right
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample
	interp_filter(it2,1)=1.-real((it2-1))/real(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  write(0,*) "Dimensions:"
  write(0,'(A7,I6)') " nz = ",nz
  write(0,'(A7,I6)') " nx = ",nx
  write(0,'(A7,I6)') " ny = ",ny
  write(0,'(A7,I6)') " nt = ",nt
  write(0,'(A7,I6)') " ns = ",ns
  write(0,'(A7,I6)') " ng = ",ng
  write(0,*) ""
  
  if (wantwavefield) write(0,*) "#Program will be writing wavefield as requested#"
  
  
  do is = 1,ns
  	old = 0.; cur = 0.
  	oldpert = 0.; curpert = 0.
  	sec_source_left=0.; sec_source_right=0.
  	scat_sou_right = 0.
	write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the source
	scaled_wavelet = wavelet_fine
	call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_op(scaled_wavelet,vel_pad) !Adjoint = Forward (input-output same variable)
	call system_clock(count_rate=cr)
	call system_clock(count1)
	write(0,*)"Propagating..."
	!Propagate source field first step
		it=0
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
		endif	

		!saving source wavefield first step
		if (wantwavefield) then
			wc=srite("wavefields",pcur,nz*nx*ny*4)
		endif

		!Apply the adjoint of imaging condition (i.e., Born operator)
		scat_sou_right(:,:,:) = pcur(:,:,:) * pertvel_pad(:,:,:)*sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., scat_sou_right(:,:,:), psec_source_left(:,:,:))

		do it2 = 1, desample
			call Prop_step_op(.false., pold, pcur, pnew)
			!Adding secondary source
			if (twoD) then
				call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			else
				call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			endif			
			!Swapping pointers
			pold => pcur 
			pcur => pnew
			pnew => pold
		enddo

		
		!saving first step scattered wavefield = 0.
		if (wantwavefield) then
			wc=srite("wavefield",pcurpert,nz*nx*ny*4)
		endif

		
	do itw = 2, nw-1 !Forward
		it = (itw-1)*desample
		!Propagate source field
		if (twoD) then
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
		else
			call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
		endif	

		!saving source wavefield
		if (wantwavefield) then
			wc=srite("wavefields",pcur,nz*nx*ny*4)
		endif

		!Apply the adjoint of imaging condition (i.e., Born operator)
		scat_sou_right(:,:,:) = pcur(:,:,:) * pertvel_pad(:,:,:)*sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., scat_sou_right(:,:,:), psec_source_right(:,:,:))

		do it2 = 1, desample
			call Prop_step_op(.false., pold, pcur, pnew)
			!Adding secondary source
			if (twoD) then
				call Add_source2D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			else
				call Add_source3D_op(.true.,.true.,pnew,scaled_wavelet(it+it2))
			endif	
			
			!Swapping pointers
			pold => pcur 
			pcur => pnew
			pnew => pold
		enddo
		
		
		
		!Propagate scattered wavefield
		it = (itw-2)*desample
		do it2 = 1, desample
			call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)
			if (twoD) then
				!Taking perturbed wavefield
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			else
				!Taking perturbed wavefield
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			endif
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pcurpert,nz*nx*ny*4)
		endif
		
		!Swap secondary source left and right pointers
		ptemp => psec_source_left
		psec_source_left => psec_source_right
		psec_source_right => ptemp
		
		if (mod(itw,(nw/10))==0) then
        	write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif

      	
	enddo
		!saving source wavefield last step
		if (wantwavefield) then
			wc=srite("wavefields",pcur,nz*nx*ny*4)
		endif

		!last step scattered wavefield
 		 it = (nw-2)*desample
  		!Apply the adjoint of imaging condition (i.e., Born operator)
		scat_sou_right(:,:,:) = pcur(:,:,:) * pertvel_pad(:,:,:)*sec_scale(:,:,:)
		!Scale secondary source
		call weight_sec_op(.false., .false., scat_sou_right(:,:,:), psec_source_right(:,:,:))
		do it2 = 1, desample
			call Prop_step_op(.false., poldpert, pcurpert, pnewpert)
			!Adding source (perturbed and source wavefield)
			pnewpert(:,:,:) = pnewpert(:,:,:) + psec_source_left(:,:,:)*interp_filter(it2,1) + psec_source_right(:,:,:)*interp_filter(it2,2)
			if (twoD) then
				!Taking perturbed wavefield
				do ig = 1, ng
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			else
				!Taking perturbed wavefield
				do ig = 1,ng
					call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
					call Add_source3D_op(.false.,.false.,pcurpert,recfield_fine(it+it2,ig))
				enddo
			endif
			!Swapping pointers
			poldpert => pcurpert 
			pcurpert => pnewpert
			pnewpert => poldpert
		enddo
		!saving scattered wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pcurpert,nz*nx*ny*4)
		endif
	write(0,'(a)',advance="yes")""
	call system_clock(count2)
	write(0,*) "Propagation time: ",(count2-count1)/cr
	write(0,'(A44,I5)') " Writing receiver wavefield for source n°: ", is
	!Resample receivers' wavefield
	do ig = 1,ng
		if (dw .le. dt) then
			wc=srite("out",recfield_fine(:,ig),nw*4)
		else
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		endif
	enddo
   	write(0,*) ""

  enddo
  

  !Closing used routines
  call Prop_step_close()
  call weight_sec_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		if (desample .eq. 1) then
			call to_history("n3",nt,"wavefield")
		else
			call to_history("n3",nt/desample+1,"wavefield")
		endif
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  		call to_history("n4",ns,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Source number","wavefield")  	

		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		if (desample .eq. 1) then
			call to_history("n3",nt,"wavefields")
		else
			call to_history("n3",nt/desample+1,"wavefields")
		endif
  		call to_history("o3",ow,"wavefields")
 	 	call to_history("d3",dt*desample,"wavefields")
  		call to_history("label3","Time [s]","wavefields")
  		call to_history("n4",ns,"wavefields")
  		call to_history("o4",1.,"wavefields")
  		call to_history("d4",1.,"wavefields")
  		call to_history("label4","Source number","wavefields")  

  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefield")
		else
			call to_history("n4",nt/desample+1,"wavefield")
		endif
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")

		call to_history("n1",nz,"wavefields")
  		call to_history("o1",oz,"wavefields")
  		call to_history("d1",dz,"wavefields")
  		call to_history("label1","z [m]","wavefields")
		call to_history("n2",nx,"wavefields")
	  	call to_history("o2",ox,"wavefields")
  		call to_history("d2",dx,"wavefields")
  		call to_history("label2","x [m]","wavefields")
		call to_history("n3",ny,"wavefields")
	  	call to_history("o3",oy,"wavefields")
	  	call to_history("d3",dy,"wavefields")
  		call to_history("label3","y [m]","wavefields")
		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefields")
		else
			call to_history("n4",nt/desample+1,"wavefields")
		endif
  		call to_history("o4",ow,"wavefields")
 	 	call to_history("d4",dt*desample,"wavefields")
  		call to_history("label4","Time [s]","wavefields")
  		call to_history("n5",ns,"wavefields")
  		call to_history("o5",1.,"wavefields")
  		call to_history("d5",1.,"wavefields")
  		call to_history("label5","Source number","wavefields")
  	endif
  endif
  write(0,*) "End propagation"

  !Deallocating memory
  deallocate(interp_filter)
  deallocate(sec_source_right,sec_source_left)
  deallocate(sec_scale,scat_sou_right)
  deallocate(curpert,oldpert)
  deallocate(cur,old)
  deallocate(scaled_wavelet,wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(pertvel_pad,vel_pad)
  deallocate(wavelet_buff,wavelet)
  call sep_close()
  call exit (0)
end program Born_fw


