!Module to compute central derivative by averaging staggered operator 1/2(Dp+Dm) (10th order approximation)
!Model dimension (z,x,y) Input: wavefield 
!Data dimension (z,x,y) Output: derivative
!D(z,x,y) = D10pt_0_(z,x,y)(Model(z,x,y))
!Model(z,x,y) = D10pt_0_(z,x,y)*(D(z,x,y))
module D10pt_0_mod

  implicit none
  !Module variables
  integer, private									:: nz, nx, ny
  real, private										:: dz, dx, dy
  logical, private									:: twoD
  real, private										:: C1z, C2z, C3z, C4z, C5z, C1x, C2x, C3x, C4x, C5x, C1y, C2y, C3y, C4y, C5y
  real, dimension(:,:,:),allocatable,private		:: tmp

  contains
  
  		subroutine D10pt_0_init(nz_in,nx_in,ny_in,dz_in,dx_in,dy_in,twoD_in)
			!Input variables	
			integer, intent(in)						:: nz_in, nx_in, ny_in
			real, intent(in)						:: dz_in, dx_in, dy_in
			logical, intent(in)						:: twoD_in
			!Internal variables
			integer 								:: buffer
			
			parameter(buffer=5)
			
			!Assigning module's variables
			nz = nz_in
			nx = nx_in
			ny = ny_in
			
			dz = dz_in
			dx = dx_in
			dy = dy_in
			
			twoD = twoD_in
			!D0 coefficients application --> C1(tmp(i+1)-tmp(i-1))+C2(tmp(i+2)-tmp(i-2))+C3(tmp(i+3)-tmp(i-3))......
			!Defining D0 coefficients
			C1z = 0.560760494944436/(dz)
			C2z = -0.037939449331637/(dz)
			C3z = 0.006038554375894/(dz)
			C4z = -8.234893506317208e-04/(dz)
			C5z = 5.933959872487745e-05/(dz)
			
			C1x = 0.560760494944436/(dx)
			C2x = -0.037939449331637/(dx)
			C3x = 0.006038554375894/(dx)
			C4x = -8.234893506317208e-04/(dx)
			C5x = 5.933959872487745e-05/(dx)
			
			C1y = 0.560760494944436/(dy)
			C2y = -0.037939449331637/(dy)
			C3y = 0.006038554375894/(dy)
			C4y = -8.234893506317208e-04/(dy)
			C5y = 5.933959872487745e-05/(dy)
		
			!Allocating memory for temporary array
			if (.not. allocated(tmp)) then
				if (twoD) then
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1))
				else
					allocate(tmp(1-buffer:nz+buffer,1-buffer:nx+buffer,1-buffer:ny+buffer))
				endif
			endif
			tmp = 0.
		
			return
		end subroutine D10pt_0_init
		
		
		subroutine D0z_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call D0z_op(adj,add,mm,dd)
  		end subroutine D0z_call
  		
  		
  		
  		subroutine D0z_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (D0z*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) + (C1z*(tmp(iz+1,ix,1)-tmp(iz-1,ix,1))  &
										      		  + C2z*(tmp(iz+2,ix,1)-tmp(iz-2,ix,1)) &
										      		  + C3z*(tmp(iz+3,ix,1)-tmp(iz-3,ix,1)) &
										      		  + C4z*(tmp(iz+4,ix,1)-tmp(iz-4,ix,1)) &
										      		  + C5z*(tmp(iz+5,ix,1)-tmp(iz-5,ix,1)))
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) = mm(iz,ix,iy) + (C1z*(tmp(iz+1,ix,iy)-tmp(iz-1,ix,iy))  &
										      		 			  	+ C2z*(tmp(iz+2,ix,iy)-tmp(iz-2,ix,iy)) &
										      		  				+ C3z*(tmp(iz+3,ix,iy)-tmp(iz-3,ix,iy)) &
										      		  				+ C4z*(tmp(iz+4,ix,iy)-tmp(iz-4,ix,iy)) &
										      		  				+ C5z*(tmp(iz+5,ix,iy)-tmp(iz-5,ix,iy)))
       	     							
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying D0z
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C1z*(tmp(iz+1,ix,1)-tmp(iz-1,ix,1))   &
										    		  + C2z*(tmp(iz+2,ix,1)-tmp(iz-2,ix,1)) &
										      		  + C3z*(tmp(iz+3,ix,1)-tmp(iz-3,ix,1)) &
										      		  + C4z*(tmp(iz+4,ix,1)-tmp(iz-4,ix,1)) &
										      		  + C5z*(tmp(iz+5,ix,1)-tmp(iz-5,ix,1)) 
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = dd(iz,ix,iy) + C1z*(tmp(iz+1,ix,iy)-tmp(iz-1,ix,iy))   &
										    		 			  	+ C2z*(tmp(iz+2,ix,iy)-tmp(iz-2,ix,iy)) &
										      		      		  	+ C3z*(tmp(iz+3,ix,iy)-tmp(iz-3,ix,iy)) &
										      		  			  	+ C4z*(tmp(iz+4,ix,iy)-tmp(iz-4,ix,iy)) &
										      		  			  	+ C5z*(tmp(iz+5,ix,iy)-tmp(iz-5,ix,iy)) 
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine D0z_op
   		
   		
   		subroutine D0x_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call D0x_op(adj,add,mm,dd)
  		end subroutine D0x_call
  		
  		
  		
  		subroutine D0x_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
        		if(twoD) then
					!Applying (D0z*)
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			mm(iz,ix,1) = mm(iz,ix,1) + (C1x*(tmp(iz,ix+1,1)-tmp(iz,ix-1,1))&
										      		  + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-2,1)) &
										      		  + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-3,1)) &
										      		  + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-4,1)) &
										      		  + C5x*(tmp(iz,ix+5,1)-tmp(iz,ix-5,1)))
		           		enddo
		     		enddo
		        	!$OMP END PARALLEL DO
				
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       					    		do iz=1,nz
       					    			mm(iz,ix,iy) = mm(iz,ix,iy) + (C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix-1,iy))&
										      		 			  	+ C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-2,iy)) &
										      		  				+ C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-3,iy)) &
										      		  				+ C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-4,iy)) &
										      		  				+ C5x*(tmp(iz,ix+5,iy)-tmp(iz,ix-5,iy)))
       	     							
            						enddo
           					enddo
				 	enddo
        	 		!$OMP END PARALLEL DO
				
				endif


			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				if(twoD) then
					!Applying D0z
					!$OMP PARALLEL DO private(iz,ix)
					do ix=1,nx
            			do iz=1,nz
                			dd(iz,ix,1) = dd(iz,ix,1) + C1x*(tmp(iz,ix+1,1)-tmp(iz,ix-1,1)) &
										    		  + C2x*(tmp(iz,ix+2,1)-tmp(iz,ix-2,1)) &
										      		  + C3x*(tmp(iz,ix+3,1)-tmp(iz,ix-3,1)) &
										      		  + C4x*(tmp(iz,ix+4,1)-tmp(iz,ix-4,1)) &
										      		  + C5x*(tmp(iz,ix+5,1)-tmp(iz,ix-5,1)) 
		        	    enddo
		     	   enddo
		      	  !$OMP END PARALLEL DO
				else
					!$OMP PARALLEL DO private(iz,ix,iy)
					do iy = 1,ny
        					do ix=1,nx
       				    			do iz=1,nz
       	     							dd(iz,ix,iy) = dd(iz,ix,iy) + C1x*(tmp(iz,ix+1,iy)-tmp(iz,ix-1,iy)) &
										    		 			  	+ C2x*(tmp(iz,ix+2,iy)-tmp(iz,ix-2,iy)) &
										      		      		  	+ C3x*(tmp(iz,ix+3,iy)-tmp(iz,ix-3,iy)) &
										      		  			  	+ C4x*(tmp(iz,ix+4,iy)-tmp(iz,ix-4,iy)) &
										      		  			  	+ C5x*(tmp(iz,ix+5,iy)-tmp(iz,ix-5,iy)) 
            						enddo
           					enddo
			 		enddo
        	 		!$OMP END PARALLEL DO
				
				endif
		
			endif
   			
   			
   		end subroutine D0x_op
   		
   		
   		subroutine D0y_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:) 	 :: mm,dd
    		call D0y_op(adj,add,mm,dd)
  		end subroutine D0y_call
  		
  		
  		!###############################################
		!D0y partial derivative along y axis (3D ONLY)
		!###############################################
		
  		subroutine D0y_op(adj,add,mm,dd)
    		!Input-Output variables
			logical, intent(in)					:: adj, add
			real, dimension(nz,nx,ny)			:: mm !Input 
			real, dimension(nz,nx,ny)			:: dd !Output 
			!Internal variables
			integer								:: iz, ix, iy
			
			
			if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model
        		tmp(1:nz,1:nx,1:ny) = dd
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       				    		do iz=1,nz
       	     						mm(iz,ix,iy) = mm(iz,ix,iy)	+(C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy-1)) &
											   					+ C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-2)) &
																+ C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-3)) &
																+ C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-4)) &
																+ C5y*(tmp(iz,ix,iy+5)-tmp(iz,ix,iy-5)))
            					enddo
           				enddo
			 	enddo
         		!$OMP END PARALLEL DO

			else	     !Forward
				if(.not. add) dd = 0. !Zero-out the data
				tmp(1:nz,1:nx,1:ny) = mm
				!$OMP PARALLEL DO private(iz,ix,iy)
				do iy = 1,ny
        				do ix=1,nx
       			    			do iz=1,nz
       	     						dd(iz,ix,iy) = dd(iz,ix,iy) + C1y*(tmp(iz,ix,iy+1)-tmp(iz,ix,iy-1)) &
									 			  	    		+ C2y*(tmp(iz,ix,iy+2)-tmp(iz,ix,iy-2)) &
									   			   	    		+ C3y*(tmp(iz,ix,iy+3)-tmp(iz,ix,iy-3)) &
									   			   	    		+ C4y*(tmp(iz,ix,iy+4)-tmp(iz,ix,iy-4)) &
									   			   	    		+ C5y*(tmp(iz,ix,iy+5)-tmp(iz,ix,iy-5))
           						enddo
        				enddo
		 		enddo
        	 	!$OMP END PARALLEL DO
				
			endif

   			
   			
   		end subroutine D0y_op


		subroutine D10pt_0_close()
			if(allocated(tmp))deallocate(tmp)
			return
		end subroutine D10pt_0_close


end  module D10pt_0_mod