!!$=head1 NAME
!!$
!!$Interp2D - Program to linearly interpolate 2D grids (forward and adjoint)
!!$
!!$=head1 SYNOPSIS
!!$
!!$Interp2D.x <model_coarse.H npoints= [adj=0] >model_fine.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Program to interpolate bilinearly a 2D model from coarse to fine and adjoint if asked
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history
!!$
!!$=item from par
!!$
!!$B<integer> I<npoints> - [no default] Number of points to add between points in the fine grid per dimension
!!$
!!$B<integer> I<adj> - [0] apply the forward (coarse->fine); 1 apply the adjoint (fine->coarse)
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES
!!$
!!$SEP
!!$
!!$=head1 SEE ALSO
!!$
!!$
!!$=cut
!%
program Interp2D

  !Used modules
  use sep

  implicit none

  integer							  :: nz_coarse,nx_coarse,nz_fine,nx_fine,np,adj,npoints
  real								  :: oz,ox,op,dz_coarse,dx_coarse,dp,dz_fine,dx_fine
  real,dimension(:,:,:),allocatable   :: model_coarse,model_fine
  integer                             :: iz,ix,ip,idz,idx
  real                                :: fz,fx,wz,wx
  call sep_init (SOURCE)

  call from_par("npoints",npoints)
  call from_par("adj",adj,0)
  if(adj)then
    !Reading from fine model space
    call from_history("n1",nz_fine)
    call from_history("d1",dz_fine)
    call from_history("o1",oz)
    call from_history("n2",nx_fine)
    call from_history("d2",dx_fine)
    call from_history("o2",ox)
    call from_history("n3",np)
    call from_history("d3",dp)
    call from_history("o3",op)

    !Computing grid parameters
    dz_coarse=dz_fine*(npoints+1)
    dx_coarse=dx_fine*(npoints+1)
    nz_coarse=(nz_fine+npoints)/(npoints+1)
    nx_coarse=(nx_fine+npoints)/(npoints+1)

    !Writing from coarse model space
    call to_history("n1",nz_coarse)
    call to_history("d1",dz_coarse)
    call to_history("o1",oz)
    call to_history("n2",nx_coarse)
    call to_history("d2",dx_coarse)
    call to_history("o2",ox)
    call to_history("n3",np)
    call to_history("d3",dp)
    call to_history("o3",op)
  else
    !Reading from fine model space
    call from_history("n1",nz_coarse)
    call from_history("d1",dz_coarse)
    call from_history("o1",oz)
    call from_history("n2",nx_coarse)
    call from_history("d2",dx_coarse)
    call from_history("o2",ox)
    call from_history("n3",np)
    call from_history("d3",dp)
    call from_history("o3",op)

    !Computing grid parameters
    dz_fine=dz_coarse/(npoints+1)
    dx_fine=dx_coarse/(npoints+1)
    nz_fine=(nz_coarse-1)*npoints+nz_coarse
    nx_fine=(nx_coarse-1)*npoints+nx_coarse

    !Writing from coarse model space
    call to_history("n1",nz_fine)
    call to_history("d1",dz_fine)
    call to_history("o1",oz)
    call to_history("n2",nx_fine)
    call to_history("d2",dx_fine)
    call to_history("o2",ox)
    call to_history("n3",np)
    call to_history("d3",dp)
    call to_history("o3",op)
  endif

  !Allocate memory
  allocate(model_coarse(nz_coarse+1,nx_coarse+1,np),model_fine(nz_fine,nx_fine,np))
  model_fine=0.
  model_coarse=0.


  if(adj)then
    call sep_read(model_fine)

    do ip = 1,np
      do ix = 1,nx_fine
        fx = 1. + (ix-1)*dx_fine/dx_coarse; idx = fx; wx = fx - idx
        do iz = 1,nz_fine
          fz = 1. + (iz-1)*dz_fine/dz_coarse; idz = fz; wz = fz - idz
          model_coarse(idz,idx,ip)       = model_coarse(idz,idx,ip)       + model_fine(iz,ix,ip)*(1.-wz)*(1.-wx)
          model_coarse(idz,idx+1,ip)     = model_coarse(idz,idx+1,ip)     + model_fine(iz,ix,ip)*(1.-wz)*(wx)
          model_coarse(idz+1,idx,ip)     = model_coarse(idz+1,idx,ip)     + model_fine(iz,ix,ip)*(wz)*(1.-wx)
          model_coarse(idz+1,idx+1,ip)   = model_coarse(idz+1,idx+1,ip)   + model_fine(iz,ix,ip)*(wz)*(wx)
        enddo
      enddo
    enddo

    call sep_write(model_coarse(1:nz_coarse,1:nx_coarse,:))
  else !Forward (coarse to fine)
    call sep_read(model_coarse(1:nz_coarse,1:nx_coarse,:))

    do ip = 1,np
      !$OMP PARALLEL DO private(iz,ix,fx,idx,wx,fz,idz,wz)
      do ix = 1,nx_fine
        fx = 1. + (ix-1)*dx_fine/dx_coarse; idx = fx; wx = fx - idx
        do iz = 1,nz_fine
          fz = 1. + (iz-1)*dz_fine/dz_coarse; idz = fz; wz = fz - idz
          model_fine(iz,ix,ip)=model_fine(iz,ix,ip) + model_coarse(idz,idx,ip)*(1.-wz)*(1.-wx) + &
  					 	                                        model_coarse(idz,idx+1,ip)*(1.-wz)*(wx)  + &
  						  	                                    model_coarse(idz+1,idx,ip)*(wz)*(1.-wx)  + &
  					  	  		                                model_coarse(idz+1,idx+1,ip)*(wz)*(wx)
        enddo
      enddo
      !$OMP END PARALLEL DO
    enddo

    call sep_write(model_fine)
  endif


  deallocate(model_fine,model_coarse)
  call sep_close()
  call exit (0)

end program Interp2D
