!Module for solving one step of the elastic wave equation for displacement formulation
module Prop_elastic_disp_step

  use D10pt_0_mod
  use D10pt_mod
  use S_mod
  use weight
  use G
  
  implicit none
  !Module variables
  integer, private								:: nz, nx, ny, nt
  real, private									:: dt
  real,allocatable,dimension(:,:,:), private	:: scale_dtrho, ptemp, ptemp1, ptemp2, lame_mu_scalez, mu_scalez, lame_mu_scalex, mu_scalex
  real, dimension(:,:,:), pointer, private		:: lame_scale, mu_scale, scale_vel
  logical, private								:: twoD
  
  
  
  contains
  
  	subroutine Prop_elastic_disp_step_init(nz_in, nx_in, ny_in, nt_in, dt_in, dz_in, dx_in, dy_in, twoD_in, rho_in, lame_in, mu_in,nv_pad_in)
  		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in, nt_in, nv_pad_in 
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		real, dimension(:,:,:), pointer			:: lame_in, mu_in, rho_in
		nz = nz_in
		nx = nx_in
		ny = ny_in
		nt = nt_in
		dt = dt_in
		twoD = twoD_in
		
		!Initializing other modules
		call D10pt_0_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD_in)
		call D10pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD_in)
		call S_init(nz,nx,ny,twoD_in)
		
		if(.not. associated(scale_vel)) allocate(scale_vel(nz,nx,ny))
		if(.not. allocated(scale_dtrho)) allocate(scale_dtrho(nz,nx,ny))
		if(.not. allocated(lame_mu_scalez)) allocate(lame_mu_scalez(nz,nx,ny))
		if(.not. allocated(mu_scalez)) allocate(mu_scalez(nz,nx,ny))
		if(.not. allocated(lame_mu_scalex)) allocate(lame_mu_scalex(nz,nx,ny))
		if(.not. allocated(mu_scalex)) allocate(mu_scalex(nz,nx,ny))
		if(.not. allocated(ptemp)) allocate(ptemp(nz,nx,ny))
		if(.not. allocated(ptemp1)) allocate(ptemp1(nz,nx,ny))
		if(.not. allocated(ptemp2)) allocate(ptemp2(nz,nx,ny))
		if(associated(lame_in))then
			lame_scale => lame_in
		else
			write(0,*)"Lame parameter pointer not associated in Prop_elastic_disp_init"
			stop
		endif
		if(associated(mu_in))then
			mu_scale => mu_in
		else
			write(0,*)"Shear modulus pointer not associated in Prop_elastic_disp_init"
			stop
		endif
		
		!Absorbing boundary
		scale_vel=sqrt((lame_in+2.0*mu_in)/rho_in) 
  		call weight_init(nz, nx, ny, dt, dz_in, dx_in, dy_in, nv_pad_in, scale_vel,twoD)
  		call G_init(nz, nx, ny, dt, dz_in, dx_in, dy_in, nv_pad_in,twoD)
		scale_vel=scale_vel*dt
		
		!Computing scaling factors
		scale_dtrho = 1.0/rho_in*(dt*dt)
		call Sz_op(.false., .false., (lame_in+2.0*mu_in), lame_mu_scalez)
		call Sz_op(.false., .false., mu_in, mu_scalez)
		call Sx_op(.false., .false., (lame_in+2.0*mu_in), lame_mu_scalex)
		call Sx_op(.false., .false., mu_in, mu_scalex)
		
		
	end subroutine Prop_elastic_disp_step_init
	
	
	
	subroutine Prop_elastic_step_disp_op(adj,pux_old,puz_old,pux_cur,puz_cur,pux_new,puz_new)
		!Input-Output variables
		logical, intent(in)								:: adj
		real, dimension(nz,nx,ny), intent(inout)		:: pux_old,puz_old,pux_cur,puz_cur,pux_new,puz_new
		
		
		
		if(adj) then !Adjoint 
		
			if (twoD) then
				!Uz portion 1
				ptemp2 = puz_cur*scale_dtrho
				
				call Dpz_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*lame_mu_scalez
				call Dmz_op(.false.,.false.,ptemp,ptemp1)
				
				call Dpx_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*mu_scalex
				call Dmx_op(.false.,.true.,ptemp,ptemp1) 
				
				puz_new = ptemp1 + 2.0*puz_cur - puz_old
				
				!Ux portion 2
				call D0z_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*lame_scale
				call D0x_op(.false.,.false.,ptemp,ptemp1)
				
				call D0x_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*mu_scale
				call D0z_op(.false.,.true.,ptemp,ptemp1)
				
				pux_new = ptemp1 + 2.0*pux_cur -  pux_old
				
				!Uz portion 2
				ptemp2 = pux_cur*scale_dtrho
				
				call D0x_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*lame_scale
				call D0z_op(.false.,.false.,ptemp,ptemp1)
			
				call D0z_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*mu_scale
				call D0x_op(.false.,.true.,ptemp,ptemp1)
				
				puz_new = puz_new + ptemp1
				
				!Ux portion 1
				call Dpx_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*lame_mu_scalex
				call Dmx_op(.false.,.false.,ptemp,ptemp1) 
				
				call Dpz_op(.false.,.false.,ptemp2,ptemp) 
				ptemp = ptemp*mu_scalez
				call Dmz_op(.false.,.true.,ptemp,ptemp1)
				
				pux_new = pux_new + ptemp1
				
			
			endif
		
		else	     !Forward
		
			if (twoD) then
				!Uz portion 1
				call Dpz_op(.false.,.false.,puz_cur,ptemp) 
				ptemp = ptemp*lame_mu_scalez
				call Dmz_op(.false.,.false.,ptemp,ptemp1) 
			
				call Dpx_op(.false.,.false.,puz_cur,ptemp) 
				ptemp = ptemp*mu_scalex
				call Dmx_op(.false.,.true.,ptemp,ptemp1) 
				
				!Uz portion 2
				call D0x_op(.false.,.false.,pux_cur,ptemp) 
				ptemp = ptemp*lame_scale
				call D0z_op(.false.,.true.,ptemp,ptemp1)
			
				call D0z_op(.false.,.false.,pux_cur,ptemp) 
				ptemp = ptemp*mu_scale
				call D0x_op(.false.,.true.,ptemp,ptemp1)
			
				!puz_new = ptemp1*scale_dtrho + 2.0*puz_cur - puz_old
				ptemp1 = ptemp1*scale_dtrho + 2.0*puz_cur
				
				!Operator: (I-w)C2
				call weight_op(.false.,.false.,ptemp1,ptemp)
				ptemp1 = ptemp1 - ptemp
				!Operator: w(I-v*dt*G)
				call G_op(.false.,.false.,puz_cur,ptemp)
				ptemp = puz_cur-scale_vel*ptemp
				call weight_op(.false.,.true.,ptemp,ptemp1)
				
				!Operator: (w-I) on old
				call weight_op(.false.,.false.,puz_old,ptemp)
				ptemp = ptemp - puz_old
				puz_new = ptemp1 + ptemp
			
				!Ux portion 1
				call Dpx_op(.false.,.false.,pux_cur,ptemp) 
				ptemp = ptemp*lame_mu_scalex
				call Dmx_op(.false.,.false.,ptemp,ptemp1) 
				
				call Dpz_op(.false.,.false.,pux_cur,ptemp) 
				ptemp = ptemp*mu_scalez
				call Dmz_op(.false.,.true.,ptemp,ptemp1)
				
				!Ux portion 2
				call D0z_op(.false.,.false.,puz_cur,ptemp) 
				ptemp = ptemp*lame_scale
				call D0x_op(.false.,.true.,ptemp,ptemp1)
				
				call D0x_op(.false.,.false.,puz_cur,ptemp) 
				ptemp = ptemp*mu_scale
				call D0z_op(.false.,.true.,ptemp,ptemp1)
				
				!pux_new = ptemp1*scale_dtrho + 2.0*pux_cur - pux_old
				ptemp1 = ptemp1*scale_dtrho + 2.0*pux_cur
				
				!Operator: (I-w)C2
				call weight_op(.false.,.false.,ptemp1,ptemp)
				ptemp1 = ptemp1 - ptemp
				!Operator: w(I-v*dt*G)
				call G_op(.false.,.false.,pux_cur,ptemp)
				ptemp = pux_cur-scale_vel*ptemp
				call weight_op(.false.,.true.,ptemp,ptemp1)
				
				!Operator: (w-I) on old
				call weight_op(.false.,.false.,pux_old,ptemp)
				ptemp = ptemp - pux_old
				pux_new = ptemp1 + ptemp
				
			endif
		
		
		endif
	
	
	
	end subroutine Prop_elastic_step_disp_op
	
	
	subroutine Prop_elastic_step_disp_close()
	
		call weight_close()
  		if(allocated(ptemp2))deallocate(ptemp2)
  		if(allocated(ptemp1))deallocate(ptemp1)
  		if(allocated(ptemp))deallocate(ptemp)
  		if(allocated(mu_scalex))deallocate(mu_scalex)
  		if(allocated(lame_mu_scalex))deallocate(lame_mu_scalex)
  		if(allocated(mu_scalez))deallocate(mu_scalez)
  		if(allocated(lame_mu_scalez))deallocate(lame_mu_scalez)
  		if(allocated(scale_dtrho))deallocate(scale_dtrho)
  		if(associated(scale_vel))deallocate(scale_vel)
  		call D10pt_0_close()
  		
  	end subroutine  Prop_elastic_step_disp_close




end module Prop_elastic_disp_step