!Module for solving one step of the elastic wave equation
module Prop_elastic_step


  use D10pt_mod
  use S_mod

  implicit none
  !Module variables
  integer, private														:: nz, nx, ny, nt, buffer=5
  real, private																:: dt
  real,allocatable,dimension(:,:,:), private	:: scalevz, scalevx, scale_lame_mu, scale_lame, scale_mu_xz, ptemp
	real,allocatable,dimension(:,:), private		:: scalefree
	logical, private														:: free=.false.


  contains

  	subroutine Prop_elastic_step_init(nz_in,nx_in,ny_in,nt_in,nv_pad_in,dt_in,dz_in,dx_in, dy_in,twoD_in,rhox_in,rhoz_in,lame_in,mu_in,free_in)
  		!Input variables
		integer, intent(in)										:: nz_in, nx_in, ny_in, nt_in, nv_pad_in
		real																	:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)										:: twoD_in
		logical, intent(in), optional					:: free_in
		real, dimension(:,:,:), pointer				:: lame_in, mu_in
		real, dimension(:,:,:), allocatable		:: rhox_in, rhoz_in
		nz = nz_in
		nx = nx_in
		ny = ny_in
		nt = nt_in
		dt = dt_in


		!Initializing other modules
		call D10pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD_in)
		call S_init(nz,nx,ny,twoD_in)

		if(.not. allocated(scalevz)) allocate(scalevz(nz,nx,ny))
		if(.not. allocated(scalevx)) allocate(scalevx(nz,nx,ny))
		if(.not. allocated(scale_lame_mu)) allocate(scale_lame_mu(nz,nx,ny))
		if(.not. allocated(scale_lame)) allocate(scale_lame(nz,nx,ny))
		if(.not. allocated(scale_mu_xz)) allocate(scale_mu_xz(nz,nx,ny))
		if(.not. allocated(ptemp)) allocate(ptemp(nz,nx,ny))

		!Computing the scaling factors of the derivatives
		scalevx = dt/rhox_in
		scalevz = dt/rhoz_in
		call Sx_op(.true., .false., scale_lame, dt*mu_in)
		call Sz_op(.true., .false., scale_mu_xz, scale_lame)
		scale_lame = dt*lame_in
		scale_lame_mu = dt*(lame_in+2.0*mu_in)
		!Checking for free surface
		!Implementing free surface from Robertsson (1996)
		if(present(free_in)) free = free_in
		if (free) then
			if(.not. allocated(scalefree)) allocate(scalefree(nx,ny))
			scalefree(:,:)=dt*(lame_in(buffer+1,:,:)*lame_in(buffer+1,:,:))/(lame_in(buffer+1,:,:)+2.0*mu_in(buffer+1,:,:))
		endif

  	end subroutine Prop_elastic_step_init


  	subroutine Prop_elastic_step_op(adj, pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz, pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt)
  		!Input-Output variables
		logical, intent(in)													:: adj
		real, dimension(nz,nx,ny), intent(inout)		:: pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz
		real, dimension(nz,nx,ny), intent(inout)		:: pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt

		if(adj) then !Adjoint
			!vx time derivative
			ptemp=pcur_xx*scale_lame_mu
			if (free) then
				!Free surface at constant depth of buffer+1 or z=0
				ptemp(buffer+1,:,:) = ptemp(buffer+1,:,:)-pcur_xx(buffer+1,:,:)*scalefree(:,:)
			endif
			call Dpx_op(.true.,.false.,pvx_dt,ptemp)
			ptemp=pcur_zz*scale_lame
			call Dpx_op(.true.,.true.,pvx_dt,ptemp)
			ptemp=pcur_xz*scale_mu_xz
			call Dmz_op(.true.,.true.,pvx_dt,ptemp)
			!vz time derivative
			ptemp=pcur_xx*scale_lame
			if (free) then
				!Free surface at constant depth of buffer+1 or z=0
				ptemp(buffer+1,:,:) = 0.0
			endif
			call Dpz_op(.true.,.false.,pvz_dt,ptemp)
			ptemp=pcur_zz*scale_lame_mu
			call Dpz_op(.true.,.true.,pvz_dt,ptemp)
			ptemp=pcur_xz*scale_mu_xz
			call Dmx_op(.true.,.true.,pvz_dt,ptemp)
			!sigmaxx derivative
			ptemp=pcur_vx*scalevx
			call Dmx_op(.true.,.false.,pxx_dt,ptemp)
			!sigmazz derivative
			ptemp=pcur_vz*scalevz
			call Dmz_op(.true.,.false.,pzz_dt,ptemp)
			!sigmaxz time derivative
			ptemp=pcur_vx*scalevx
			call Dpz_op(.true.,.false.,pxz_dt,ptemp)
			ptemp=pcur_vz*scalevz
			call Dpx_op(.true.,.true.,pxz_dt,ptemp)
			!Adjoint condition of free surface
			if (free) then
				!Sigma zz (defined on x,z)
				pzz_dt(11,:,:) = pzz_dt(11,:,:) - pzz_dt(1,:,:)
				pzz_dt(10,:,:) = pzz_dt(10,:,:) - pzz_dt(2,:,:)
				pzz_dt(9,:,:)  = pzz_dt(9,:,:)  - pzz_dt(3,:,:)
				pzz_dt(8,:,:)  = pzz_dt(8,:,:)  - pzz_dt(4,:,:)
				pzz_dt(7,:,:)  = pzz_dt(7,:,:)  - pzz_dt(5,:,:)
				pzz_dt(1,:,:) = 0.0
				pzz_dt(2,:,:) = 0.0
				pzz_dt(3,:,:) = 0.0
				pzz_dt(4,:,:) = 0.0
				pzz_dt(5,:,:) = 0.0
				pzz_dt(buffer+1,:,:) = 0.0
				!Sigma xz (defined at x-1/2,z-1/2)
				pxz_dt(11,:,:) = pxz_dt(11,:,:) - pxz_dt(2,:,:)
				pxz_dt(10,:,:) = pxz_dt(10,:,:) - pxz_dt(3,:,:)
				pxz_dt(9,:,:)  = pxz_dt(9,:,:)  - pxz_dt(4,:,:)
				pxz_dt(8,:,:)  = pxz_dt(8,:,:)  - pxz_dt(5,:,:)
				pxz_dt(7,:,:)  = pxz_dt(7,:,:)  - pxz_dt(6,:,:)
				pxz_dt(2,:,:) = 0.0
				pxz_dt(3,:,:) = 0.0
				pxz_dt(4,:,:) = 0.0
				pxz_dt(5,:,:) = 0.0
				pxz_dt(6,:,:) = 0.0
				!Vz (defined at x,z-1/2)
				pvz_dt(11,:,:) = pvz_dt(11,:,:) + pvz_dt(2,:,:)
				pvz_dt(10,:,:) = pvz_dt(10,:,:) + pvz_dt(3,:,:)
				pvz_dt(9,:,:)  = pvz_dt(9,:,:)  + pvz_dt(4,:,:)
				pvz_dt(8,:,:)  = pvz_dt(8,:,:)  + pvz_dt(5,:,:)
				pvz_dt(7,:,:)  = pvz_dt(7,:,:)  + pvz_dt(6,:,:)
				pvz_dt(2,:,:) = 0.0
				pvz_dt(3,:,:) = 0.0
				pvz_dt(4,:,:) = 0.0
				pvz_dt(5,:,:) = 0.0
				pvz_dt(6,:,:) = 0.0
				!Vx (defined at x-1/2,z)
				pvx_dt(11,:,:) = pvx_dt(11,:,:) + pvx_dt(1,:,:)
				pvx_dt(10,:,:) = pvx_dt(10,:,:) + pvx_dt(2,:,:)
				pvx_dt(9,:,:)  = pvx_dt(9,:,:)  + pvx_dt(3,:,:)
				pvx_dt(8,:,:)  = pvx_dt(8,:,:)  + pvx_dt(4,:,:)
				pvx_dt(7,:,:)  = pvx_dt(7,:,:)  + pvx_dt(5,:,:)
				pvx_dt(1,:,:) = 0.0
				pvx_dt(2,:,:) = 0.0
				pvx_dt(3,:,:) = 0.0
				pvx_dt(4,:,:) = 0.0
				pvx_dt(5,:,:) = 0.0
			endif



		else	     !Forward
			!Imaging method 2 by Crase (1990)
			!Particle velocities must be continuous at the interface
			!set sigmazz(x,z=0) = 0.0
			!make sigmazz and sigma xz odd functions w.r.t. free surface
			!make vx and vz even functions w.r.t. free surface
			if (free) then
				!Sigma zz (defined on x,z)
				pcur_zz(1,:,:) = -pcur_zz(11,:,:)
				pcur_zz(2,:,:) = -pcur_zz(10,:,:)
				pcur_zz(3,:,:) = -pcur_zz(9,:,:)
				pcur_zz(4,:,:) = -pcur_zz(8,:,:)
				pcur_zz(5,:,:) = -pcur_zz(7,:,:)
				pcur_zz(buffer+1,:,:) = 0.0
				!Sigma xz (defined at x-1/2,z-1/2)
				pcur_xz(2,:,:) = -pcur_xz(11,:,:)
				pcur_xz(3,:,:) = -pcur_xz(10,:,:)
				pcur_xz(4,:,:) = -pcur_xz(9,:,:)
				pcur_xz(5,:,:) = -pcur_xz(8,:,:)
				pcur_xz(6,:,:) = -pcur_xz(7,:,:)
				!Vz (defined at x,z-1/2)
				pcur_vz(2,:,:) = pcur_vz(11,:,:)
				pcur_vz(3,:,:) = pcur_vz(10,:,:)
				pcur_vz(4,:,:) = pcur_vz(9,:,:)
				pcur_vz(5,:,:) = pcur_vz(8,:,:)
				pcur_vz(6,:,:) = pcur_vz(7,:,:)
				!Vx (defined at x-1/2,z)
				pcur_vx(1,:,:) = pcur_vx(11,:,:)
				pcur_vx(2,:,:) = pcur_vx(10,:,:)
				pcur_vx(3,:,:) = pcur_vx(9,:,:)
				pcur_vx(4,:,:) = pcur_vx(8,:,:)
				pcur_vx(5,:,:) = pcur_vx(7,:,:)
			endif
			!vx time derivative
			call Dmx_op(.false.,.false.,pcur_xx,pvx_dt)
			call Dpz_op(.false.,.true.,pcur_xz,pvx_dt)
			pvx_dt = pvx_dt*scalevx
			!vz time derivative
			call Dmz_op(.false.,.false.,pcur_zz,pvz_dt)
			call Dpx_op(.false.,.true.,pcur_xz,pvz_dt)
			pvz_dt = pvz_dt*scalevz
			!sigmaxx and sigmazz time derivative
			call Dpz_op(.false.,.false.,pcur_vz,ptemp)
			pxx_dt = ptemp*scale_lame
			pzz_dt = ptemp*scale_lame_mu
			if (free) then
				!Free surface at constant depth of buffer+1 or z=0
				pxx_dt(buffer+1,:,:) = 0.0
			endif
			call Dpx_op(.false.,.false.,pcur_vx,ptemp)
			pxx_dt = pxx_dt + ptemp*scale_lame_mu
			pzz_dt = pzz_dt + ptemp*scale_lame
			if (free) then
				!Free surface at constant depth of buffer+1 or z=0
				pxx_dt(buffer+1,:,:) = pxx_dt(buffer+1,:,:)-ptemp(buffer+1,:,:)*scalefree(:,:)
			endif
			!sigmaxz time derivative
			call Dmx_op(.false.,.false.,pcur_vz,pxz_dt)
			call Dmz_op(.false.,.true.,pcur_vx,pxz_dt)
			pxz_dt = pxz_dt*scale_mu_xz

		endif

  	end subroutine Prop_elastic_step_op

		subroutine dotprod_Prop_elastic_step()
			!Subroutine to test dot-product (unit-test)
			integer																:: nz_dot=300,nx_dot=300,ny_dot=1,nt_dot=1,nv_pad_dot=50
			real																	:: dt_dot=1e-4,dz_dot=5.0,dx_dot=5.0,dy_dot=5.0
			logical																:: twoD_dot=.true.
			real, dimension(:,:,:), pointer				:: lame_dot, mu_dot
			real, dimension(:,:,:), allocatable		:: rhox_dot, rhoz_dot
			!Model and data variables for dot_product
			real, dimension(:,:,:), pointer				:: pcur_xx1, pcur_zz1, pcur_xz1, pcur_vx1, pcur_vz1
			real, dimension(:,:,:), pointer				:: pcur_xx2, pcur_zz2, pcur_xz2, pcur_vx2, pcur_vz2
			real, dimension(:,:,:), pointer				:: pxx_dt1, pzz_dt1,pxz_dt1, pvx_dt1, pvz_dt1
			real, dimension(:,:,:), pointer				:: pxx_dt2, pzz_dt2,pxz_dt2, pvx_dt2, pvz_dt2
			real*8																:: dotm, dotd

			allocate(lame_dot(nz_dot,nx_dot,ny_dot),mu_dot(nz_dot,nx_dot,ny_dot),rhox_dot(nz_dot,nx_dot,ny_dot),rhoz_dot(nz_dot,nx_dot,ny_dot))
			!Creating background model with random numbers
			call random_number(lame_dot)
			lame_dot=(lame_dot*0.1+8.5)*1e9
			call random_number(mu_dot)
			mu_dot=(mu_dot*0.1+2.0)*1e9
			call random_number(rhox_dot)
			rhox_dot=(rhox_dot*100+2000.)
			call random_number(rhoz_dot)
			rhoz_dot=(rhoz_dot*100+2000.)

			!Allocating dot-product variables and filling with random numbers
			allocate(pcur_xx1(nz_dot,nx_dot,ny_dot),pcur_zz1(nz_dot,nx_dot,ny_dot),pcur_xz1(nz_dot,nx_dot,ny_dot),pcur_vx1(nz_dot,nx_dot,ny_dot),pcur_vz1(nz_dot,nx_dot,ny_dot))
			allocate(pcur_xx2(nz_dot,nx_dot,ny_dot),pcur_zz2(nz_dot,nx_dot,ny_dot),pcur_xz2(nz_dot,nx_dot,ny_dot),pcur_vx2(nz_dot,nx_dot,ny_dot),pcur_vz2(nz_dot,nx_dot,ny_dot))
			allocate(pxx_dt1(nz_dot,nx_dot,ny_dot),pzz_dt1(nz_dot,nx_dot,ny_dot),pxz_dt1(nz_dot,nx_dot,ny_dot),pvx_dt1(nz_dot,nx_dot,ny_dot),pvz_dt1(nz_dot,nx_dot,ny_dot))
			allocate(pxx_dt2(nz_dot,nx_dot,ny_dot),pzz_dt2(nz_dot,nx_dot,ny_dot),pxz_dt2(nz_dot,nx_dot,ny_dot),pvx_dt2(nz_dot,nx_dot,ny_dot),pvz_dt2(nz_dot,nx_dot,ny_dot))
			call random_number(pcur_xx1);call random_number(pcur_zz1);call random_number(pcur_xz1);call random_number(pcur_vx1);call random_number(pcur_vz1)
			call random_number(pxx_dt2);call random_number(pzz_dt2);call random_number(pxz_dt2);call random_number(pvx_dt2);call random_number(pvz_dt2)
			pcur_xx2=0.0;pcur_zz2=0.0;pcur_xz2=0.0;pcur_vx2=0.0;pcur_vz2=0.0
			pxx_dt1=0.0;pzz_dt1=0.0;pxz_dt1=0.0;pvx_dt1=0.0;pvz_dt1=0.0

			write(0,*)"Testing dot-product with free surface for elastic stepper"
			call Prop_elastic_step_init(nz_dot,nx_dot,ny_dot,nt_dot,nv_pad_dot,dt_dot,dz_dot,dx_dot,dy_dot,twoD_dot,rhox_dot,rhoz_dot,lame_dot,mu_dot,.true.)
			!Apply forward fields to derivatives
			write(0,*)"Applying forward"
			call Prop_elastic_step_op(.false., pcur_xx1, pcur_zz1, pcur_xz1, pcur_vx1, pcur_vz1, pxx_dt1, pzz_dt1,pxz_dt1, pvx_dt1, pvz_dt1)
			!Apply adjoint derivatives to fields
			write(0,*)"Applying adjoint"
			call Prop_elastic_step_op(.true., pxx_dt2, pzz_dt2, pxz_dt2, pvx_dt2, pvz_dt2, pcur_xx2, pcur_zz2,pcur_xz2, pcur_vx2, pcur_vz2)

			dotm = sum(pcur_xx1*pcur_xx2+pcur_zz1*pcur_zz2+pcur_xz1*pcur_xz2+pcur_vx1*pcur_vx2+pcur_vz1*pcur_vz2)
			dotd = sum(pxx_dt2*pxx_dt1+pzz_dt2*pzz_dt1+pxz_dt2*pxz_dt1+pvx_dt2*pvx_dt1+pvz_dt2*pvz_dt1)
			write(0,*)"Dot product field side = ", dotm
			write(0,*)"Dot product  derivative side = ", dotd
			write(0,*)"Abs Error =", dotm-dotd
			write(0,*)"Rel Error =", (dotm-dotd)/dotm

			pcur_xx2=0.0;pcur_zz2=0.0;pcur_xz2=0.0;pcur_vx2=0.0;pcur_vz2=0.0
			pxx_dt1=0.0;pzz_dt1=0.0;pxz_dt1=0.0;pvx_dt1=0.0;pvz_dt1=0.0

			write(0,*)"Testing dot-product without free surface for elastic stepper"
			call Prop_elastic_step_init(nz_dot,nx_dot,ny_dot,nt_dot,nv_pad_dot,dt_dot,dz_dot,dx_dot,dy_dot,twoD_dot,rhox_dot,rhoz_dot,lame_dot,mu_dot,.false.)
			!Apply forward fields to derivatives
			write(0,*)"Applying forward"
			call Prop_elastic_step_op(.false., pcur_xx1, pcur_zz1, pcur_xz1, pcur_vx1, pcur_vz1, pxx_dt1, pzz_dt1,pxz_dt1, pvx_dt1, pvz_dt1)
			!Apply adjoint derivatives to fields
			write(0,*)"Applying adjoint"
			call Prop_elastic_step_op(.true., pxx_dt2, pzz_dt2, pxz_dt2, pvx_dt2, pvz_dt2, pcur_xx2, pcur_zz2,pcur_xz2, pcur_vx2, pcur_vz2)

			dotm = sum(pcur_xx1*pcur_xx2+pcur_zz1*pcur_zz2+pcur_xz1*pcur_xz2+pcur_vx1*pcur_vx2+pcur_vz1*pcur_vz2)
			dotd = sum(pxx_dt2*pxx_dt1+pzz_dt2*pzz_dt1+pxz_dt2*pxz_dt1+pvx_dt2*pvx_dt1+pvz_dt2*pvz_dt1)
			write(0,*)"Dot product field side = ", dotm
			write(0,*)"Dot product  derivative side = ", dotd
			write(0,*)"Abs Error =", dotm-dotd
			write(0,*)"Rel Error =", (dotm-dotd)/dotm


			deallocate(lame_dot,mu_dot,rhox_dot,rhoz_dot)
			deallocate(pcur_xx1,pcur_zz1,pcur_xz1,pcur_vx1,pcur_vz1)
			deallocate(pcur_xx2,pcur_zz2,pcur_xz2,pcur_vx2,pcur_vz2)
			deallocate(pxx_dt1,pzz_dt1,pxz_dt1,pvx_dt1,pvz_dt1)
			deallocate(pxx_dt2,pzz_dt2,pxz_dt2,pvx_dt2,pvz_dt2)
		end subroutine dotprod_Prop_elastic_step


  	subroutine Prop_elasti_step_close()
  		call D10pt_close()
			if(allocated(scalefree))deallocate(scalefree)
  		if(allocated(ptemp))deallocate(ptemp)
  		if(allocated(scale_mu_xz))deallocate(scale_mu_xz)
  		if(allocated(scale_lame))deallocate(scale_lame)
  		if(allocated(scale_lame_mu))deallocate(scale_lame_mu)
  		if(allocated(scalevx))deallocate(scalevx)
  		if(allocated(scalevz))deallocate(scalevz)
  	end subroutine  Prop_elasti_step_close



end module Prop_elastic_step
