!Module for solving the acoustic wave equation constant density without pointers
!Model dimension (z,x,y,t) Source wavefield
!Data dimension (z,x,y,t) Wavefield
!D(z,x,y,t) = Propag(Model(z,x,y,t))
!Model(z,x,y,t) = Propag*(D(z,x,y,t))
module Propag

  use C2
  use weight
  use G
  use C5

  implicit none
  !Module variables
  integer, private								:: nz, nx, ny, nt
  real,allocatable,dimension(:,:,:),private		:: tmp1,tmp2

  contains


	!Initialization routine
	subroutine Propag_init(nz_in, nx_in, ny_in, nt_in, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		
		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in, nt_in
		real, dimension(:,:,:), pointer			:: vel_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in

		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		nt = nt_in
		if(.not. allocated(tmp1)) allocate(tmp1(nz,nx,ny))
		if(.not. allocated(tmp2)) allocate(tmp2(nz,nx,ny))
 		
		!Initializing other modules
		call C2_init(nz, nx, ny, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)
		call C5_init(nz, nx, ny, vel_in, dt_in, dz_in, dx_in, dy_in, twoD_in)	
		

		return
	end subroutine Propag_init


  	subroutine Propag_call(adj,add,mm,dd)
    		logical,intent(in)   :: adj,add
   			real,dimension(:)  :: mm,dd
    		call Propag_op(adj,add,mm,dd)
  	end subroutine Propag_call

	
	!Operator routine
	subroutine Propag_op(adj, add, mm, dd)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny,nt), intent(inout)		:: mm !Source wavefield
		real, dimension(nz,nx,ny,nt), intent(inout)		:: dd !Wavefield
		!Internal variables
		integer											:: it

    		if(adj) then !Adjoint
        		if (.not. add) mm = 0. !Zero-out the model

			call C5_op(.true.,.false.,tmp1(:,:,:),mm(:,:,:,nt))
			mm(:,:,:,nt-1) = mm(:,:,:,nt-1) + dd(:,:,:,nt) + tmp1(:,:,:)
			do it = nt-2, 1, -1     
			      call C5_op(.true.,.false.,tmp1(:,:,:),mm(:,:,:,it+1))
			      !Operator: (w-I)
			      call weight_op(.true.,.false.,tmp2(:,:,:),mm(:,:,:,it+2))
			      mm(:,:,:,it) = mm(:,:,:,it) + tmp1(:,:,:) + tmp2(:,:,:) + dd(:,:,:,it+1) - mm(:,:,:,it+2) !Adding the source term too
			enddo

		else	     !Forward
			if(.not. add) dd = 0. !Zero-out the data
	
			call C5_op(.false.,.false.,dd(:,:,:,1),tmp1(:,:,:))
			dd(:,:,:,2) = dd(:,:,:,2) + mm(:,:,:,1) + tmp1(:,:,:)

			do it = 3, nt
			      call C5_op(.false.,.false.,dd(:,:,:,it-1),tmp1(:,:,:))
			      !Operator: (w-I)
			      call weight_op(.false.,.false.,dd(:,:,:,it-2),tmp2(:,:,:))
			      dd(:,:,:,it) = dd(:,:,:,it) + tmp1(:,:,:) + tmp2(:,:,:) + mm(:,:,:,it-1) - dd(:,:,:,it-2) !Adding the source term too
			enddo
		endif

		return
	end subroutine Propag_op


	!Subroutine for deallocating variables
	subroutine Propag_close()
		call C2_close()
	  	call C5_close()
	  	if(allocated(tmp2)) deallocate(tmp2)
	  	if(allocated(tmp1)) deallocate(tmp1)
	end subroutine Propag_close


end module Propag

