!!$=head1 NAME
!!$
!!$Propagator Forward - Full acoustic propagator (i.e., varying density) for composite sources.
!!$
!!$=head1 SYNOPSIS
!!$
!!$Progator_fw <bulk.H rho=rho.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H threads= >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Wave equation solver for full acoustic approximation with absorbing boundary. It takes the source and receiver positions from the parameters file created by Velpar.x and propagates the wavelet into
!!$the model and return the data.H. It propagates composite source (i.e., all the sources in the parameter file at once)
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<bulk.H> - Bulk modulus model in SEP3D format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<rho> - density model in SEP3D format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source wavelet to be propagated.
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save wavefield if wantwavefield=1
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<verbose> - [0] Program will write propagation information
!!$
!!$B<integer> I<mod_padded> - [0] Model parameters were padded (100 elements), useful for FWI
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Propagator_acoustic_comp_fw
  
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use Add_source2D
  use Add_source3D
  use resample
  use fourier
  use Sou_scale
  use Prop_acoustic_step


  implicit none
  !Input variables
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: bulk, rho  !bulk modulus and density models 
  real, dimension(:), allocatable						:: wavelet !Source wavelet 
  integer												:: wantwavefield, num_threads
  !Internal variables
  logical												:: twoD
  real, dimension(:,:,:), allocatable, target			:: cur, old
  real, dimension(:,:,:), pointer						:: pnew, pcur, pold
  real, dimension(:,:,:), pointer						:: bulk_pad, rho_pad !Padded models
  real													:: dt !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min velocity
  real, dimension(:,:), allocatable						:: wavelet_fine !Source wavelet fine grid (Composite source)
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad !Padding for the models 
  integer												:: is, ig, it !Cycles variables
  integer												:: wc, srite, rc, sreed
  integer												:: desample
  real, dimension(:,:), allocatable						:: recfield_fine !Receiver field 
  real, dimension(:), allocatable						:: recfield !Receiver field 
  real													:: cr
  integer												:: count1, count2,ierr, verbose, mod_padded
  !Output variables 
  

  parameter( nv_pad=100) !Padding for the absording boundary
  call sep_init (SOURCE)
  
  call from_par ("verbose",verbose,0)		  

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)"#################### Full Acoustic 2D/3D wave propagator #####################"
  if(verbose)write(0,*)"####################          |Composite Source|         #####################"  
  if(verbose)write(0,*)"####################                Forward              #####################"  
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("mod_padded",mod_padded,0)
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("dt",dt,0.)
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the source wavelet
  allocate(wavelet(nw))
  wavelet = 0.
  rc=sreed("wavelet" , wavelet , nw*4)

  !Reading the bulk modulus and rho models and padding if necessary
  if(mod_padded .eq. 0) then
  	allocate(bulk(nz,nx,ny))
  	bulk(:,:,:) = 0.
  	rc=sreed("in" , bulk , nz*nx*ny*4)
  	bulk(:,:,:) = bulk(:,:,:)*1e9 ! Conversion from GPa to Pa
  	allocate(rho(nz,nx,ny))
  	rho(:,:,:) = 0.
  	rc=sreed("rho" , rho , nz*nx*ny*4)
  	!Adding absorbing boundaries to the bulk model
  	call Pad_velocity(bulk,bulk_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	!Adding absorbing boundaries to the density model
  	call Pad_perturbation(rho,rho_pad,nv_pad)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	allocate(bulk_pad(nz,nx,ny))
  	bulk_pad(:,:,:) = 0.
  	rc=sreed("in" , bulk_pad , nz*nx*ny*4)
  	bulk_pad(:,:,:) = bulk_pad(:,:,:)*1e9 ! Conversion from GPa to Pa
  	allocate(rho_pad(nz,nx,ny))
  	rho_pad(:,:,:) = 0.
  	rc=sreed("rho" , rho_pad , nz*nx*ny*4)
  endif  
  !Computing max and min velocities
  minvel = minval(sqrt(bulk_pad/rho_pad))
  maxvel = maxval(sqrt(bulk_pad/rho_pad))
  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(2.))/2.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.50*(min(dx,dy,dz)/maxvel*sqrt(3.))/3.
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minvel/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  if(verbose)write(0,*)"	Min vel [m/s]:",minvel
  if(verbose)write(0,*)"	Max vel [m/s]:",maxvel
  if(verbose)write(0,*) ""
  
  !Scaling source wavelet by 1/(dz*dx) for 2D or 1/(dz*dx*dy) (discretization of delta function!)
  if(twoD)then
  	wavelet=wavelet/(dz*dx)
  else
  	wavelet=wavelet/(dz*dx*dy)
  endif

  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw
  	 allocate(wavelet_fine(nt,ns))
  	 do is = 1,ns
  	 	wavelet_fine(:,is) = wavelet(:)
  	 enddo
  	 desample = 1
  else
 	nt = (nw-1)*desample+1
  	!Source wavelet interpolation to fine time grid
	allocate(wavelet_fine(nt,ns))
  	wavelet_fine = 0.
	!call resample_init(nw, nt)
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet,wavelet_fine(:,1))
	do is = 2,ns
		wavelet_fine(:,is) = wavelet_fine(:,1)
	enddo
  endif
  
  
  !Initialzing the fixed-variable routines
  call Prop_acoustic_step_init(nz, nx, ny, nt, nv_pad, bulk_pad, rho_pad, dt, dz, dx, dy, twoD)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_fine(nt,ng))
  recfield = 0.
  recfield_fine = 0.

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old(nz,nx,ny),cur(nz,nx,ny))
  pold => old
  pcur => cur
  pnew => old
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  
  if (wantwavefield) then
  	if(verbose)write(0,*) "#Program will be writing wavefield as requested#"
  endif


  !Scaling composite source
  do is = 1,ns
	!Scaling the source
	call Sou_scale_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_op(wavelet_fine(:,is),sqrt(bulk_pad/rho_pad)) !Adjoint = Forward (input-output same variable)
	!wavelet_fine(:,is) = wavelet_fine(:,1)
  enddo
  old = 0.; cur = 0.
  if(verbose)write(0,*) "Starting propagation for composite source"
  
  
  call system_clock(count_rate=cr)
  call system_clock(count1) 
  if(verbose)write(0,*)"Propagating..."
  do it = 1, nt !Forward
		!Propagate
		call Prop_acoustic_step_op(.false., pold, pcur, pnew)
		!Adding composite source and Taking wavefield
		if (twoD) then
			do is = 1, ns
				call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
				call Add_source2D_op(.true.,.true.,pnew,wavelet_fine(it,is))
			enddo
			do ig = 1, ng
				call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
				call Add_source2D_op(.false.,.false.,pcur,recfield_fine(it,ig))
			enddo
		else
			do is = 1, ns
				call Add_source3D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy)
				call Add_source3D_op(.true.,.true.,pnew,wavelet_fine(it,is))
			enddo
			do ig = 1,ng
				call Add_source3D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, oy, dz, dx, dy)
				call Add_source3D_op(.false.,.false.,pcur,recfield_fine(it,ig))
			enddo
		endif
		if (wantwavefield) then
			if (mod(it,desample) .eq. 1 .or. desample .eq. 1) then
				wc=srite("wavefield",pcur,nz*nx*ny*4)
			endif
		endif
		if (mod(it,(nt/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(it*100./nt),"%..."
        	close(0)
      	endif
		!Swapping pointers
		pold => pcur 
		pcur => pnew
		pnew => pold
  enddo
  if(verbose)write(0,'(a)',advance="yes")""
  call system_clock(count2)
  if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
  if(verbose)write(0,*) "Writing receiver wavefield for composite source"
  !Resample receivers' wavefield
  do ig = 1,ng
		if (dw .le. dt) then
			wc=srite("out",recfield_fine(:,ig),nw*4)
		else
			call resample_op(.true.,.false.,recfield,recfield_fine(:,ig))
			wc=srite("out",recfield,nw*4)
		endif
  enddo
  if(verbose)write(0,*) ""


  !Closing used routines
  call Prop_acoustic_step_close()
  call Sou_scale_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",ng)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Receiver number")
  if (wantwavefield) then
  
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		if (desample .eq. 1) then
			call to_history("n3",nt,"wavefield")
		else
			call to_history("n3",nt/desample+1,"wavefield")
		endif
  		call to_history("o3",ow,"wavefield")
 	 	call to_history("d3",dt*desample,"wavefield")
  		call to_history("label3","Time [s]","wavefield")
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefield")
		else
			call to_history("n4",nt/desample+1,"wavefield")
		endif
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dt*desample,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"

  !Deallocating memory
  deallocate(cur,old)
  deallocate(wavelet_fine)
  deallocate(recfield_fine,recfield)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(rho_pad,bulk_pad)
  deallocate(wavelet)
  call sep_close()
  call exit (0)
end program Propagator_acoustic_comp_fw


