!!$=head1 NAME
!!$
!!$Propagator Adjoint - Elastic propagator (Lame's parameter, shear modulus, density parametrization).
!!$
!!$=head1 SYNOPSIS
!!$
!!$Propagator_elastic_adj.x <lame.H mu=mu.H rho=rho.H data=data.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H >wavelet.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Elastic wave equation propagator for velocity-stress formulation. Based on both Vireux's (1986) and Alves' (2015) works. Back-propagates data and returns wavelets.
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history
!!$
!!$B<filename> I<model.H> - model parameters in SEP3D format (axis order: z,x,y,parameter).
!!$
!!$=item from par
!!$
!!$B<integer> I<mod_par> - [0] parameterization: 0 -> lame[GPa], mu[GPa], rho[Kg/m3]; 1 -> Vp[m/s], Vs[m/s], rho[Kg/m3]; 2 -> Ip[N s/m3], Is[N s/m3],rho[Kg/m3] 
!!$
!!$B<filename> I<data> - [data.H] Data to be back-propagated (Time,Sigmaxx,sigmazz,sigmaxz,vx,vz,Receiver,Source).
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save wavefield if wantwavefield=1 only velocity and their time derivatives (x,z,dxdt,dzdt)
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<verbose> - [0] Program will write propagation information
!!$
!!$B<integer> I<free_surf> - [0] Applying free-surface boundary conditions at the surface
!!$
!!$B<integer> I<mod_padded> - [0] Model parameters were padded (120 elements), useful for FWI. If free_surf is requested pad model by 5 samples at the top.
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%


program Propagator_elastic_adj

  !Used modules
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use resample
  use S_mod
  use Prop_elastic_step
  use Sou_scale_acoustic
  use Add_source2D
  use absorbing

  implicit none
  character*256											              	  :: sources, recs !Files containing source and receiver positions
  integer												                  :: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													                  :: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling
  integer												                  :: nw !Time sample of the source
  real													                  :: ow, dw !Origin and sampling of the source
  real, dimension(:), allocatable						      			  :: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					    			  :: lame, mu, rho  !lame, shear modulus and density models
  real, dimension(:), allocatable						      			  :: wavelet !Source wavelets coarse scale
  real, dimension(:,:), allocatable						    			  :: recfield_xx_fine, recfield_zz_fine, recfield_xz_fine, recfield_vx_fine, recfield_vz_fine !Receiver field at fine time scale
  real, dimension(:), allocatable						      			  :: recfield !Receiver field to write
  integer												                  :: wc, srite, rc, sreed, sk, sseek_block
  integer												                  :: wantwavefield, num_threads
  logical												                  :: twoD, inj_xx=.false., inj_zz=.false., inj_xz=.false., inj_vx=.false., inj_vz=.false.
  integer												                  :: is, ig !Cycles variables
  integer												                  :: it2, itw, it
  !Arrays needed for propagation
  real, dimension(:,:,:), allocatable, target							  :: cur_xx, old_xx, cur_zz, old_zz, cur_xz, old_xz, cur_vx, old_vx, cur_vz, old_vz
  real, dimension(:,:,:), pointer						      			  :: pnew_xx, pcur_xx, pold_xx, pnew_zz, pcur_zz, pold_zz, pnew_xz, pcur_xz, pold_xz, pnew_vx, pcur_vx, pold_vx, pnew_vz, pcur_vz, pold_vz, ptemp
  real, dimension(:,:,:), pointer						      			  :: lame_pad, mu_pad, rho_pad !Padded models
  real													                  :: dt, dt_inv !Temporal sampling of the propagator
  real													                  :: maxvel, minvel !max and min p-velocity
  real													                  :: minvels, minimum_vel !min s-velocity
  real, dimension(:), allocatable						      			  :: wavelet_xx_fine, wavelet_zz_fine, wavelet_xz_fine, wavelet_vx_fine, wavelet_vz_fine !Source wavelet fine grid
  integer												                  :: nt !Number of time samples for the propagation
  integer												                  :: nv_pad, pad_offset !Padding for the models and offset from absorbing boundary
  integer												                  :: desample, mod_padded
  integer												                  :: verbose, free_in
  logical                                         						  :: free=.false.
  real, dimension(:,:), allocatable						    		   	  :: interp_filter
  real, dimension(:,:,:), allocatable					    			  :: rhoz, rhox, rhoy !Staggered density
  real													                  :: cr
  integer												                  :: count1, count2,ierr, icount, mod_par
  real, dimension(:,:,:), pointer						     		      :: pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt
  real, dimension(:,:,:), pointer						      			  :: pvx_r, pvx_l, pvz_r, pvz_l, pvxdt_r, pvxdt_l, pvzdt_r, pvzdt_l





  parameter( nv_pad=120, pad_offset=20) !Padding parameters for the absording boundary
  call sep_init (SOURCE)


  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("verbose",verbose,0)
  call from_par ("mod_padded",mod_padded,0)
  call from_par ("mod_par",mod_par,0)
  call from_par ("dt",dt,0.)
  !Free surface control
  call from_par ("free_surf",free_in,0)
  if (free_in .ne. 0) free = .true.

  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)"####################      Elastic 2D wave propagator     #####################"
  if(verbose)write(0,*)"####################                Adjoint              #####################"
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""

  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)


  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("data","n1", nw)
  call from_aux ("data","o1", ow);
  call from_aux ("data","d1", dw);

  !Reading the source wavelets for each component (sigmaxx, sigmazz, sigmaxz, vx, vz)
  allocate(wavelet(nw))
  wavelet = 0.

  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)

  !Reading the lame, shear modulus and rho models and padding if necessary
  if(mod_padded .eq. 0) then
  	!Reading the lame modulus
  	allocate(lame(nz,nx,ny))
  	lame(:,:,:) = 0.
  	rc=sreed("in" , lame , nz*nx*ny*4)
  	!Reading the shear modulus
  	allocate(mu(nz,nx,ny))
  	mu(:,:,:) = 0.
 	  rc=sreed("in" , mu , nz*nx*ny*4)
  	!Reading the density model
  	allocate(rho(nz,nx,ny))
  	rho(:,:,:) = 0.
  	rc=sreed("in" , rho , nz*nx*ny*4)
  	!Adding absorbing boundaries to the lame model
  	call Pad_velocity(lame,lame_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD,free)
  	!Adding absorbing boundaries to the mu model
  	call Pad_perturbation(mu,mu_pad,nv_pad,free)
  	!Adding absorbing boundaries to the density model
  	call Pad_perturbation(rho,rho_pad,nv_pad,free)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	!Reading the lame modulus
  	allocate(lame_pad(nz,nx,ny))
  	lame_pad(:,:,:) = 0.
  	rc=sreed("in" , lame_pad , nz*nx*ny*4)
  	!Reading the shear modulus
  	allocate(mu_pad(nz,nx,ny))
  	mu_pad(:,:,:) = 0.
  	rc=sreed("in" , mu_pad , nz*nx*ny*4)
  	!Reading the density model
  	allocate(rho_pad(nz,nx,ny))
  	rho_pad(:,:,:) = 0.
  	rc=sreed("in" , rho_pad , nz*nx*ny*4)
  endif
  !Computing max and min velocities
  minvel = minval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  maxvel = maxval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  if (any(mu_pad .eq. 0.0)) then
  	minvels = 0.0
  	minimum_vel = minvel
  else
  	!minvels = minval(sqrt(mu_pad/rho_pad), mask=(sqrt(mu_pad/rho_pad).gt.0.))
  	minvels = minval(sqrt(mu_pad/rho_pad))
  	minimum_vel = minvels
  endif
  
  !Converting model parameters to lame/mu/rho if they were provided in a different space
  if(mod_par .eq. 0) then
  	lame_pad(:,:,:) = lame_pad(:,:,:)*1e9 ! Conversion from GPa to Pa
  	mu_pad(:,:,:) = mu_pad(:,:,:)*1e9     ! Conversion from GPa to Pa
  elseif(mod_par .eq. 1) then
  	!Converting Vp/Vs/rho to lame/mu/rho
  	mu_pad(:,:,:) = mu_pad(:,:,:)*mu_pad(:,:,:)*rho_pad(:,:,:)
  	lame_pad(:,:,:) = lame_pad(:,:,:)*lame_pad(:,:,:)*rho_pad(:,:,:) - 2.0*mu_pad(:,:,:)
  elseif(mod_par .eq. 2) then
  	!Converting Ip/Is/rho to lame/mu/rho
  	mu_pad(:,:,:) = mu_pad(:,:,:)*mu_pad(:,:,:)/rho_pad(:,:,:)
  	lame_pad(:,:,:) = lame_pad(:,:,:)*lame_pad(:,:,:)/rho_pad(:,:,:) - 2.0*mu_pad(:,:,:)
  endif

  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.5*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		desample = desample*2
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.25*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		desample = desample*2
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
	desample = desample*2
  endif

  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minimum_vel/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  if(verbose)write(0,*)"	Min S vel [m/s]:",minvels
  if(verbose)write(0,*)"	Min P vel [m/s]:",minvel
  if(verbose)write(0,*)"	Max P vel [m/s]:",maxvel
  if(verbose)write(0,*) ""


  !Necessary for time derivative
  dt_inv =1.0/dt


  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw*2 !Because we advance only dt/2 in this implementation
  	 desample = 1
  	 allocate(wavelet_xx_fine(nt),wavelet_zz_fine(nt),wavelet_xz_fine(nt),wavelet_vx_fine(nt),wavelet_vz_fine(nt))
  	 wavelet_xx_fine = 0.
  	 wavelet_zz_fine = 0.
  	 wavelet_xz_fine = 0.
  	 wavelet_vx_fine = 0.
  	 wavelet_vz_fine = 0.
  else
 	nt = (nw-1)*desample+1
  	!Source wavelets interpolation to fine time grid
	allocate(wavelet_xx_fine(nt),wavelet_zz_fine(nt),wavelet_xz_fine(nt),wavelet_vx_fine(nt),wavelet_vz_fine(nt))
  	wavelet_xx_fine = 0.
  	wavelet_zz_fine = 0.
  	wavelet_xz_fine = 0.
  	wavelet_vx_fine = 0.
  	wavelet_vz_fine = 0.
	call resample_init(nt, nw, desample)
  endif

  !Initialzing the fixed-variable routines
  call S_init(nz,nx,ny,twoD)
  call absorbing_init(nz, nx, ny, nv_pad, pad_offset,twoD,free)

  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_xx_fine(nt,ng),recfield_zz_fine(nt,ng),recfield_xz_fine(nt,ng),recfield_vx_fine(nt,ng),recfield_vz_fine(nt,ng))
  recfield = 0.
  recfield_xx_fine = 0.; recfield_zz_fine = 0.; recfield_xz_fine = 0.; recfield_vx_fine = 0.; recfield_vz_fine = 0.

  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))

  if (verbose)then
    if(free)write(0,*)"Requested free surface boundary conditions"
  endif

  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old_xx(nz,nx,ny),cur_xx(nz,nx,ny),old_zz(nz,nx,ny),cur_zz(nz,nx,ny),old_xz(nz,nx,ny),cur_xz(nz,nx,ny),old_vx(nz,nx,ny),cur_vx(nz,nx,ny),old_vz(nz,nx,ny),cur_vz(nz,nx,ny))
  allocate(pxx_dt(nz,nx,ny), pzz_dt(nz,nx,ny), pxz_dt(nz,nx,ny), pvx_dt(nz,nx,ny), pvz_dt(nz,nx,ny))
  pold_xx => old_xx; pold_zz => old_zz; pold_xz => old_xz; pold_vx => old_vx; pold_vz => old_vz
  pcur_xx => cur_xx; pcur_zz => cur_zz; pcur_xz => cur_xz; pcur_vx => cur_vx; pcur_vz => cur_vz
  pnew_xx => old_xx; pnew_zz => old_zz; pnew_xz => old_xz; pnew_vx => old_vx; pnew_vz => old_vz
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""




  !Staggering density perturbation
  allocate(rhoz(nz,nx,ny), rhox(nz,nx,ny))
  call Sz_op(.true., .false., rhoz, rho_pad)
  call Sx_op(.true., .false., rhox, rho_pad)
  !Staggering on y if not twoD problem
  if(.not. twoD) then
  	allocate(rhoy(nz,nx,ny))
  	call Sy_op(.true., .false., rhoy, rho_pad)
  endif


  !allocating memory for linear interpolation of velocity wavefields
  if (wantwavefield) then
  	allocate(pvx_r(nz,nx,ny), pvx_l(nz,nx,ny), pvz_r(nz,nx,ny), pvz_l(nz,nx,ny), pvxdt_r(nz,nx,ny), pvxdt_l(nz,nx,ny), pvzdt_r(nz,nx,ny), pvzdt_l(nz,nx,ny))
  endif


  call Prop_elastic_step_init(nz, nx, ny, nt, nv_pad, dt, dz, dx, dy, twoD, rhox, rhoz, lame_pad, mu_pad, free)


  if (wantwavefield) then
  	if(verbose)write(0,*) "#Program will be writing wavefield as requested#"
  		do is = 1,ns
			do it = 1, nw
				wc=srite("wavefield",pcur_vx,nz*nx*ny*4)
				wc=srite("wavefield",pcur_vz,nz*nx*ny*4)
				wc=srite("wavefield",pcur_vx,nz*nx*ny*4)
				wc=srite("wavefield",pcur_vx,nz*nx*ny*4)
			enddo
		enddo
  endif

  do is = 1,ns
  	old_xx = 0.; cur_xx = 0.
  	old_zz = 0.; cur_zz = 0.
  	old_xz = 0.; cur_xz = 0.
  	old_vx = 0.; cur_vx = 0.
  	old_vz = 0.; cur_vz = 0.
  	if (wantwavefield) then
  		pvx_r = 0.; pvx_l = 0.
  		pvz_r = 0.; pvz_l = 0.
  		pvxdt_r = 0.; pvxdt_l = 0.
  		pvzdt_r = 0.; pvzdt_l = 0.
  	endif
  	icount = 0
  	if (wantwavefield) sk = sseek_block("wavefield",4*nw*(is),4*nz*nx*ny,0)
    if(verbose)write(0,'(A42,I5)') "Starting back propagation for source n°: ", is


	!Creating adjoint source at the receivers
	if (dw .le. dt) then
		do ig = 1,ng
			rc=sreed("data",recfield,nw*4)
			recfield_xx_fine(:,ig) = recfield
			rc=sreed("data",recfield,nw*4)
			recfield_zz_fine(:,ig) = recfield
			rc=sreed("data",recfield,nw*4)
			recfield_xz_fine(:,ig) = recfield
			rc=sreed("data",recfield,nw*4)
			recfield_vx_fine(:,ig) = recfield
			rc=sreed("data",recfield,nw*4)
			recfield_vz_fine(:,ig) = recfield
		enddo
	else
		do ig = 1,ng
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_xx_fine(:,ig))
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_zz_fine(:,ig))
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_xz_fine(:,ig))
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_vx_fine(:,ig))
			rc=sreed("data",recfield,nw*4)
			call resample_op(.false.,.false.,recfield,recfield_vz_fine(:,ig))
		enddo
	endif
	!Checking with component is non-zero
	if (any(recfield_xx_fine .ne. 0.0)) inj_xx=.true.
	if (any(recfield_zz_fine .ne. 0.0)) inj_zz=.true.
	if (any(recfield_xz_fine .ne. 0.0)) inj_xz=.true.
	if (any(recfield_vx_fine .ne. 0.0)) inj_vx=.true.
	if (any(recfield_vz_fine .ne. 0.0)) inj_vz=.true.


	call system_clock(count_rate=cr)
	call system_clock(count1)

	do itw = nw, 2, -1 !Adjoint
		it = (itw-2)*desample
		!Propagate receiver field
		do it2 = desample+1, 2, -1


			!Applying absorbing boundary condition
			call absorbing_op(pnew_xx)
			call absorbing_op(pnew_zz)
			call absorbing_op(pnew_xz)
			call absorbing_op(pnew_vx)
			call absorbing_op(pnew_vz)


			!computing time derivative for each component
			call Prop_elastic_step_op(.true., pcur_xx, pcur_zz, pcur_xz, pcur_vx, pcur_vz, pxx_dt, pzz_dt,pxz_dt, pvx_dt, pvz_dt)

			!Adding receiver data to the time derivative of each component
			if (inj_xx .or. inj_zz) then
				do ig = 1, ng
					!Adding sigmaxx and sigmazz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pxx_dt,recfield_xx_fine(it+it2,ig))
					call Add_source2D_op(.true.,.true.,pzz_dt,recfield_zz_fine(it+it2,ig))
				enddo
			endif

			if (inj_xz) then
				do ig = 1, ng
					!Adding sigmaxz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz-(dz/2.0), ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pxz_dt,recfield_xz_fine(it+it2,ig))
				enddo
			endif

			if (inj_vx) then
				do ig = 1, ng
					!Adding sigmavx receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox-(dx/2.0), dz, dx)
					call Add_source2D_op(.true.,.true.,pvx_dt,recfield_vx_fine(it+it2,ig))
				enddo
			endif

			if (inj_vz) then
				do ig = 1, ng
					!Adding sigmavz receivers
					call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz-(dz/2.0), ox, dz, dx)
					call Add_source2D_op(.true.,.true.,pvz_dt,recfield_vz_fine(it+it2,ig))
				enddo
			endif


			if (wantwavefield) then
				!Apply adjoint(fine to coarse) of linear interpolation
				!On Velocities
				pvx_l = pvx_l + pcur_vx*interp_filter(it2,1)
				pvx_r = pvx_r + pcur_vx*interp_filter(it2,2)
				pvz_l = pvz_l + pcur_vz*interp_filter(it2,1)
				pvz_r = pvz_r + pcur_vz*interp_filter(it2,2)
				!On derivatives
				pvxdt_l = pvxdt_l + pvx_dt*interp_filter(it2,1)*dt_inv
				pvxdt_r = pvxdt_r + pvx_dt*interp_filter(it2,2)*dt_inv
				pvzdt_l = pvzdt_l + pvz_dt*interp_filter(it2,1)*dt_inv
				pvzdt_r = pvzdt_r + pvz_dt*interp_filter(it2,2)*dt_inv
			endif


			!Computing the new time step
			pnew_xx = pold_xx + pxx_dt
			pnew_zz = pold_zz + pzz_dt
			pnew_xz = pold_xz + pxz_dt
			pnew_vz = pold_vz + pvz_dt
			pnew_vx = pold_vx + pvx_dt

			!Taking the wavefields of each component at the source position
			!Taking sigmaxx
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
			call Add_source2D_op(.false.,.false.,pcur_xx,wavelet_xx_fine(it+it2))
			!Taking sigmazz
			call Add_source2D_op(.false.,.false.,pcur_zz,wavelet_zz_fine(it+it2))

			!Taking sigmaxz
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz-(dz/2.0), ox-(dx/2.0), dz, dx)
			call Add_source2D_op(.false.,.false.,pcur_xz,wavelet_xz_fine(it+it2))

			!Taking vx
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox-(dx/2.0), dz, dx)
			call Add_source2D_op(.false.,.false.,pcur_vx,wavelet_vx_fine(it+it2))

			!Taking vz
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz-(dz/2.0), ox, dz, dx)
			call Add_source2D_op(.false.,.false.,pcur_vz,wavelet_vz_fine(it+it2))



			!Swapping pointers
			pold_xx => pcur_xx
			pcur_xx => pnew_xx
			pnew_xx => pold_xx
			pold_zz => pcur_zz
			pcur_zz => pnew_zz
			pnew_zz => pold_zz
			pold_xz => pcur_xz
			pcur_xz => pnew_xz
			pnew_xz => pold_xz
			pold_vx => pcur_vx
			pcur_vx => pnew_vx
			pnew_vx => pold_vx
			pold_vz => pcur_vz
			pcur_vz => pnew_vz
			pnew_vz => pold_vz

		enddo

		!saving source wavefield
		if (wantwavefield) then

			sk = sseek_block("wavefield",-4,4*nz*nx*ny,1)
			wc=srite("wavefield",pvx_r*dt/(rhox*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvz_r*dt/(rhoz*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvxdt_r*dt/(rhox*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvzdt_r*dt/(rhoz*desample),nz*nx*ny*4)
			sk = sseek_block("wavefield",-4,4*nz*nx*ny,1)

			!Swap source wavefield left and right pointers
			ptemp => pvx_l
			pvx_l => pvx_r
			pvx_r => ptemp
			pvx_l = 0.

			ptemp => pvz_l
			pvz_l => pvz_r
			pvz_r => ptemp
			pvz_l = 0.

			ptemp => pvxdt_l
			pvxdt_l => pvxdt_r
			pvxdt_r => ptemp
			pvxdt_l = 0.

			ptemp => pvzdt_l
			pvzdt_l => pvzdt_r
			pvzdt_r => ptemp
			pvzdt_l = 0.
		endif

		icount = icount + 1
		if (mod(icount,(nw/10))==0) then
	        	if(verbose)write(0,'(I3,A4)',advance="no")nint((icount)*100./nw),"%..."
	        	close(0)
		endif


	enddo

		!Last step of the source wavefield just saving source wavefield
		if (wantwavefield) then
			pvx_r(:,:,:) = pvx_r(:,:,:) + pcur_vx*interp_filter(1,2)
			pvz_r(:,:,:) = pvz_r(:,:,:) + pcur_vz*interp_filter(1,2)
			pvzdt_r(:,:,:) = pvzdt_r(:,:,:) + pvz_dt*interp_filter(1,2)*dt_inv
			pvxdt_r(:,:,:) = pvxdt_r(:,:,:) + pvx_dt*interp_filter(1,2)*dt_inv
			sk = sseek_block("wavefield",-4,4*nz*nx*ny,1)
			wc=srite("wavefield",pvx_r*dt/(rhox*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvz_r*dt/(rhoz*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvxdt_r*dt/(rhox*desample),nz*nx*ny*4)
			wc=srite("wavefield",pvzdt_r*dt/(rhoz*desample),nz*nx*ny*4)
			sk = sseek_block("wavefield",-4,4*nz*nx*ny,1)
		endif

		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	if(verbose)write(0,'(a)',advance="yes")""


	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing source wavelet for source n°: ", is

	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, 1.0, twoD)
	!Scaling source wavelet by 1/(dz*dx) for 2D (discretization of delta function!)
	wavelet_xx_fine=wavelet_xx_fine/(dz*dx)
	wavelet_zz_fine=wavelet_zz_fine/(dz*dx)
	wavelet_xz_fine=wavelet_xz_fine/(dz*dx)
	wavelet_vx_fine=wavelet_vx_fine/(dz*dx)
	wavelet_vz_fine=wavelet_vz_fine/(dz*dx)
	
	!Resample source wavelets and scaling operation
	if (dw .le. dt) then
		wavelet_xx_fine=dt*wavelet_xx_fine
		wc=srite("out",wavelet_xx_fine,nw*4)
		wavelet_zz_fine=dt*wavelet_zz_fine
		wc=srite("out",wavelet_zz_fine,nw*4)
		wavelet_xz_fine=dt*wavelet_xz_fine
		wc=srite("out",wavelet_xz_fine,nw*4)
		wavelet_vx_fine=dt*wavelet_vx_fine
		wc=srite("out",wavelet_vx_fine,nw*4)
		wavelet_vz_fine=dt*wavelet_vz_fine
		wc=srite("out",wavelet_vz_fine,nw*4)
	else
		call resample_op(.true.,.false.,wavelet,wavelet_xx_fine)
		wavelet=dt*wavelet
		wc=srite("out",wavelet,nw*4)
		call resample_op(.true.,.false.,wavelet,wavelet_zz_fine)
		wavelet=dt*wavelet
		wc=srite("out",wavelet,nw*4)
		call resample_op(.true.,.false.,wavelet,wavelet_xz_fine)
		wavelet=dt*wavelet
		wc=srite("out",wavelet,nw*4)
		call resample_op(.true.,.false.,wavelet,wavelet_vx_fine)
		call Sou_scale_acoustic_op(wavelet,dt/rho_pad)
		wc=srite("out",wavelet,nw*4)
		call resample_op(.true.,.false.,wavelet,wavelet_vz_fine)
		call Sou_scale_acoustic_op(wavelet,dt/rho_pad)
		wc=srite("out",wavelet,nw*4)
	endif
   	if(verbose)write(0,*) ""



  enddo

  call Prop_elasti_step_close()
  call Sou_scale_acoustic_close()
  call absorbing_close()

  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  call to_history("n2",5)
  call to_history("o2",1.)
  call to_history("d2",1.)
  call to_history("label2","Component [Sigmaxx,Sigmazz,Sigmaxz,Vx,Vz]")
  call to_history("n3",ns)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Source number")
  if (wantwavefield) then
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
  		call to_history("n3",4,"wavefield")
  		call to_history("o3",1.,"wavefield")
  		call to_history("d3",1.,"wavefield")
  		call to_history("label3","Component [Vx,Vz,dVxdt,dVzdt]","wavefield")
  		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefield")
		else
			call to_history("n4",nt/desample+1,"wavefield")
		endif
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dw,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
  		call to_history("n4",6,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Component [Vx,Vy,Vz,dVxdt.dVydt,dVzdt]","wavefield")
  		if (desample .eq. 1) then
			call to_history("n5",nt,"wavefield")
		else
			call to_history("n5",nt/desample+1,"wavefield")
		endif
  		call to_history("o5",ow,"wavefield")
 	 	call to_history("d5",dw,"wavefield")
  		call to_history("label5","Time [s]","wavefield")
  		call to_history("n6",ns,"wavefield")
  		call to_history("o6",1.,"wavefield")
  		call to_history("d6",1.,"wavefield")
  		call to_history("label6","Source number","wavefield")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"



  !Deallocating memory
  if (wantwavefield) deallocate(pvzdt_l,pvzdt_r,pvxdt_l,pvxdt_r,pvz_l,pvz_r,pvx_l,pvx_r)
  if(allocated(rhoy)) deallocate(rhoy)
  deallocate(rhox,rhoz)
  deallocate(pvz_dt, pvx_dt, pxz_dt, pxx_dt, pzz_dt)
  deallocate(cur_vz,old_vz,cur_vx,old_vx,cur_xz,old_xz,cur_zz,old_zz,cur_xx,old_xx)
  deallocate(interp_filter)
  deallocate(recfield_vz_fine,recfield_vx_fine,recfield_xz_fine,recfield_zz_fine,recfield_xx_fine,recfield)
  deallocate(wavelet_vz_fine,wavelet_vx_fine,wavelet_xz_fine,wavelet_zz_fine,wavelet_xx_fine)
  deallocate(rho_pad,mu_pad,lame_pad)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(wavelet)
  call sep_close()
  call exit (0)


end program Propagator_elastic_adj
