!!$=head1 NAME
!!$
!!$Propagator Forward - Elastic propagator for displacement formulation (Lame's parameter, shear modulus, density parametrization).
!!$
!!$=head1 SYNOPSIS
!!$
!!$Propagator_elastic_disp_fw.x <lame.H mu=mu.H rho=rho.H wavelet=wavelet.H sources=input_sources.p recs=input_recs.p wantwavefield=0 wavefield=wavefield.H >data.H 
!!$
!!$=head1 DESCRIPTION
!!$
!!$Elastic wave equation propagator for displacement formulation. 
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<lame.H> - lame parameter model in SEP3D format (theoretically [GPa]).
!!$
!!$=item from par
!!$
!!$B<filename> I<mu> - shear modulus in SEP3D format (theoretically [GPa]).
!!$
!!$B<filename> I<rho> - density model in SEP3D format (theoretically [Kg/m^3]).
!!$
!!$B<filename> I<wavelet> - [wavelet.H] Source force wavelet to be propagated for each component (fz,fx).
!!$
!!$B<filename> I<sources> - [input_sources.p] File containing number of sources and their positions.
!!$
!!$B<filename> I<recs> - [input_recs.p] File containing number of receivers and their positions.
!!$
!!$B<integer> I<wantwavefield> - [0] If 1 the program saves the wavefield.
!!$
!!$B<filename> I<wavefield> - [wavefield.H] File where to save source wavefield if wantwavefield=1 (ux,uz)
!!$
!!$B<real> I<dt> - Override internal dt definition (optional) NB!mod(dt_wavelet/dt_wavefield)=0
!!$
!!$B<integer> I<threads> - [16] Number of threads to use during propagation.
!!$
!!$B<integer> I<inj_p> - [0] If 1 the program will inject the source components as a explosive source (i.e., dipole)
!!$
!!$B<integer> I<rec_p> - [0] If 1 the program will record the pressure as well as the particle displacements
!!$
!!$B<integer> I<mod_padded> - [0] Model parameters were padded (120 elements), useful for FWI
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, read_pos, Pad_interp3D, Pad_interp3D, resample
!!$
!!$=head1 SEE ALSO
!!$
!!$Velpar, Vel, Geometry.sh
!!$
!!$=cut
!%

program Propagator_elastic_disp_fw
  
  !Used modules 
  use sep
  use read_pos
  use Pad_vel
  use Pad_pert
  use resample 
  use Prop_elastic_disp_step
  use Sou_scale_acoustic
  use Add_source2D_pressure
  use Add_source2D
  !use absorbing


  implicit none
  character*256											:: sources, recs !Files containing source and receiver positions
  integer												:: nz, nx, ny, ns, ng !Dimension of the model and # of sources and receivers
  real													:: oz, ox, oy, dz, dx, dy !Origin of the model and spatial sampling 
  integer												:: nw !Time sample of the source
  real													:: ow, dw !Origin and sampling of the source 
  real, dimension(:), allocatable						:: sx, sy, sz, gx, gy, gz !Position of sources and receivers
  real, dimension(:,:,:), allocatable					:: lame, mu, rho  !lame, shear modulus and density models 
  real, dimension(:), allocatable						:: wavelet_fx, wavelet_fz !Source wavelets for the different (force) components 
  real, dimension(:,:), allocatable						:: recfield_ux_fine, recfield_uz_fine, recfield_P_fine !Receiver field at fine time scale
  real, dimension(:), allocatable						:: recfield !Receiver field to write
  integer												:: wc, srite, rc, sreed
  integer												:: wantwavefield, num_threads, inj_p, rec_p
  logical												:: twoD
  integer												:: is, ig !Cycles variables
  integer												:: it2, itw, it
  !Arrays needed for propagation
  real, dimension(:,:,:), allocatable, target			:: cur_ux, old_ux, cur_uz, old_uz
  real, dimension(:,:,:), pointer						:: pnew_ux, pcur_ux, pold_ux, pnew_uz, pcur_uz, pold_uz, ptemp
  real, dimension(:,:,:), pointer						:: lame_pad, mu_pad, rho_pad !Padded models
  real													:: dt, dt_inv !Temporal sampling of the propagator
  real													:: maxvel, minvel !max and min p-velocity
  real													:: minvels, minimum_vel !min s-velocity
  real, dimension(:), allocatable						:: wavelet_fx_fine, wavelet_fz_fine !Source wavelet fine grid 
  real, dimension(:), allocatable						:: scaled_fx_wavelet, scaled_fz_wavelet
  integer												:: nt !Number of time samples for the propagation
  integer												:: nv_pad, pad_offset !Padding for the models and offset from absorbing boundary
  integer												:: desample
  integer												:: verbose
  real*8, dimension(:,:), allocatable					:: interp_filter
  real, dimension(:,:,:), allocatable					:: rhoz, rhox, rhoy !Staggered density
  real													:: cr
  integer												:: count1, count2,ierr, mod_padded
  real, dimension(:,:,:), pointer						:: pux_r, pux_l, puz_r, puz_l


  parameter(nv_pad=120, pad_offset=20) !Padding parameters for the absording boundary
  call sep_init (SOURCE)
  
  
  !Reading parameters
  call from_par ("sources",sources,"input_sources.p")
  call from_par ("recs",recs,"input_recs.p")
  call from_par ("wantwavefield",wantwavefield,0)
  call from_par ("verbose",verbose,0)
  call from_par ("mod_padded",mod_padded,0)
  call from_par ("dt",dt,0.)
  call from_par ("inj_p",inj_p,0)
  call from_par ("rec_p",rec_p,0)
  
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)"####################      Elastic 2D wave propagator     #####################"
  if(verbose)write(0,*)"####################       Displacement formulation      #####################"
  if(verbose)write(0,*)"####################               Forward               #####################"  
  if(verbose)write(0,*)"##############################################################################"
  if(verbose)write(0,*)""
  if(verbose)write(0,*)""
  
  !Setting number of threads
  call from_par("threads",num_threads,16)
  call omp_set_num_threads(num_threads)
  
  !Reading information from input files
  call from_history("n1", nz)
  call from_history("o1", oz);
  call from_history("d1", dz);
  call from_history("n2", nx)
  call from_history("o2", ox);
  call from_history("d2", dx);
  call from_history("n3", ny)
  call from_history("o3", oy);
  call from_history("d3", dy); if(dy .eq. 0.) dy = min(dz,dx); ! In case d3 is 0.0
  call from_aux ("wavelet","n1", nw)
  call from_aux ("wavelet","o1", ow);
  call from_aux ("wavelet","d1", dw);

 !Reading the source wavelets for each component (fz, fx)
  allocate(wavelet_fx(nw),wavelet_fz(nw))
  wavelet_fx = 0.; wavelet_fz = 0.;
  rc=sreed("wavelet" , wavelet_fz , nw*4)
  rc=sreed("wavelet" , wavelet_fx , nw*4)
  
  !Reading position of sources and receivers from files
  call reading_pos(sources,ns,sx,sy,sz)
  call reading_pos(recs,ng,gx,gy,gz)
  
  !Reading the lame, shear modulus and rho models and padding if necessary
  if(mod_padded .eq. 0) then
  	!Reading the lame modulus
  	allocate(lame(nz,nx,ny))
  	lame(:,:,:) = 0.
  	rc=sreed("in" , lame , nz*nx*ny*4)
  	lame(:,:,:) = lame(:,:,:)*1e9 ! Conversion from GPa to Pa
  	!Reading the shear modulus
  	allocate(mu(nz,nx,ny))
  	mu(:,:,:) = 0.
 	rc=sreed("mu" , mu , nz*nx*ny*4)
  	mu(:,:,:) = mu(:,:,:)*1e9 ! Conversion from GPa to Pa
  	!Reading the density model
  	allocate(rho(nz,nx,ny))
  	rho(:,:,:) = 0.
  	rc=sreed("rho" , rho , nz*nx*ny*4)
  	!Adding absorbing boundaries to the lame model
  	call Pad_velocity(lame,lame_pad,nv_pad,oz,oy,ox,nz,nx,ny,dz,dx,dy,twoD)
  	!Adding absorbing boundaries to the mu model
  	call Pad_perturbation(mu,mu_pad,nv_pad)
  	!Adding absorbing boundaries to the density model
  	call Pad_perturbation(rho,rho_pad,nv_pad)
  else
  	if ((nz .eq. 1) .or. (nx .eq. 1) .or. (ny .eq. 1)) twoD=.true.
  	!Reading the lame modulus
  	allocate(lame_pad(nz,nx,ny))
  	lame_pad(:,:,:) = 0.
  	rc=sreed("in" , lame_pad , nz*nx*ny*4)
  	lame_pad(:,:,:) = lame_pad(:,:,:)*1e9 ! Conversion from GPa to Pa
  	!Reading the shear modulus
  	allocate(mu_pad(nz,nx,ny))
  	mu_pad(:,:,:) = 0.
  	rc=sreed("mu" , mu_pad , nz*nx*ny*4)
  	mu_pad(:,:,:) = mu_pad(:,:,:)*1e9 ! Conversion from GPa to Pa
  	!Reading the density model
  	allocate(rho_pad(nz,nx,ny))
  	rho_pad(:,:,:) = 0.
  	rc=sreed("rho" , rho_pad , nz*nx*ny*4)
  endif  
  !Computing max and min velocities
  minvel = minval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  maxvel = maxval(sqrt((lame_pad+2.0*mu_pad)/rho_pad))
  if (any(mu_pad .eq. 0.0)) then
  	minvels = 0.0
  	minimum_vel = minvel
  else
  	!minvels = minval(sqrt(mu_pad/rho_pad), mask=(sqrt(mu_pad/rho_pad).gt.0.))
  	minvels = minval(sqrt(mu_pad/rho_pad))
  	minimum_vel = minvels
  endif
  
  
  if(verbose)write(0,*)"Temporal information:"
  !Computing dt for numerical stability
  !Checking if 2D or 3D propagation
  if (dt .eq. 0.) then
  	if(twoD) then
		dt = 0.50*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	2D propagation with dt = ", dt,"[s]"
  	else
		dt = 0.25*min(dx,dy,dz)/(sqrt(2.0)*maxvel)
		desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
		dt = dw/desample
		if(verbose)write(0,*) "	3D propagation with dt = ", dt,"[s]"
  	endif
  else
  	if(verbose)write(0,*)"	Overriding internal dt definition, dt = ", dt,"[s]"
  	desample = floor(dw/dt);if(mod(dw,dt).ne.0)desample=desample+1
	dt = dw/desample
  endif
  
  
  if(verbose)write(0,*) "	Temporal Nyquist frequency: ", 1./(2.*dt),"[Hz]"
  if(verbose)write(0,*) "	Number of samples for 50 Hz wave = ", (minimum_vel/(50.*max(dx,dy,dz)))
  if(verbose)write(0,*) "	CFL number: ",dt*maxvel/min(dx,dy,dz)
  if(verbose)write(0,*)"	Min S vel [m/s]:",minvels
  if(verbose)write(0,*)"	Min P vel [m/s]:",minvel
  if(verbose)write(0,*)"	Max P vel [m/s]:",maxvel
  if(verbose)write(0,*) ""
  
  !Scaling source wavelet by 1/(dz*dx) for 2D (discretization of delta function!)
  wavelet_fx=wavelet_fx/(dz*dx)
  wavelet_fz=wavelet_fz/(dz*dx)
  
  
  !Computing number of samples for the propagation
  if (dw .le. dt) then
  	 if(verbose)write(0,*)"Changing temporal step to original dt: ", dw
  	 nt = nw
  	 dt = dw 
  	 desample = 1
  	 allocate(wavelet_fx_fine(nt),wavelet_fz_fine(nt))
  	 allocate(scaled_fx_wavelet(nt),scaled_fz_wavelet(nt))
  	  wavelet_fx_fine = wavelet_fx;
  	  wavelet_fz_fine = wavelet_fz;
  else
 	nt = (nw-1)*desample+1
  	!Source wavelets interpolation to fine time grid
	allocate(wavelet_fx_fine(nt),wavelet_fz_fine(nt))
	allocate(scaled_fx_wavelet(nt),scaled_fz_wavelet(nt))
	!call resample_init(nw, nt)
	call resample_init(nt, nw, desample)
	call resample_op(.false.,.false.,wavelet_fx,wavelet_fx_fine)
	call resample_op(.false.,.false.,wavelet_fz,wavelet_fz_fine)
  endif
  
  !Initialzing the fixed-variable routines
  !call absorbing_init(nz, nx, ny, nv_pad, pad_offset,twoD)
  !Allocating space for receiver trace
  allocate(recfield(nw),recfield_ux_fine(nt,ng),recfield_uz_fine(nt,ng),recfield_P_fine(nt,ng))
  recfield = 0.
  recfield_ux_fine = 0.; recfield_uz_fine = 0.; recfield_P_fine = 0.;
  
  !precompute the linear interpolation filter
  allocate(interp_filter(desample+1,2))
  do it2=1,desample+1
	interp_filter(it2,1)=1.-float((it2-1))/float(desample)
	interp_filter(it2,2)=1.-interp_filter(it2,1)
  end do
  interp_filter = interp_filter * sqrt(float(nw)/float(nt))  
  
  if(verbose)write(0,*)"Allocating memory for propagation"
  allocate(old_ux(nz,nx,ny),cur_ux(nz,nx,ny),cur_uz(nz,nx,ny),old_uz(nz,nx,ny))
  pold_ux => old_ux; pold_uz => old_uz;
  pcur_ux => cur_ux; pcur_uz => cur_uz;
  pnew_ux => old_ux; pnew_uz => old_uz;
  if(verbose)write(0,*) "Dimensions:"
  if(verbose)write(0,'(A7,I6)') " nz = ",nz
  if(verbose)write(0,'(A7,I6)') " nx = ",nx
  if(verbose)write(0,'(A7,I6)') " ny = ",ny
  if(verbose)write(0,'(A7,I6)') " nt = ",nt
  if(verbose)write(0,'(A7,I6)') " ns = ",ns
  if(verbose)write(0,'(A7,I6)') " ng = ",ng
  if(verbose)write(0,*) ""
  
  !allocating memory for linear interpolation of velocity wavefields
  if (wantwavefield) then
  	allocate(pux_r(nz,nx,ny), pux_l(nz,nx,ny), puz_r(nz,nx,ny), puz_l(nz,nx,ny))
  endif
  call Prop_elastic_disp_step_init(nz, nx, ny, nt, dt, dz, dx, dy, twoD, rho_pad, lame_pad, mu_pad, nv_pad-pad_offset)
  call Add_source2D_pressure_init2(lame_pad,mu_pad,nz,nx,ny)
  
  if(verbose) then   
  	if (wantwavefield) write(0,*) "#Program will be writing wavefield as requested#"
  endif
  
  do is = 1,ns
  	old_ux = 0.; cur_ux = 0.
  	old_uz = 0.; cur_uz = 0.
  	if (wantwavefield) then
  		pux_r = 0.; pux_l = 0.
  		puz_r = 0.; puz_l = 0.
  	endif	
	if(verbose)write(0,'(A38,I5)') "Starting propagation for source n°: ", is
	!Scaling the force source
	scaled_fx_wavelet = wavelet_fx_fine
	scaled_fz_wavelet = wavelet_fz_fine
	call Sou_scale_acoustic_init(nz, nx, ny, sz(is), sx(is), sy(is), oz, ox, oy, dz, dx, dy, dt, twoD)
	call Sou_scale_acoustic_op(scaled_fx_wavelet,1.0/rho_pad*(dt*dt)) !Adjoint = Forward (input-output same variable)
	call Sou_scale_acoustic_op(scaled_fz_wavelet,1.0/rho_pad*(dt*dt))
	!Timing
	call system_clock(count_rate=cr)
	call system_clock(count1)
	
	do itw = 1, nw-1 !Forward
		it = (itw-1)*desample
		!Propagate source field
		do it2 = 1, desample
			!computing time derivative for each component
			call Prop_elastic_step_disp_op(.false., pold_ux,pold_uz,pcur_ux,pcur_uz,pnew_ux,pnew_uz)
			!Adding source to each component
			call Add_source2D_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
			call Add_source2D_pressure_init(nz, nx, ny, nt, sz(is), sx(is), sy(is), oz, ox, dz, dx)
			if (inj_p) then
				!Adding fz source
				call Add_source2D_pressurez_op(.true.,.true.,pnew_uz,wavelet_fz_fine(it+it2))
				!Adding fx source
				call Add_source2D_pressurex_op(.true.,.true.,pnew_ux,wavelet_fx_fine(it+it2))
			else
				!Adding fz source
				call Add_source2D_op(.true.,.true.,pnew_uz,wavelet_fz_fine(it+it2))
				!Adding fx source
				call Add_source2D_op(.true.,.true.,pnew_ux,wavelet_fx_fine(it+it2))
			endif

		
			
			if (wantwavefield) then
				!Apply adjoint(fine to coarse) of linear interpolation
				!On displacements
				pux_l = pux_l + pcur_ux*interp_filter(it2,1)
				pux_r = pux_r + pcur_ux*interp_filter(it2,2)
				puz_l = puz_l + pcur_uz*interp_filter(it2,1)
				puz_r = puz_r + pcur_uz*interp_filter(it2,2)
			endif
			
			
			
			!Taking the wavefields of each component
			do ig = 1, ng
				!ux and uz
				call Add_source2D_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
				call Add_source2D_op(.false.,.false.,pcur_ux,recfield_ux_fine(it+it2,ig))
				call Add_source2D_op(.false.,.false.,pcur_uz,recfield_uz_fine(it+it2,ig))
			enddo
			
			!Recording pressure
			if (rec_p) then
				do ig = 1, ng
					call Add_source2D_pressure_init(nz, nx, ny, nt, gz(ig), gx(ig), gy(ig), oz, ox, dz, dx)
					call Add_source2D_pressurez_op(.false.,.false.,pcur_uz,recfield_P_fine(it+it2,ig))
					call Add_source2D_pressurex_op(.false.,.true.,pcur_ux,recfield_P_fine(it+it2,ig))
				enddo
			endif
			
			!Applying absorbing boundary condition
			!call absorbing_op(pnew_ux)
			!call absorbing_op(pnew_uz)
			
			
			!Swapping pointers
			pold_ux => pcur_ux 
			pcur_ux => pnew_ux
			pnew_ux => pold_ux
			pold_uz => pcur_uz
			pcur_uz => pnew_uz
			pnew_uz => pold_uz
		
		enddo
		
		!saving source wavefield
		if (wantwavefield) then
			wc=srite("wavefield",pux_l,nz*nx*ny*4)
			wc=srite("wavefield",puz_l,nz*nx*ny*4)
			!Swap source wavefield left and right pointers
			ptemp => pux_l
			pux_l => pux_r
			pux_r => ptemp
			pux_r = 0.
		
			ptemp => puz_l
			puz_l => puz_r
			puz_r => ptemp
			puz_r = 0.
		
		endif
		
		if (mod(itw,(nw/10))==0) then
        	if(verbose)write(0,'(I3,A4)',advance="no")nint(itw*100./nw),"%..."
        	close(0)
      	endif
      	
		
	enddo
	
		!Last step of the source wavefield just saving source wavefield
		if (wantwavefield) then
			pux_l(:,:,:) = pux_l(:,:,:) + pcur_ux*interp_filter(1,1)
			puz_l(:,:,:) = puz_l(:,:,:) + pcur_uz*interp_filter(1,1)
			wc=srite("wavefield",pux_l,nz*nx*ny*4)
			wc=srite("wavefield",puz_l,nz*nx*ny*4)
		endif
		
		if(verbose)write(0,'(A1,I3,A1)',advance="no")" ",100,"%"
        close(0)
	if(verbose)write(0,'(a)',advance="yes")""
	
	
	call system_clock(count2)
	if(verbose)write(0,*) "Propagation time: ",(count2-count1)/cr
	if(verbose)write(0,'(A44,I5)') " Writing receiver data for source n°: ", is
	
	!Resample receivers' wavefield
	if (dw .le. dt) then
		do ig = 1,ng
			wc=srite("out",recfield_uz_fine(:,ig),nw*4)
			wc=srite("out",recfield_ux_fine(:,ig),nw*4)
			if (rec_p) wc=srite("out",recfield_P_fine(:,ig),nw*4)
		enddo
	else
		do ig = 1,ng
			call resample_op(.true.,.false.,recfield,recfield_uz_fine(:,ig))
			wc=srite("out",recfield,nw*4)
			call resample_op(.true.,.false.,recfield,recfield_ux_fine(:,ig))
			wc=srite("out",recfield,nw*4)
			if (rec_p) then
				call resample_op(.true.,.false.,recfield,recfield_P_fine(:,ig))
				wc=srite("out",recfield,nw*4)
			endif
		enddo
	endif
   	if(verbose)write(0,*) ""
  
  
  
  enddo
  
  call Add_source2D_pressure_close()
  call Prop_elastic_step_disp_close()
  call Sou_scale_acoustic_close()
  !call absorbing_close()
  
  !Writing information to headers
  call to_history("n1",nw)
  call to_history("o1",ow)
  call to_history("d1",dw)
  call to_history("label1","Time [s]")
  if (rec_p) then
  	call to_history("n2",3)
  	call to_history("o2",1.)
  	call to_history("d2",1.)
  	call to_history("label2","Component [Uz,Ux,P]")
  else
  	call to_history("n2",2)
  	call to_history("o2",1.)
  	call to_history("d2",1.)
  	call to_history("label2","Component [Uz,Ux]")
  endif
  call to_history("n3",ng)
  call to_history("o3",1.)
  call to_history("d3",1.)
  call to_history("label3","Receiver number")
  call to_history("n4",ns)
  call to_history("o4",1.)
  call to_history("d4",1.)
  call to_history("label4","Source number")
  if (wantwavefield) then
  	if(twoD) then
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
  		call to_history("n3",2,"wavefield")
  		call to_history("o3",1.,"wavefield")
  		call to_history("d3",1.,"wavefield")
  		call to_history("label3","Component [Ux,Uz]","wavefield")	
  		if (desample .eq. 1) then
			call to_history("n4",nt,"wavefield")
		else
			call to_history("n4",nt/desample+1,"wavefield")
		endif
  		call to_history("o4",ow,"wavefield")
 	 	call to_history("d4",dw,"wavefield")
  		call to_history("label4","Time [s]","wavefield")
  		call to_history("n5",ns,"wavefield")
  		call to_history("o5",1.,"wavefield")
  		call to_history("d5",1.,"wavefield")
  		call to_history("label5","Source number","wavefield")  	
  	else
		call to_history("n1",nz,"wavefield")
  		call to_history("o1",oz,"wavefield")
  		call to_history("d1",dz,"wavefield")
  		call to_history("label1","z [m]","wavefield")
		call to_history("n2",nx,"wavefield")
	  	call to_history("o2",ox,"wavefield")
  		call to_history("d2",dx,"wavefield")
  		call to_history("label2","x [m]","wavefield")
		call to_history("n3",ny,"wavefield")
	  	call to_history("o3",oy,"wavefield")
	  	call to_history("d3",dy,"wavefield")
  		call to_history("label3","y [m]","wavefield")
  		call to_history("n4",3,"wavefield")
  		call to_history("o4",1.,"wavefield")
  		call to_history("d4",1.,"wavefield")
  		call to_history("label4","Component [Ux,Uy,Uz]","wavefield")
  		if (desample .eq. 1) then
			call to_history("n5",nt,"wavefield")
		else
			call to_history("n5",nt/desample+1,"wavefield")
		endif
  		call to_history("o5",ow,"wavefield")
 	 	call to_history("d5",dw,"wavefield")
  		call to_history("label5","Time [s]","wavefield")
  		call to_history("n6",ns,"wavefield")
  		call to_history("o6",1.,"wavefield")
  		call to_history("d6",1.,"wavefield")
  		call to_history("label6","Source number","wavefield")
  	endif
  endif
  if(verbose)write(0,*) "End propagation"
  

  !Deallocating memory
  if (wantwavefield) deallocate(puz_l,puz_r,pux_l,pux_r)
  deallocate(cur_uz,old_uz,cur_ux,old_ux)
  deallocate(interp_filter)
  deallocate(recfield_uz_fine,recfield_ux_fine,recfield_P_fine,recfield)
  deallocate(scaled_fz_wavelet,scaled_fx_wavelet)
  deallocate(wavelet_fz_fine,wavelet_fx_fine)
  deallocate(rho_pad,mu_pad,lame_pad)
  deallocate(sx,sy,sz,gx,gy,gz)
  deallocate(wavelet_fz,wavelet_fx)
  call sep_close()
  call exit (0)


end program Propagator_elastic_disp_fw
