!!$=head1 NAME
!!$
!!$Fortran Solver with wrapper - Linear/Non-linear solvers with fortran wrapper to run shell commands
!!$
!!$=head1 SYNOPSIS
!!$
!!$Solver_run.x initial_model=initial_model.H data=data.H forward_file=Par/fw_commands.p adjoint_file=Par/adj_commands.p input_file_fw=input_model_fw.H output_file_fw=output_data_fw.H input_file_adj=input_data_adj.H output_file_adj=output_model_adj.H nmodel= n_model= d_model= o_model= ndata= n_data= d_data= o_data= niter= solver=CG model_iter=model_iter.H residual_iter=residual_iter.H cost_iter=cost_iter.H >inverted_model.H
!!$
!!$=head1 DESCRIPTION
!!$
!!$Solver for linear/non-linear least-squares problems
!!$
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 3
!!$
!!$=item from history 
!!$
!!$B<filename> I<none> - not necessary (it breaks the fortran wrapper)
!!$
!!$=item from par
!!$
!!$B<filename> I<initial_model> - starting model
!!$
!!$B<filename> I<data> - data to be inverted
!!$
!!$B<filename> I<forward_file> - parameter file containing the commands to run forward operator (structure = first line: #commands,other lines: commands)
!!$
!!$B<filename> I<adjoint_file> - parameter file containing the commands to run adjoint operator (structure = first line: #commands,other lines: commands)
!!$
!!$B<filename> I<input_file_fw> - name of the input file for running the forward operator
!!$
!!$B<filename> I<output_file_fw> - name of the output file after running the forward operator (!NOTE: put out=output_file_fw.H@ in the last command to run the wrapper correctly)
!!$
!!$B<filename> I<input_file_adj> - name of the input file for running the adjoint operator
!!$
!!$B<filename> I<output_file_adj> - name of the output file after running the adjoint operator (!NOTE: put out=output_file_adj.H@ in the last command to run the wrapper correctly)
!!$
!!$B<integer> I<nmodel> - number of axes of the model
!!$
!!$B<integer/real> I<n_model,d_model,o_model> - usual sep axis declarations, to be consistent with nmodel (e.g.,nmodel=1 n1model= d1model= o1model=)
!!$
!!$B<integer> I<ndata> - number of axes of the data
!!$
!!$B<integer/real> I<n_data,d_data,o_data> - usual sep axis declarations, to be consistent with ndata 
!!$
!!$B<integer> I<niter> - number of iterations to run
!!$
!!$B<integer> I<solver> - solver to be run (avail: 1 = CG (default), 2 = CG-mkl, 3 = Dot-Prod)
!!$
!!$B<filename> I<model_iter> - inverted model at each iteration
!!$
!!$B<filename> I<residual_iter> - residual vector at each iteration
!!$
!!$B<filename> I<cost_iter> - cost function at each iteration
!!$
!!$
!!$
!!$=back
!!$
!!$=head1 USED MODULES 
!!$
!!$sep, command_run_mod, solver_mod, cgstep_mod, CG_mkl_mod, dot_mod
!!$
!!$=head1 SEE ALSO
!!$
!!$None
!!$
!!$=cut
!%
program Solver_run

	use sep
	use command_run_mod
	use solver_mod
	use cgstep_mod
	use CG_mkl_mod
	use dot_mod

	implicit none
	character*2048						:: input_file_fw=" ", output_file_fw=" ", input_file_adj=" ", output_file_adj=" "
	character*1024						:: forward_file=" ",adjoint_file=" ",model_iter=" ",residual_iter=" ",cost_iter=" "
	character*8, parameter,dimension(7)	:: n_name_model = (/'n1model','n2model','n3model','n4model','n5model','n6model','n7model'/)
	character*8, parameter,dimension(7)	:: d_name_model = (/'d1model','d2model','d3model','d4model','d5model','d6model','d7model'/)
	character*8, parameter,dimension(7)	:: o_name_model = (/'o1model','o2model','o3model','o4model','o5model','o6model','o7model'/)
	character*8, parameter,dimension(7)	:: n_name_data = (/'n1data','n2data','n3data','n4data','n5data','n6data','n7data'/)
	character*8, parameter,dimension(7)	:: d_name_data = (/'d1data','d2data','d3data','d4data','d5data','d6data','d7data'/)
	character*8, parameter,dimension(7)	:: o_name_data = (/'o1data','o2data','o3data','o4data','o5data','o6data','o7data'/)
	character*8, parameter,dimension(7) :: n_name = (/'n1','n2','n3','n4','n5','n6','n7'/)
	character*8, parameter,dimension(7) :: d_name = (/'d1','d2','d3','d4','d5','d6','d7'/)
	character*8, parameter,dimension(7) :: o_name = (/'o1','o2','o3','o4','o5','o6','o7'/)
	integer								:: solver=1
	integer								:: nmodel, ndata
	integer								:: n_model(7)=0, n_data(7)=0, idim
	real								:: d_model(7)=0.0, o_model(7)=0.0, d_data(7)=0.0, o_data(7)=0.0
	!Internal variables
	real, allocatable					:: mm(:), dd(:), mm0(:), resd(:)
	integer						 		:: wc, srite, rc, sreed, stat, auxinout
	integer								:: niter
	
	
	call sep_init(SOURCE)
	
	
	!Getting variables from parameter files
	call from_param("forward_file",forward_file)
	call from_param("model_iter",model_iter,"model_iter.H")
	call from_param("residual_iter",residual_iter,"residual_iter.H")
	call from_param("cost_iter",cost_iter,"cost_iter.H")
	call from_param("adjoint_file",adjoint_file)
	call from_param("input_file_fw",input_file_fw)
	call from_param("output_file_fw",output_file_fw)
	call from_param("input_file_adj",input_file_adj)
	call from_param("output_file_adj",output_file_adj)
	call from_param("solver",solver,1)
	call from_param("niter",niter)
	!Getting axis information
	call from_param("nmodel",nmodel)
	do idim = 1, nmodel
		call from_param(n_name_model(idim),n_model(idim))
		call from_param(d_name_model(idim),d_model(idim),1.0)
		call from_param(o_name_model(idim),o_model(idim),0.0)
	enddo
	call from_param("ndata",ndata)
	do idim = 1, ndata
		call from_param(n_name_data(idim),n_data(idim))
		call from_param(d_name_data(idim),d_data(idim),1.0)
		call from_param(o_name_data(idim),o_data(idim),0.0)
	enddo
	
	!Call the init routine before reading model or data in the main otherwise the input/output file are only read or write only type
	call command_run_init(forward_file,adjoint_file, input_file_fw, output_file_fw, input_file_adj, output_file_adj, nmodel, ndata, n_model, d_model, o_model, n_data, d_data, o_data)
	
	
	!reading the model
	allocate(mm(product(n_model(1:nmodel))),mm0(product(n_model(1:nmodel))))
	rc = sreed("initial_model",mm0,4*product(n_model(1:nmodel)))
	!data part
	allocate(dd(product(n_data(1:ndata))),resd(product(n_data(1:ndata))))
	rc = sreed("data",dd,4*product(n_data(1:ndata)))

	select case(solver)
		case(1)
			write(0,*)"Start running CG solver"
			call solver_tiny_init(nmodel,n_model,model_iter,cost_iter)
			call solver_tiny( mm, dd, command_run_fun_call, cgstep, niter, mm0,resd)
			write(0,*)"Solver has finished"		
		case(2)
			write(0,*)"Start running CG-mkl solver"
			mm = mm0
			call cg_mkl_init(nmodel,n_model, model_iter,residual_iter,cost_iter)
			call cg_mkl(command_run_fun_call,mm,dd,niter)
			write(0,*)"Solver has finished"		
		case(3)
			write(0,*)"Start running dot-product test"
			call dotp(command_run_call,size(mm),size(dd))
		case default
       	 	write(0,*)'Not a solver type.'
       		call exit(0)
	end select

	
	wc=srite("out",mm,4*product(n_model(1:nmodel)))
	do idim = 1,nmodel
		call to_history(n_name(idim),n_model(idim))
		call to_history(d_name(idim),d_model(idim))
		call to_history(o_name(idim),o_model(idim))
	enddo
	
	call command_run_close()
	deallocate(resd,dd,mm0,mm)

	call sep_close()
	call exit(0)

end program Solver_run
