!Module for the scaling of the source (adjoint=forward)
!Model dimension (z,x,y,t)
!Data dimension (z,x,y,t)
!D(z,x,y,t) = Source_scale(Model(z,x,y,t))
!Model(z,x,y,t) = Source_scale*(D(z,x,y,t))

module Source_scale

  use weight

  implicit none
  !Module variables
  real, private										:: x, y, z, x0, y0, z0, dx, dy, dz, dt
  integer, private									:: nz, nx, ny, nt
  logical, private									:: twoD
  real, dimension(:,:,:), allocatable, private		:: tmp
  real												:: dt2

  contains
	
	!Initialization routine
	subroutine Source_scale_init(nz_in, nx_in, ny_in, nt_in, z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in, dt_in,twoD_in)
		
		!Input variables	
		real, intent(in)				:: z_in, x_in, y_in, z0_in, x0_in, y0_in, dz_in, dx_in, dy_in, dt_in
		integer, intent(in)				:: nz_in, nx_in, ny_in, nt_in
		logical, intent(in)				:: twoD_in

		!Assigning the internal variables
		x = x_in
		y = y_in
		z = z_in
		x0 = x0_in
		y0 = y0_in
		z0 = z0_in
		dx = dx_in
		dy = dy_in
		dz = dz_in
		nx = nx_in
		ny = ny_in
		nz = nz_in
		nt = nt_in
		dt = dt_in
		twoD = twoD_in
		dt2= dt*dt

		if(.not. allocated(tmp)) allocate(tmp(nz,nx,ny))
		tmp = 0.

		return
	end subroutine Source_scale_init

	!Operator routine
	subroutine Source_scale_op(adj,add,mm,vel)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny,nt), intent(inout)		:: mm !Padded source (z,x,y,t)
		real, dimension(:,:,:), intent(in)				:: vel
		!Internal variables
		integer											:: it, zi, xi, yi


		xi = 1. + (x-x0)/dx
		yi = 1. + (y-y0)/dy
		zi = 1. + (z-z0)/dz
		
		if (twoD) then
			mm(zi,xi,1,:)       = -mm(zi,xi,1,:)*dt2*vel(zi,xi,1)*vel(zi,xi,1)       
			mm(zi,xi+1,1,:)     = -mm(zi,xi+1,1,:)*dt2*vel(zi,xi+1,1)*vel(zi,xi+1,1)   
			mm(zi+1,xi,1,:)     = -mm(zi+1,xi,1,:)*dt2*vel(zi+1,xi,1)*vel(zi+1,xi,1)     
			mm(zi+1,xi+1,1,:)   = -mm(zi+1,xi+1,1,:)*dt2*vel(zi+1,xi+1,1)*vel(zi+1,xi+1,1)   
		
		else
			mm(zi,xi,yi,:)       = -mm(zi,xi,yi,:)*dt2*vel(zi,xi,yi)*vel(zi,xi,yi)       
			mm(zi,xi+1,yi,:)     = -mm(zi,xi+1,yi,:)*dt2*vel(zi,xi+1,yi)*vel(zi,xi+1,yi)
			mm(zi,xi,yi+1,:)     = -mm(zi,xi,yi+1,:)*dt2*vel(zi,xi,yi+1)*vel(zi,xi,yi+1)    
			mm(zi,xi+1,yi+1,:)   = -mm(zi,xi+1,yi+1,:)*dt2*vel(zi,xi+1,yi+1)*vel(zi,xi+1,yi+1)   
			mm(zi+1,xi,yi,:)     = -mm(zi+1,xi,yi,:)*dt2*vel(zi+1,xi,yi)*vel(zi+1,xi,yi)     
			mm(zi+1,xi+1,yi,:)   = -mm(zi+1,xi+1,yi,:)*dt2*vel(zi+1,xi+1,yi)*vel(zi+1,xi+1,yi)   
			mm(zi+1,xi,yi+1,:)   = -mm(zi+1,xi,yi+1,:)*dt2*vel(zi+1,xi,yi+1)*vel(zi+1,xi,yi+1)  
			mm(zi+1,xi+1,yi+1,:) = -mm(zi+1,xi+1,yi+1,:)*dt2*vel(zi+1,xi+1,yi+1)*vel(zi+1,xi+1,yi+1)	
		endif

		if(adj) then
			do it = 1,nt
				call weight_op(adj,add,tmp(:,:,:),mm(:,:,:,it))
				mm(:,:,:,it)=tmp-mm(:,:,:,it)
			enddo
		else
			do it = 1,nt
				call weight_op(adj,add,mm(:,:,:,it),tmp(:,:,:))
				mm(:,:,:,it)=tmp-mm(:,:,:,it)
			enddo
		endif
		
		return
	end subroutine Source_scale_op
	

	!Subroutine for deallocating variables
	subroutine Source_scale_close()
		if (allocated(tmp)) deallocate(tmp)
	end subroutine Source_scale_close

end module Source_scale
