!Module for computing the weight operator (see Ali's note) w (weight in the boundary region)
!Model dimension (z,x,y) Input wavefield
!Data dimension (z,x,y) Ouput wavefield
!D(z,x,y) = weight(Model(z,x,y))
!Model(z,x,y) = weight*(D(z,x,y))
!NOTE ADJOINT = FORWARD
module absorbing

  implicit none
  !Module variables
  integer, private																:: nz, nx, ny, npad
  real, dimension(:,:,:), allocatable, private		:: wz, wx, wy
  logical, private																:: twoD,free=.false.
  real, private																		:: alpha=0.05

  contains


	!Initialization routine
	subroutine absorbing_init(nz_in, nx_in, ny_in, nv_pad_in, pad_offset_in,twoD_in,free_in)

		!Input variables
		integer, intent(in)										:: nz_in, nx_in, ny_in, nv_pad_in, pad_offset_in
		real																	:: dz_in, dx_in, dy_in, argexp
		logical, intent(in)										:: twoD_in
		logical, intent(in) ,optional					:: free_in
		!Internal variables
		integer																:: ii


		!Assigning the internal variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		npad = nv_pad_in - pad_offset_in
		twoD = twoD_in
		!Checking for free surface
		if(present(free_in)) free = free_in
		if (free) then
			if (.not. allocated(wz)) allocate(wz(npad,nx,ny))
	   	if (.not. allocated(wx)) allocate(wx(nz,npad*2,ny))
	  	wz  = 0.; wx = 0.
			!Z-axis weight	and X-axis weight
			!$OMP PARALLEL DO private(ii)
			do ii=1,npad
				argexp = dfloat(npad-ii+1)/dfloat(npad)
	  		wz(npad-ii+1,:,:)		=  exp(-alpha*argexp*argexp)
	  		wx(:,ii,:)					=  exp(-alpha*argexp*argexp)
	  		wx(:,2*npad-ii+1,:)	=  exp(-alpha*argexp*argexp)
			enddo
			!$OMP END PARALLEL DO
		else
	   	if (.not. allocated(wz)) allocate(wz(npad*2,nx,ny))
	   	if (.not. allocated(wx)) allocate(wx(nz,npad*2,ny))
	  	wz  = 0.; wx = 0.

			!Z-axis weight	and X-axis weight
			!$OMP PARALLEL DO private(ii)
			do ii=1,npad
				argexp = dfloat(npad-ii+1)/dfloat(npad)
	      wz(ii,:,:)					=  exp(-alpha*argexp*argexp)
	  		wz(2*npad-ii+1,:,:)	=  exp(-alpha*argexp*argexp)
	  		wx(:,ii,:)					=  exp(-alpha*argexp*argexp)
	  		wx(:,2*npad-ii+1,:)	=  exp(-alpha*argexp*argexp)
			enddo
			!$OMP END PARALLEL DO
		endif


		if(.not. twoD) then
			if (.not. allocated(wy)) allocate(wy(nz,nx,npad*2))
			wy = 0.
			!Y-axis weight
			!$OMP PARALLEL DO private(ii)
			do ii=1,npad
				argexp = dfloat(npad-ii+1)/dfloat(npad)
		    wy(:,:,ii)						= exp(-alpha*argexp*argexp)
   		 	wy(:,:,2*npad-ii+1)		= exp(-alpha*argexp*argexp)
			enddo
			!$OMP END PARALLEL DO
		endif

		return
	end subroutine absorbing_init



	!Operator routine
	subroutine absorbing_op(mm) !Forward = Adjoint (diagonal operator)

		!Input-Output variables
		real, dimension(nz,nx,ny)			:: mm
		!Internal variables

		!Z-axis weighting
		if (free) then
			mm(nz-npad+1:nz,:,:) 	= mm(nz-npad+1:nz,:,:)*wz(1:npad,:,:)
		else
	    mm(1:npad,:,:) 				= mm(1:npad,:,:)*wz(1:npad,:,:)
			mm(nz-npad+1:nz,:,:) 	= mm(nz-npad+1:nz,:,:)*wz(npad+1:2*npad,:,:)
		endif
		!X-axis weighting
    mm(:,1:npad,:) 				= mm(:,1:npad,:)*wx(:,1:npad,:)
		mm(:,nx-npad+1:nx,:) 	= mm(:,nx-npad+1:nx,:)*wx(:,npad+1:2*npad,:)
		if(.not. twoD) then
			!Y-axis weighting
    	mm(:,:,1:npad) 				= mm(:,:,1:npad)*wy(:,:,1:npad)
			mm(:,:,ny-npad+1:ny) 	= mm(:,:,ny-npad+1:ny)*wy(:,:,npad+1:2*npad)
		endif

		return
	end subroutine absorbing_op


	!Subroutine for deallocating variables
	subroutine absorbing_close()
		if (allocated(wy)) deallocate(wy)
		if (allocated(wx)) deallocate(wx)
		if (allocated(wz)) deallocate(wz)
	end subroutine absorbing_close


end module absorbing
