!Program to compute particle displacements and element strain tensor caused by a pressure field
!Note that the displacement are shifted by half grid space in the increasing direction on each axis
!(i.e., ux(i+1/2,j,k), uy(i,j+1/2,k), uz(i,j,k+1/2))
!Note the element strain is shift by half grid in two direction for the shear
!for example: Ezx(i+1/2,i,k+1/2)
!Normal Strain are on regular grid
!for example: Exx(i,j,k)

program displace_strain

	!used modules
	use sep
	use D10pt_mod
	use S_mod
	use Pad_pert
	
	!variables
	implicit none
	real												:: dz, dx, dy, oz, ox, oy, ot, dt
	integer												:: nz, nx, ny, nt, ns, nzrho, nxrho, nyrho
	integer												:: is, it
	logical												:: twoD = .true.
	real, dimension(:,:,:,:), allocatable				:: pressure
	real, dimension(:,:,:), allocatable,target			:: rho
	real, dimension(:,:,:), allocatable					:: epsi, u_vec, rhoz, rhox, rhoy, disp, vel
	real, dimension(:,:,:), pointer						:: rho_pad, tmp
	integer												:: wc, srite, rc, sreed, sk, sseek
	integer*8											:: rbit !necessary for reading big files
	
	
	
	!initialization SEP
	call sep_init (SOURCE)
	
	
	write(0,*)""
  	write(0,*)""
  	write(0,*)"##############################################################################"
  	write(0,*)"####################    Particle velocity computation    #####################"
  	write(0,*)"####################            & element strain         #####################" 
  	write(0,*)"####################          from pressure waves        #####################" 
  	write(0,*)"##############################################################################"
  	write(0,*)""
  	write(0,*)""
	!Reading information from input files
	call from_aux("rho","n3",nyrho)
	if (nyrho .gt. 1) twoD = .false.
	if (twoD) then
  		call from_history("n1", nz)
  		call from_history("o1", oz);
  		call from_history("d1", dz);
  		call from_history("n2", nx)
  		call from_history("o2", ox);
  		call from_history("d2", dx);
  		call from_history("n3", nt)
  		call from_history("o3", ot);
  		call from_history("d3", dt);
  		call from_history("n4", ns)
  		ny = 1
  		oy = ox
  		dy = dx
  	else
  		call from_history("n1", nz)
  		call from_history("o1", oz);
  		call from_history("d1", dz);
  		call from_history("n2", nx)
  		call from_history("o2", ox);
  		call from_history("d2", dx);
  		call from_history("n3", ny)
  		call from_history("o3", oy);
  		call from_history("d3", dy);
  		call from_history("n4", nt)
  		call from_history("o4", ot);
  		call from_history("d4", dt);
  		call from_history("n5", ns)
  	endif
	
	!allocating memory for the program
	allocate(pressure(nz,nx,ny,nt),disp(nz,nx,ny),vel(nz,nx,ny),u_vec(nz,nx,ny),epsi(nz,nx,ny))
	pressure = 0.0
	epsi = 0.0
	u_vec = 0.0
	disp = 0.0
	!Check if rho needs to be padded
	call from_aux("rho","n1",nzrho)
	call from_aux("rho","n2",nxrho)
	if (nzrho .lt. nz) then
		allocate(rho(nzrho,nxrho,nyrho)); rho = 0.
		rc = sreed("rho",rho,nzrho*nxrho*nyrho*4)
		call Pad_perturbation(rho,rho_pad,nz-nzrho)
		allocate(rho(nz,nx,ny)); rho = rho_pad
		deallocate(rho_pad)
	elseif (nzrho .gt. nz) then
		write(0,*)"Code error: mismatching dimension (nzrho=",nzrho,"> nz=", nz,")"
		stop
	else
		allocate(rho(nz,nx,ny)); rho=0.0
		rc = sreed("rho",rho,nz*nx*ny*4)
	endif
	rho = 1.0/rho
	tmp => rho
	!initialization routines
	call D10pt_init(nz,nx,ny,dz,dx,dy,twoD)
	call S_init(nz,nx,ny,twoD)
	
	!Staggering density perturbation
  	allocate(rhoz(nz,nx,ny), rhox(nz,nx,ny))
  	call Sz_op(.true., .false., rhoz, rho)
  	call Sx_op(.true., .false., rhox, rho)
  	!Staggering on y if not twoD problem
  	if(.not. twoD) then
  		allocate(rhoy(nz,nx,ny))
  		call Sy_op(.true., .false., rhoy, rho)
  	endif
	
	rbit = nz*nx*ny
	do is=1,ns
		write(0,'(A62,I5)') "    Computing displacements and strain tensor for source n°: ", is
		rc=sreed("in" , pressure , rbit*nt*4)
		do it=1,nt
			!###########################################################################
			!double time integration u = \int (\int grad P dt) dt 
			!Using Leibniz disp rule we can interchange integration with differentiation 
			if (it .eq. 1) then
				disp = 0.0
				vel = 0.0
			else
				disp(:,:,:) = disp(:,:,:) + vel(:,:,:)*dt + 0.25*(pressure(:,:,:,it-1)+pressure(:,:,:,it))*dt*dt
				vel(:,:,:) = vel(:,:,:) + 0.5*(pressure(:,:,:,it-1)+pressure(:,:,:,it))*dt
			endif
			!###########################################################################
			!Uz particle displacement field
			call Dmz_op(.false.,.false.,disp(:,:,:),u_vec)
			call Dmz_op(.false.,.false.,vel(:,:,:),tmp)
			!Density scaling
			u_vec = rhoz*u_vec;
			tmp = rhoz*tmp;
			!writing Vz on file
			wc=srite("out",tmp,nz*nx*ny*4)
			!Computation of the normal strain zz
			call Dpz_op(.false.,.false.,u_vec,epsi)
			!writing Ezz on file
			wc=srite("strain",epsi,nz*nx*ny*4)
			!Computation of the shear strain zx
			call Dmx_op(.false.,.false.,u_vec,epsi)
			!writing Ezx on file
			wc=srite("strain",epsi,nz*nx*ny*4)
			if (.not. twoD) then
				!Computation of the shear strain zy
				call Dpy_op(.false.,.false.,u_vec,epsi)
				!writing Ezy on file
				wc=srite("strain",epsi,nz*nx*ny*4)
			endif
			!###########################################################################
			!Ux particle displacement field
			call Dmx_op(.false.,.false.,disp(:,:,:),u_vec)
			call Dmx_op(.false.,.false.,vel(:,:,:),tmp)
			!Density scaling
			u_vec = rhox*u_vec;
			tmp = rhox*tmp;
			!writing Vx on file
			wc=srite("out",tmp,nz*nx*ny*4)
			!Computation of the normal strain xx
			call Dpx_op(.false.,.false.,u_vec,epsi)
			!writing Exx on file
			wc=srite("strain",epsi,nz*nx*ny*4)
			!Computation of the shear strain zx
			call Dpx_op(.false.,.false.,u_vec,epsi)
			if (.not. twoD) then
				!Computation of the shear strain xy
				call Dpy_op(.false.,.false.,u_vec,epsi)
				!writing Exy on file
				wc=srite("strain",epsi,nz*nx*ny*4)
			endif
			!###########################################################################
			if (.not. twoD) then
				call Dpy_op(.false.,.false.,disp(:,:,:),u_vec)
				call Dpy_op(.false.,.false.,vel(:,:,:),tmp)
				!Density scaling
				u_vec = rhoy*u_vec;
				tmp = rhoy*tmp;
				!writing Uy on file
				wc=srite("out",tmp,nz*nx*ny*4)
				!Computation of the normal strain yy
				call Dmy_op(.false.,.false.,u_vec,epsi)
				!writing Eyy on file
				wc=srite("strain",epsi,nz*nx*ny*4)
			endif
		enddo	
	enddo
	
	
	!Writing header of output
  	if (twoD) then
  		call to_history("n1",nz)
  		call to_history("o1",oz)
  		call to_history("d1",dz)
  		call to_history("label1","z [m]")
		call to_history("n2",nx)
  		call to_history("o2",ox)
 		call to_history("d2",dx)
  		call to_history("label2",'x [m]')
  		call to_history("n4",nt)
  		call to_history("o4",ot)
 		call to_history("d4",dt)
  		call to_history("label4","Time [s]")
  		call to_history("n3",2)
  		call to_history("o3",1.)
  		call to_history("d3",1.)
  		call to_history("label3","Component (Uz,Ux)")
  		call to_history("n5",ns)
  		call to_history("o5",1.)
  		call to_history("d5",1.)
  		call to_history("label5","Source number")
  	else
  		call to_history("n1",nz)
  		call to_history("o1",oz)
  		call to_history("d1",dz)
  		call to_history("label1","z [m]")
		call to_history("n2",nx)
  		call to_history("o2",ox)
 		call to_history("d2",dx)
  		call to_history("label2",'x [m]')
  		call to_history("n3",ny)
		call to_history("o3",oy)
		call to_history("d3",dy)
  		call to_history("label3","y [m]")
  		call to_history("n5",nt)
  		call to_history("o5",ot)
 		call to_history("d5",dt)
  		call to_history("label5","Time [s]")
  		call to_history("n4",3)
  		call to_history("o4",1.)
  		call to_history("d4",1.)
  		call to_history("label4","Component (Uz,Ux,Uy)")
  		call to_history("n6",ns)
  		call to_history("o6",1.)
  		call to_history("d6",1.)
  		call to_history("label6","Source number")
  	endif
  	
  	!Writing header of strain
  	if (twoD) then
  		call to_history("n1",nz,"strain")
  		call to_history("o1",oz,"strain")
  		call to_history("d1",dz,"strain")
	  	call to_history("label1","z [m]","strain")
		call to_history("n2",nx,"strain")	
  		call to_history("o2",ox,"strain")
 		call to_history("d2",dx,"strain")
 		call to_history("label2",'x [m]',"strain")
 		call to_history("n4",nt,"strain")
  		call to_history("o4",ot,"strain")
 		call to_history("d4",dt,"strain")
  		call to_history("label4","Time [s]","strain")
  		call to_history("n3",3,"strain")
  		call to_history("o3",1.,"strain")
  		call to_history("d3",1.,"strain")
  		call to_history("label3","Component (Ezz,Ezx,Exx)","strain")
  		call to_history("n5",ns,"strain")
  		call to_history("o5",1.,"strain")
  		call to_history("d5",1.,"strain")
  		call to_history("label5","Source number","strain")
  	else
  		call to_history("n1",nz,"strain")
  		call to_history("o1",oz,"strain")
  		call to_history("d1",dz,"strain")
  		call to_history("label1","z [m]","strain")
		call to_history("n2",nx,"strain")
  		call to_history("o2",ox,"strain")
 		call to_history("d2",dx,"strain")
  		call to_history("label2",'x [m]',"strain")
		call to_history("n3",ny,"strain")
		call to_history("o3",oy,"strain")
		call to_history("d3",dy,"strain")	
	  	call to_history("label3","y [m]","strain")
		call to_history("n5",nt,"strain")
  		call to_history("o5",ot,"strain")
 		call to_history("d5",dt,"strain")
  		call to_history("label5","Time [s]","strain")
  		call to_history("n4",6,"strain")
  		call to_history("o4",1.,"strain")
  		call to_history("d4",1.,"strain")
  		call to_history("label4","Component (Ezz,Ezx,Ezy,Exx,Exy,Eyy)","strain")
  		call to_history("n6",ns,"strain")
  		call to_history("o6",1.,"strain")
  		call to_history("d6",1.,"strain")
  		call to_history("label6","Source number","strain")
  	endif
  	
	
	call D10pt_close()
	if(allocated(rhoy))deallocate(rhoy)
  	deallocate(rhox,rhoz)
	deallocate(rho,epsi,u_vec,vel,disp,pressure)
	call sep_close()
	call exit (0)
	
end program displace_strain