!Module to compute elastic gradient from pressure waves

module elastic_imaging

	use S_mod
	use sep
	!use D8pt_mod
	use D10pt_mod

	implicit none
	logical, private									:: twoD
	real, private										:: dt
	integer, private									:: nz, nx, ny, itw, nw
	!Arrays for elastic gradient computation
 	real, dimension(:,:,:), pointer, private			:: vel_sou, vel_rec, disp_rec
 	real, dimension(:,:,:), pointer, private			:: acce_temp
 	real, dimension(:,:,:), allocatable, private		:: rhoz, rhox, rhoy
 	real, dimension(:,:,:), pointer, private			:: dstrainzz_s, dstrainxx_s, dstrainyy_s
  	real, dimension(:,:,:), pointer, private			:: strainzz_r, strainxx_r, strainyy_r
  	real, dimension(:,:,:), pointer, private			:: dstrainxz_s, strainxz_r
  	real, dimension(:,:,:), pointer, private			:: ptemp, ptemp1
  	real, dimension(:,:,:), pointer, private			:: lame_pad, mu_pad
	
	
	
	contains
	
	
		!Initialization routines
		subroutine elastic_imaging_init1(nz_in,nx_in,ny_in,nw_in,dt_in,dz_in,dx_in,dy_in,rho_in,temp_array1,lame_pad_in,mu_pad_in,twoD_in)
			integer, intent(in)					:: nz_in,nx_in,ny_in,nw_in
			real, intent(in)					:: dt_in,dz_in,dx_in,dy_in
			logical, intent(in)					:: twoD_in
			real, dimension(:,:,:), pointer		:: rho_in, temp_array1
			real, dimension(:,:,:), pointer		:: lame_pad_in, mu_pad_in
			
			twoD = twoD_in
			dt = dt_in
			nz = nz_in
			nx = nx_in
			ny = ny_in
			nw = nw_in
		
			call S_init(nz_in,nx_in,ny_in,twoD)
			!call D8pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD)
			call D10pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD)
			
			!Allocating memory for imaging condition (particle velocity and displacements)
  			allocate(vel_sou(nz,nx,ny), vel_rec(nz,nx,ny), disp_rec(nz,nx,ny), acce_temp(nz,nx,ny))
			
			!Staggering density
  			allocate(rhoz(nz,nx,ny), rhox(nz,nx,ny))
  			call Sz_op(.true., .false., rhoz, 1.0/rho_in)
  			call Sx_op(.true., .false., rhox, 1.0/rho_in)
 			!Staggering on y if not twoD problem
  			if(.not. twoD) then
  				allocate(rhoy(nz,nx,ny))
  				call Sy_op(.true., .false., rhoy, 1.0/rho_in)
  			endif
  			!Allocating strain arrays
  			allocate(dstrainxx_s(nz,nx,ny),dstrainxz_s(nz,nx,ny))
  			allocate(strainzz_r(nz,nx,ny),strainxx_r(nz,nx,ny),strainxz_r(nz,nx,ny))
  			
  			!Assigning temporary array
  			ptemp1 => temp_array1
			dstrainzz_s => ptemp1
			lame_pad => lame_pad_in
			mu_pad => mu_pad_in
			
			!call to_history("n1",nz,"acoustic_image_strainzz.H")
    		!call to_history("n2",nx,"acoustic_image_strainzz.H")
    		!call to_history("n3",nw,"acoustic_image_strainzz.H")
    		!call to_history("d3",dt,"acoustic_image_strainzz.H")
		
		end subroutine elastic_imaging_init1
		
		subroutine elastic_imaging_init2(temp_array,itw_in)
			real, dimension(:,:,:), pointer		:: temp_array
			integer, intent(in)					:: itw_in
			
			ptemp => temp_array
			itw = itw_in
		
		end subroutine elastic_imaging_init2
		
		
		
		!Main routine
		subroutine elastic_imaging_op(source_wavefield,scat_sou,pertlame_pad,pertmu_pad,pertrho_pad)
			real, dimension(:,:,:,:),pointer	:: source_wavefield
			real, dimension(:,:,:), pointer		:: scat_sou,pertlame_pad,pertmu_pad,pertrho_pad
			integer		:: wc, srite
		
		
			!Integrating in time for particle velocity and displacements
			if (itw .eq. nw) then
				disp_rec = 0.0
				vel_rec = 0.0 
				vel_sou = 0.0
				acce_temp = scat_sou !To keep in memory the right step necessary for receiver wavefield integration
			else
				vel_sou = vel_sou + 0.5*(source_wavefield(:,:,:,itw)+source_wavefield(:,:,:,itw+1))*dt
				disp_rec = disp_rec + vel_rec*dt + 0.25*(scat_sou+acce_temp)*dt*dt
				vel_rec = vel_rec + 0.5*(scat_sou+acce_temp)*dt
				acce_temp = scat_sou !To keep in memory the right step necessary for receiver wavefield integration
			endif
			
			!Source acceleration-z
			call Dmz_op(.false.,.false., source_wavefield(:,:,:,itw), ptemp)
			!Rec velocity-z
			call Dmz_op(.false.,.false., vel_rec, ptemp1)
			!Imaging condition and returning to regular grid
			call Sz_op(.false., .true., ptemp*ptemp1, pertrho_pad) !dVz,s/dt * Vz,r
			!Source acceleration-x
			call Dmx_op(.false.,.false., source_wavefield(:,:,:,itw), ptemp)
			!Rec velocity-x
			call Dmx_op(.false.,.false., vel_rec, ptemp1)
			!Imaging condition and returning to regular grid
			call Sx_op(.false., .true., ptemp*ptemp1, pertrho_pad) !dVx,s/dt * Vx,r
			!y-axis scattering (3D ONLY) 
  			if(.not. twoD) then
  				!Source acceleration-y
				call Dmy_op(.false.,.false., source_wavefield(:,:,:,itw), ptemp)
				!Rec velocity-y
				call Dmy_op(.false.,.false., vel_rec, ptemp1)
				!Imaging condition and returning to regular grid
				call Sy_op(.false., .true., ptemp*ptemp1, pertrho_pad) !dVy,s/dt * Vy,r
  			endif
  			!Imaging condition for lames parameter and shear modulus (SO FAR 2D ONLY)
  			!Source vel-z
			call Dmz_op(.false.,.false., vel_sou, ptemp)
			ptemp = ptemp*rhoz !Scaling by density
			!Normal strain zz
			call Dpz_op(.false.,.false., ptemp, dstrainzz_s)
			!wc=srite("acoustic_image_strainzz.H",dstrainzz_s,nz*nx*4)
			!Shear strain xz
			call Dmx_op(.false.,.false., ptemp, dstrainxz_s)
			!Source vel-x
			call Dmx_op(.false.,.false., vel_sou, ptemp)
			ptemp = ptemp*rhox !Scaling by density
			!Normal strain xx
			call Dpx_op(.false.,.false., ptemp, dstrainxx_s)
			!Rec vel-z
			call Dmz_op(.false.,.false., disp_rec, ptemp)
			ptemp = ptemp*rhoz !Scaling by density
			!Normal strain zz
			call Dpz_op(.false.,.false., ptemp, strainzz_r)
			!Shear strain xz
			call Dmx_op(.false.,.false., ptemp, strainxz_r)
			!Source vel-x
			call Dmx_op(.false.,.false., disp_rec, ptemp)
			ptemp = ptemp*rhox !Scaling by density
			!Normal strain xx
			call Dpx_op(.false.,.false., ptemp, strainxx_r)
			!Lame imaging condition
			ptemp = (dstrainzz_s+dstrainxx_s)*(strainzz_r+strainxx_r)
			pertlame_pad = pertlame_pad + ptemp
			!Shear modulus imaging condition
			pertmu_pad = pertmu_pad + ptemp*lame_pad + 2.0*mu_pad*(dstrainzz_s*strainzz_r+dstrainxx_s*strainxx_r)
			!Taking the shear component of the image to the regular grid
			call Sx_op(.false., .false., dstrainxz_s*strainxz_r, ptemp)
			ptemp1 => dstrainzz_s
			call Sz_op(.false., .false., ptemp, ptemp1)
			pertmu_pad = pertmu_pad + 2.0*mu_pad*(ptemp1)
		
		end subroutine elastic_imaging_op
		
		
		
		!Closing routine
		subroutine elastic_imaging_close()
		
			!Deallocating memory
  			deallocate(strainxz_r,strainxx_r,strainzz_r)
			deallocate(dstrainxz_s,dstrainxx_s)
  			if(allocated(rhoy))deallocate(rhoy)
 			deallocate(rhox,rhoz)
  			deallocate(acce_temp,disp_rec,vel_rec,vel_sou)
  			call D10pt_close()
  			!call D8pt_close()
			
		end subroutine elastic_imaging_close





end module elastic_imaging