!Module for computing full elastic imaging condition
!NB. the perturbation are considered to be on the correct staggered grid when mulitplied
module full_elastic_imaging_mod
    
  use D10pt_mod
  
  implicit none
  !Module variables
  integer, private								:: nz, nx, ny
  real,allocatable,dimension(:,:,:), private	:: ptemp1, ptemp2
  
  
  
  contains
  
  	subroutine full_elastic_imaging_init(nz_in, nx_in, ny_in, dz_in, dx_in, dy_in, twoD_in)
  		!Input variables	
		integer, intent(in)						:: nz_in, nx_in, ny_in
		real									:: dt_in, dz_in, dx_in, dy_in
		logical, intent(in)						:: twoD_in
		
		
		nz = nz_in
		nx = nx_in
		ny = ny_in
		
		
		!Initializing other modules
		call D10pt_init(nz,nx,ny,dz_in,dx_in,dy_in,twoD_in)
		
		if(.not. allocated(ptemp1)) allocate(ptemp1(nz,nx,ny))
		if(.not. allocated(ptemp2)) allocate(ptemp2(nz,nx,ny))
  	
  	end subroutine full_elastic_imaging_init
  	
  	
  	subroutine full_elastic_imaging_op(adj,add,dvx_dt0,dvz_dt0,vx0,vz0,drhox,drhoz,dlambda,dmu,dmu_xz,dvx,dvz,dxx,dzz,dxz)
  		real, dimension(:,:,:), pointer					:: drhox, drhoz, dlambda, dmu, dmu_xz, dvx, dvz, dxx, dzz, dxz
  		real, dimension(nz,nx,ny), intent(inout)		:: dvx_dt0, dvz_dt0, vx0, vz0
  		logical											:: adj, add
  		
  		
  		if (adj) then !Adjoint (Imaging)
  			if (.not. add) then
  				drhox=0.; drhoz=0.; dlambda=0.; dmu=0.; dmu_xz=0.;
  			endif
  			!Rho imaging
  			drhox = drhox - dvx*dvx_dt0
  			drhoz = drhoz - dvz*dvz_dt0
  			!Lambda imaging
  			call Dpz_op(.false.,.false.,vz0,ptemp1)
  			call Dpx_op(.false.,.false.,vx0,ptemp2)
  			dlambda = dlambda + (ptemp1+ptemp2)*(dxx + dzz)
  			!Mu imaging
  			dmu = dmu + 2.0*(ptemp2*dxx+ptemp1*dzz)
  			call Dmx_op(.false.,.false.,vz0,ptemp1)
  			call Dmz_op(.false.,.false.,vx0,ptemp2)
  			dmu_xz = dmu_xz + dxz*(ptemp1+ptemp2)
  			
  			
  		else !Forward (Scattering)
  			if (.not. add) then
  				dvx=0.; dvz=0.; dxx=0.; dzz=0.; dxz=0.;
  			endif
  			!Vx scattering
  			dvx = dvx - drhox*dvx_dt0
  			!Vz scattering
  			dvz = dvz - drhoz*dvz_dt0
  			!Sigmaxx scattering
  			call Dpz_op(.false.,.false.,vz0,ptemp1)
  			call Dpx_op(.false.,.false.,vx0,ptemp2)
  			dxx = dxx + dlambda*(ptemp1+ptemp2) + 2.0*dmu*ptemp2
  			!Sigmazz scattering
  			dzz = dzz + dlambda*(ptemp1+ptemp2) + 2.0*dmu*ptemp1
  			!Sigmaxz scattering
  			call Dmx_op(.false.,.false.,vz0,ptemp1)
  			call Dmz_op(.false.,.false.,vx0,ptemp2)
  			dxz = dxz + dmu_xz*(ptemp1+ptemp2)
  			
  		endif
  	
  	end subroutine full_elastic_imaging_op
  	
  	!Operator for scattering condition of adjoint operator (term Dm L*(u1*)dm1 in 2nd-order term of Hessian) (Note that dxx, dzz, and dxz are zeros)
  	subroutine adjoint_elastic_scattering_op(add,dvx_dt_adj,dvz_dt_adj,xx_adj,zz_adj,xz_adj,drhox,drhoz,dlamdba,dmu,dmuxz, dvx, dvz)
  		real, dimension(:,:,:), pointer					:: drhox, drhoz, dlamdba, dmu, dmuxz !Perturbations on the corresponding grids
  		real, dimension(:,:,:), pointer					:: dvx, dvz
  		real, dimension(nz,nx,ny), intent(inout)		:: dvx_dt_adj, dvz_dt_adj, xx_adj, zz_adj, xz_adj
  		logical											:: add
  		
  		if (.not. add) then
  			dvx=0.; dvz=0.;
  		endif
  		
  		!scattered adjoint dvx
  		dvx = dvx - dvx_dt_adj*drhox
  		ptemp1=(xx_adj)*(dlamdba+2.0*dmu)
  		call Dpx_op(.true.,.true.,dvx,ptemp1)
  		ptemp1=(zz_adj)*(dlamdba)
  		call Dpx_op(.true.,.true.,dvx,ptemp1)
  		ptemp1=(xz_adj)*(dmuxz)
  		call Dmz_op(.true.,.true.,dvx,ptemp1)
  		
  		!scattered adjoint dvz
  		dvz = dvz - dvz_dt_adj*drhoz
  		ptemp1=(zz_adj)*(dlamdba+2.0*dmu)
  		call Dpz_op(.true.,.true.,dvz,ptemp1)
  		ptemp1=(xx_adj)*(dlamdba)
  		call Dpz_op(.true.,.true.,dvz,ptemp1)
  		ptemp1=(xz_adj)*(dmuxz)
  		call Dmx_op(.true.,.true.,dvz,ptemp1)
  	
  	end subroutine adjoint_elastic_scattering_op
  	
  	
  	subroutine full_elastic_imaging_close()
  		if(allocated(ptemp2))deallocate(ptemp2)
  		if(allocated(ptemp1))deallocate(ptemp1)
  		call D10pt_close()
  	end subroutine full_elastic_imaging_close


end module full_elastic_imaging_mod