#!/usr/bin/env python
#Simple self-adjoint Gaussian smoothing operator
import sys,os
sys.path.append(os.environ.get('REPOSITORY')+"/python_solver/python_modules")
import sep_util as sep
import numpy as np
import pyVector
import pyOperator

#Creating Gaussian smoothing operator
class Gauss_smooth(pyOperator.Operator):
	def __init__(self,model,sigmax,sigmaz):
		"""Overwritting default constructor"""
		assert(isinstance(sigmax,float))
		assert(isinstance(sigmaz,float))
		self.setDomainRange(model,model)
		self.sigmax=sigmax
		self.sigmaz=sigmaz
		#Precomputing Gaussian filters
		self.nz=model.naxis[0]
		self.nx=model.naxis[1]
		epsilon=1e-8 #Cut-off value for removing small filter coefficients
		#Filter on z axis
		arg = np.linspace(0.,self.nz)
		arg = -arg*arg/(sigmaz*sigmaz)
		filterz = np.exp(arg)
		#Removing extremely small values
		filterz = filterz[filterz>epsilon]
		#Symmetrizing filter
		self.filterz = np.append(np.flip(filterz[1:],0),filterz)
		#Filter on x axis
		arg = np.linspace(0.,self.nx)
		arg = -arg*arg/(sigmax*sigmax)
		filterx = np.exp(arg)
		#Removing extremely small values
		filterx = filterx[filterx>epsilon]
		#Symmetrizing filter
		self.filterx = np.append(np.flip(filterx[1:],0),filterx)
		#Allocating temporary array for dealing with the boundaries
		self.tmp_array=np.zeros((self.nx+self.filterx.size-1,self.nz+self.filterz.size-1))
		return
		
	def forward(self,add,model,data):
		"""Forward operator"""
		self.checkDomainRange(model,data)
		if(add): data_tmp=data.clone()
		data.zero()
		halfx=int(self.filterx.size/2)
		halfz=int(self.filterz.size/2)
		#Operator currently can only process vectorIC
		if(not (isinstance(model,pyVector.vectorIC) or isinstance(data,pyVector.vectorIC))): raise TypeError("ERROR! Model and data vectors must be vectorIC type!")
		self.tmp_array[halfx:halfx+self.nx,halfz:halfz+self.nz]=model.arr
		#Applying filter along z
		for ix in range(self.nx):
			for iz in range(self.nz):
				data.arr[ix,iz]=np.dot(self.tmp_array[ix+halfx,iz:iz+self.filterz.size],self.filterz)
		#Applying filter along x
		self.tmp_array[halfx:halfx+self.nx,halfz:halfz+self.nz]=data.arr
		for ix in range(self.nx):
			for iz in range(self.nz):
				data.arr[ix,iz]=np.dot(self.tmp_array[ix:ix+self.filterx.size,iz+halfz],self.filterx)
		if(add): data.scaleAdd(data_tmp)
		return

	def adjoint(self,add,model,data):
		"""Self-adjoint operator"""
		self.forward(add,data,model)
		return

if __name__ == '__main__':
    if(len(sys.argv) < 4):
        print("NAME \n")
        print("	Gaussian_smoothing - Simple self-adjoint Gaussian smoothing operator \n")
        print("SYNOPSIS")
        print("	Gaussian_smoothing.py input.H output.H sigmax sigmaz\n")
        print("DESCRIPTION")
        print("	Applies a 2D Gaussian smoothing to input.H and places the result in output.H using sigmax and sigmaz as standard deviation on each dimension.\n")
    else:
    	#reading command line for arguments
        input_file=sys.argv[1]
        output_file=sys.argv[2]
        sigmax=float(sys.argv[3])
        sigmaz=float(sys.argv[4])
        #Reading input file and creating necessary input and output vectors
        input_vec = pyVector.vectorIC(input_file)
        output_vec = input_vec.clone()
        output_vec.zero()
        #Constructing smoothing operator
        smoothing_op=Gauss_smooth(input_vec,sigmax,sigmaz)
        #Applying operator to input file
        smoothing_op.forward(False,input_vec,output_vec)
        #Writing result into output file
        output_vec.writeVec(output_file)
        









