!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! PROGRAM DESCRIPTION (only works for 2D):
!! 1. reads in an 3D file (z,x,n3) where
!! 		- i3=1 is the non-extended background model
!!		- i3=2:n3 is the extended perturbation model (n3-1 = nhz * nhx * nht)
!! 2. extracts the background and the non-extended slice of the perturbation
!! 3. sums the two slices
!! 4. applies a low-wavenumber pass filter
!! 5. spits out ONLY the low-wavenumber passed model (a non-extended 2D slice)
!! 
!! PROGRAM DESCRIPTION FOR NORMAL HUMANS:
!! This code is used in a TFWI outer loop to low-pass the output of a sequence of inner 
!! loops. It takes the output model of the inner loops, extract and sums the background 
!! model and the non-extended slice of the perturbation model, applies a low-pass 
!! wavenumber filter on this non-extended model and spits it out. 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
program scale_mixing_model

	use sep
    use fourier
	
	implicit none	
	
  	integer									  :: nz,nx,nz2,nx2,n3,n_left,n_ext,ne_index
  	integer									  :: iz,ix,fft_pad
	integer									  :: ikz,ikx,ntaper,index_okx,index_okz,buttord
	real								      :: oz,ox,dz,dx
	real									  :: dkz,dkx,okz,okx,dk_taper
	real									  :: k,kz,kx,k_taper,k_cut,k_big,k_small
	real									  :: sign1,sign2
  	real,dimension(:,:),allocatable  		  :: filter,model_bg,model_pert,model_sum,model_low,model_high
  	complex,dimension(:,:),allocatable		  :: model_sum_fft,model_sum_fft_filter,cmodel_low
	logical									  :: debug
	integer									  :: rc,sreed,sk,sseek_block
		
	call sep_init(SOURCE)

	! READ FROM HISTORY
	call from_history("n1",nz)
 	call from_history("d1",dz)	
 	call from_history("o1",oz)
 	
 	call from_history("n2",nx)
 	call from_history("d2",dx)	
 	call from_history("o2",ox) 
 	
	call from_history("n3",n3)

	call from_param("fft_pad",fft_pad,500)
	call from_param("debug",debug,.false.)

	call from_param("buttord",buttord,1)
	write(0,*) "Butterworth wavenumber low-pass filter"
	write(0,*) "filter order:",buttord		

	nz2 = nz + 2 * fft_pad
	nx2 = nx + 2 * fft_pad	
	n_ext = n3 - 1 		

	! ALLOCATE
	allocate(model_bg(nz2,nx2))
	allocate(model_pert(nz2,nx2))	
	allocate(model_sum(nz2,nx2))
	allocate(filter(nz2,nx2))
	allocate(model_sum_fft(nz2,nx2))
	allocate(model_sum_fft_filter(nz2,nx2))
	allocate(cmodel_low(nz2,nx2))
	allocate(model_low(nz2,nx2))			
	allocate(model_high(nz2,nx2))			

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! READ BACKGROUND !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	! READ BACKGROUND
  	rc = sreed("in",model_bg(fft_pad+1:fft_pad+nz,fft_pad+1:fft_pad+nx),4*nz*nx)

	! PAD THE BACKGROUND
    do ix = 1,nx
        do iz = 1,fft_pad
            model_bg(iz,ix+fft_pad)      		= model_bg(fft_pad+1,ix+fft_pad)
            model_bg(fft_pad+nz+iz,ix+fft_pad)  = model_bg(fft_pad+nz,ix+fft_pad)
        end do
    end do

    do ix = 1,fft_pad
	    do iz = 1,nz2
            model_bg(iz,ix)             = model_bg(iz,fft_pad+1)
            model_bg(iz,fft_pad+nx+ix)  = model_bg(iz,fft_pad+nx)
        end do
    end do

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! READ PERTURBATION !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	! COMPUTE THE INDEX OF THE NON-EXTENDED SLICE
	ne_index = (n_ext - 1) / 2 + 2

	! COMPUTE HOW MANY SLICES I NEED TO SKIP 
	sk = sseek_block("in",ne_index-2,4*nz*nx,1)

	! READ PERTURBATION
  	rc = sreed("in",model_pert(fft_pad+1:fft_pad+nz,fft_pad+1:fft_pad+nx),4*nz*nx)
  	
	! PAD THE PERTURBATION
    do ix = 1,nx
        do iz = 1,fft_pad
            model_pert(iz,ix+fft_pad)      		  = model_pert(fft_pad+1,ix+fft_pad)
            model_pert(fft_pad+nz+iz,ix+fft_pad)  = model_pert(fft_pad+nz,ix+fft_pad)
        end do
    end do

    do ix = 1,fft_pad
	    do iz = 1,nz2
            model_pert(iz,ix)             = model_pert(iz,fft_pad+1)
            model_pert(iz,fft_pad+nx+ix)  = model_pert(iz,fft_pad+nx)
        end do
    end do
		
	! SUM THE TWO MODELS
	model_sum = model_pert + model_bg 

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! INITIALIZE FFT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

 	! FFT PARAMETERS FOR Z
	sign1 = 1.0 
  	dkz = 1./((nz2-1)*dz)
	n_left = (nz2-(nz2+1)/2)
	okz = -1.0 * n_left * dkz 	
	write(0,*) "dkz",dkz,"[1/km]"
	
 	! FFT PARAMETERS FOR X
	sign2 = 1.0 
  	dkx = 1./((nx2-1)*dx)
	n_left = (nx2-(nx2+1)/2)
	okx = -1.0 * n_left * dkx 	
	write(0,*) "dkx",dkx,"[1/km]"	
 	
	! APPLY FFT TO MODEL SUM 
	model_sum_fft = model_sum
 	call ft_init(nz2,nx2)
 	call ft1axis(.false.,sign1,model_sum_fft)
 	call ft2axis(.false.,sign2,model_sum_fft) 

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! INITIALIZE FILTER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 	
 	
	! WAVENUMBER CUTOFF
	call from_param("k_cut",k_cut)
	write(0,*) "cutoff wavenumber:",k_cut,"[1/km]"

	! INITIALIZE FILTER TO ZERO
	filter = 0.0

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!! COMPUTE FFT LOW-PASS FILTER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	! INDEX FOR DC COMPONENT
	index_okz = nz2-(nz2+1)/2+1
	index_okx = nx2-(nx2+1)/2+1	

	! COMPUTE LOW-PASS FILTER
	do ikz = 1,nz2

 		kz = (ikz - index_okz) * dkz			
		do ikx = 1,nx2

 			kx = (ikx - index_okx) * dkx
 			k = sqrt(kz*kz+kx*kx) 

			! compute Butterworth filter
			filter(ikz,ikx) = 1 / sqrt(1 + (k/k_cut)**(2*buttord))

		end do

  	end do 	
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!! APPLY FFT LOW-PASS FILTER !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	model_sum_fft_filter = model_sum_fft * filter

	write(0,*) "max model_sum_fft_filter",maxval(cabs(model_sum_fft_filter))	
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! APPLY IFFT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!	

	! APPLY IFFT
 	cmodel_low = model_sum_fft_filter

 	call ft1axis(.true.,sign1,cmodel_low)
 	call ft2axis(.true.,sign2,cmodel_low) 	
	model_low = real(cmodel_low)

	! COMPUTE COMPLEMENT MODEL
  	model_high = model_sum - model_low

	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! WRITE OUTPUT !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	! WRITE LOW-PASSED MODEL
	call sep_write(model_low(fft_pad+1:fft_pad+nz,fft_pad+1:fft_pad+nx))

 	call to_history("n1",nz)
 	call to_history("d1",dz)
 	call to_history("o1",oz)
 	call to_history("label1","z [km]")

 	call to_history("n2",nx)
 	call to_history("d2",dx)
 	call to_history("o2",ox)
 	call to_history("label2","x [km]") 	
 
 	call to_history("n3",1)
 	call to_history("d3",0.0)
 	call to_history("o3",0.0)
 	 
	if (debug) then
	
		write(0,*) "saving intermediate steps"
	
		!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! DEBUG !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 	 
	 
		! WRITE SUM OF TWO MODELS BEFORE FILTERING
		call sep_write(model_sum(fft_pad+1:fft_pad+nz,fft_pad+1:fft_pad+nx),"model_sum")

		call to_history("n1",nz,"model_sum")
		call to_history("d1",dz,"model_sum")
		call to_history("o1",oz,"model_sum")
		call to_history("label1","z [km]","model_sum")

		call to_history("n2",nx,"model_sum")
		call to_history("d2",dx,"model_sum")
		call to_history("o2",ox,"model_sum")
		call to_history("label2","x [km]","model_sum")
 
		call to_history("n3",1,"model_sum")
		call to_history("d3",0.0,"model_sum")
		call to_history("o3",0.0,"model_sum")

		! WRITE FFT BEFORE FILTERING
		call sep_write(cabs(model_sum_fft),"model_fft")
	
		call to_history("n1",nz2,"model_fft")
		call to_history("d1",dkz,"model_fft")
		call to_history("o1",okz,"model_fft")
		call to_history("label1","kz [1/km]","model_fft")
	
		call to_history("n2",nx2,"model_fft")
		call to_history("d2",dkx,"model_fft")
		call to_history("o2",okx,"model_fft")
		call to_history("label2","kx [1/km]","model_fft")

		! WRITE FFT AFTER LOW-PASS FILTER
		call sep_write(cabs(model_sum_fft_filter),"model_fft_filter")
	
		call to_history("n1",nz2,"model_fft_filter")
		call to_history("d1",dkz,"model_fft_filter")
		call to_history("o1",okz,"model_fft_filter")
		call to_history("label1","kz [1/km]","model_fft_filter")
	 
		call to_history("n2",nx2,"model_fft_filter")
		call to_history("d2",dkx,"model_fft_filter")
		call to_history("o2",okx,"model_fft_filter")
		call to_history("label2","kx [1/km]","model_fft_filter")
	 
		! WRITE FILTER
		call sep_write(filter,"filter")
	
		call to_history("n1",nz2,"filter")
		call to_history("d1",dkz,"filter")
		call to_history("o1",okz,"filter")
	
		call to_history("n2",nx2,"filter")
		call to_history("d2",dkx,"filter")
		call to_history("o2",okx,"filter")
	
		! WRITE HIGH-PASSED MODEL
		call sep_write(model_high(fft_pad+1:fft_pad+nz,fft_pad+1:fft_pad+nx),"model_high")
	
		call to_history("n1",nz,"model_high")
		call to_history("d1",dz,"model_high")
		call to_history("o1",oz,"model_high")
		call to_history("label1","z [km]","model_high")

		call to_history("n2",nx,"model_high")
		call to_history("d2",dx,"model_high")
		call to_history("o2",ox,"model_high")
		call to_history("label2","x [km]","model_high") 
		
	end if			

 	call sep_close()

end program 

